import re
from datetime import date

class product_builder(object):
    def __init__(self, lines, *args, **kwargs):
        """Initialize a product builder, and wrap lines in a decorator that lets 
        you add calculated fields before combining them.
        
        Set up a mock line object.
        >>> class order_line(object): pass
        
        Wrapping a line still lets you read values from the original line.
        >>> raw_line = order_line()
        >>> raw_line.name = '[ABC] Alphabet blocks'
        >>> builder = product_builder([raw_line])
        >>> wrapped_lines = builder.input_lines
        >>> len(wrapped_lines)
        1
        >>> wrapped_line = wrapped_lines[0]
        >>> wrapped_line.name 
        '[ABC] Alphabet blocks'
        
        However, you can add fields to the wrapped line without affecting
        the original.
        >>> wrapped_line.promise_date = date(2011, 11, 13)
        >>> wrapped_line.promise_date
        datetime.date(2011, 11, 13)
        >>> hasattr(raw_line, 'promise_date')
        False
        
        You can replace existing field values without affecting the original.
        >>> raw_line.planned_date = date(2011, 11, 10)
        >>> wrapped_line.planned_date
        datetime.date(2011, 11, 10)
        >>> wrapped_line.planned_date = date(2011, 11, 12)
        >>> wrapped_line.planned_date
        datetime.date(2011, 11, 12)
        >>> raw_line.planned_date
        datetime.date(2011, 11, 10)
        """
        super(product_builder, self).__init__(*args, **kwargs)

        class line_wrapper(object):
            def __init__(self, content):
                self.__content = content
                
            def __getattr__(self, name):
                return getattr(self.__content, name)

        self.input_lines = [line_wrapper(line) for line in lines]
        self.fields_to_match = []
        self.fields_to_sum = []
        self.fields_to_max = []
        self.fields_to_concatenate = []
        self.fields_to_hyphenate = []
        
    def __split_name(self, product_name):
        """ Split a name from "[code] description" into code and description.
        
        Returns a tuple ("code", "description") or (None, None) if no match.
        """
        name_pattern = (
            r'^\[([^\[\]]*)\]\s*' # [code] plus optional whitespace
            r'([^\[\]]*)$') # anything else, except square brackets
        match = re.match(name_pattern, product_name)
        return match and (match.group(1), match.group(2)) or (None, None)
        
    def combine(self):
        """ Combine any grouped lines in self.input_lines.
        
        This class lets you group order lines. It assumes the following fields
        are on each line:
        grouping - a string value. Anything before the period is the group name
            and anything after it is ignored. Lines should already be sorted by
            grouping before they are passed in to this method.
        name - Must be of the format "[code] description" for any lines you want
            to combine. Combined lines have a name format 
            "[code1-code2] description1 with description2".
        
        Other fields must be specified before calling this method.

        Set up a mock product, and the builder object.
        >>> class mock(object): pass
        >>> def create_product():
        ...     product = mock()
        ...     product.type = 'product'
        ...     product.categ_id = mock()
        ...     product.categ_id.append_with = ''
        ...     return product
        
        If grouping is null or blank, don't combine.
        >>> builder = product_builder([object(), object()])
        >>> line1, line2 = builder.input_lines
        >>> line1.name = '[AB90] Angle bracket'
        >>> line1.product_id = create_product()
        >>> line1.grouping = None
        >>> line2.name = '[PS03] Power supply'
        >>> line2.product_id = create_product()
        >>> line2.grouping = None
        >>> builder.combine()
        [display_line('[AB90] Angle bracket'), display_line('[PS03] Power supply')]
        
        
        If grouping is different, don't combine.
        >>> line1.grouping = 'A'
        >>> line2.grouping = 'B'
        >>> builder.combine()
        [display_line('[AB90] Angle bracket'), display_line('[PS03] Power supply')]
        
        Assumes lines are already sorted by grouping, so won't reorder them.
        >>> line1.grouping = 'B'
        >>> line2.grouping = 'A'
        >>> builder.combine()
        [display_line('[AB90] Angle bracket'), display_line('[PS03] Power supply')]
        
        If group matches, combine them.
        >>> line1.grouping = 'A.1'
        >>> line2.grouping = 'A.2'
        >>> builder.combine()
        [display_line('[AB90-PS03] Angle bracket with Power supply')]
        >>> builder.warnings
        ''
        
        If name doesn't match expected format, it can't be combined.
        >>> builder = product_builder([line1, object()])
        >>> line1, bad_name_line = builder.input_lines 
        >>> bad_name_line.name = '(PS03) Power supply' #no [square brackets]
        >>> bad_name_line.product_id = create_product()
        >>> line1.grouping = 'A.1'
        >>> bad_name_line.grouping = 'A.2'
        >>> builder.combine()
        [display_line('[AB90] Angle bracket'), display_line('(PS03) Power supply')]
        >>> builder.warnings
        "No [product code] found in '(PS03) Power supply'."
        
        If group includes more than four items, fail to combine them.
        >>> builder = product_builder([line1, line2, object(), object(), object()])
        >>> line1, line2, line3, line4, line5 = builder.input_lines
        >>> line3.name = '[PMF3] Power adapter'
        >>> line3.product_id = create_product()
        >>> line3.grouping = 'A.3'
        >>> line4.name = '[AB90] Angle bracket'
        >>> line4.product_id = create_product()
        >>> line4.grouping = 'A.4'
        >>> line5.name = '[DC06] Data cable'
        >>> line5.product_id = create_product()
        >>> line5.grouping = 'A.5'
        >>> len(builder.combine())
        5
        >>> builder.warnings
        "More than four lines in group 'A'."
        
        To combine, both products must be stockable products.
        >>> builder = product_builder([line1, object()])
        >>> line1, service_line = builder.input_lines
        >>> service_line.name = '[CLEAN] Office cleaning'
        >>> service_line.product_id = create_product()
        >>> service_line.product_id.type = 'service'
        >>> service_line.grouping = 'A.2'
        >>> [group.name for group in builder.combine()]
        ['[AB90] Angle bracket', '[CLEAN] Office cleaning']
        >>> builder.warnings
        "No stockable product found when grouping '[CLEAN] Office cleaning'."
        
        A product category can restrict which accessories it may be grouped 
        with. It will combine with products that match the restriction.
        >>> builder = product_builder([object(), object()])
        >>> picky_line, good_line = builder.input_lines
        >>> picky_line.name = '[T-PCK] Picky product'
        >>> picky_line.product_id = create_product()
        >>> picky_line.product_id.categ_id.append_with = 'GOOD'
        >>> picky_line.grouping = 'A.1'
        >>> good_line.name = '[GOOD] good accessory'
        >>> good_line.product_id = create_product()
        >>> good_line.grouping = 'A.2'
        >>> [group.name for group in builder.combine()]
        ['[T-PCK-GOOD] Picky product with good accessory']
        
        It will not combine with products that don't match.
        >>> builder = product_builder([picky_line, object()])
        >>> picky_line, bad_line = builder.input_lines
        >>> bad_line.name = '[BAD] bad accessory'
        >>> bad_line.product_id = create_product()
        >>> bad_line.grouping = 'A.2'
        >>> [group.name for group in builder.combine()]
        ['[T-PCK] Picky product', '[BAD] bad accessory']
        >>> builder.warnings
        "Cannot append product 'BAD' to product 'T-PCK'."
        
        If group includes three or four items, check append_with against
        combined accessory items.
        >>> builder = product_builder([picky_line, line2, line3, line4])
        >>> picky_line.product_id.categ_id.append_with = 'PS03-PMF3-AB90'
        >>> [group.name for group in builder.combine()]
        ['[T-PCK-PS03-PMF3-AB90] Picky product with Power supply, Power adapter, Angle bracket']
        >>> builder.warnings
        ''
        
        Can add fields that must be matched within a group.
        >>> line1.discount = 5.0
        >>> line2.discount = 5.0
        
        They just get copied for ungrouped lines.
        >>> builder = product_builder([line1])
        >>> builder.add_fields_to_match(['discount'])
        >>> [group.discount for group in builder.combine()]
        [5.0]
        
        Also copied to group when they match.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_match(['discount'])
        >>> [group.discount for group in builder.combine()]
        [5.0]
        >>> builder.warnings
        ''
        
        Don't group if they don't match.
        >>> builder = product_builder([line1, object()])
        >>> builder.add_fields_to_match(['discount'])
        >>> line1, bad_discount_line = builder.input_lines
        >>> bad_discount_line.discount = 10.0
        >>> bad_discount_line.name = '[PMF3] Power adapter'
        >>> bad_discount_line.product_id = create_product()
        >>> bad_discount_line.grouping = 'A.2'
        >>> [group.discount for group in builder.combine()]
        [5.0, 10.0]
        >>> builder.warnings
        "Unmatched field in group 'A': discount."
        
        Can add fields to sum.
        >>> line1.price_unit = 1.00
        >>> line2.price_unit = 2.00
        >>> line1.price_subtotal = 5.00
        >>> line2.price_subtotal = 10.00
        
        They just get copied for ungrouped lines.
        >>> builder = product_builder([line1])
        >>> builder.add_fields_to_sum(['price_unit', 'price_subtotal'])
        >>> [(group.price_unit, group.price_subtotal) 
        ... for group in builder.combine()]
        [(1.0, 5.0)]
        
        They get summed for grouped lines.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_sum(['price_unit', 'price_subtotal'])
        >>> [(group.price_unit, group.price_subtotal) 
        ... for group in builder.combine()]
        [(3.0, 15.0)]
        
        Can add fields to select the maximum.
        >>> line1.delay = 1
        >>> line2.delay = 2
        >>> line1.promise_date = date(2011, 11, 1)
        >>> line2.promise_date = date(2011, 11, 2)
        
        They just get copied for ungrouped lines.
        >>> builder = product_builder([line1])
        >>> builder.add_fields_to_max(['delay', 'promise_date'])
        >>> [(group.delay, group.promise_date) 
        ... for group in builder.combine()]
        [(1, datetime.date(2011, 11, 1))]
        
        Choose the maximum from grouped lines.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_max(['delay', 'promise_date'])
        >>> [(group.delay, group.promise_date) 
        ... for group in builder.combine()]
        [(2, datetime.date(2011, 11, 2))]
        
        Can add fields to concatenate.
        >>> line1.notes = 'First note'
        >>> line2.notes = 'Second note'
        
        They just get copied for ungrouped lines.
        >>> builder = product_builder([line1])
        >>> builder.add_fields_to_concatenate(['notes'])
        >>> '---'.join([group.notes for group in builder.combine()])
        'First note'
        
        Concatenate for grouped lines.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_concatenate(['notes'])
        >>> print '---'.join([group.notes for group in builder.combine()])
        First note
        <BLANKLINE>
        Second note
        
        If only product has notes, then leave it untouched.
        >>> line1.notes = 'Only note on product'
        >>> line2.notes = None
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_concatenate(['notes'])
        >>> print '---'.join([group.notes for group in builder.combine()])
        Only note on product
        
        If only accessory has notes, then use that.
        >>> line1.notes = None
        >>> line2.notes = 'Only note on accessory'
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_concatenate(['notes'])
        >>> print '---'.join([group.notes for group in builder.combine()])
        Only note on accessory
        
        Don't build up extra blank lines
        >>> line1.notes = '''First note
        ... '''
        >>> line2.notes = 'Second note'
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_concatenate(['notes'])
        >>> print '---'.join([group.notes for group in builder.combine()])
        First note
        <BLANKLINE>
        Second note
        
        Can add fields to hyphenate.
        >>> line1.code = 'AB90'
        >>> line2.code = 'PS03'
        
        They just get copied for ungrouped lines.
        >>> builder = product_builder([line1])
        >>> builder.add_fields_to_hyphenate(['code'])
        >>> [group.code for group in builder.combine()]
        ['AB90']
        
        Hyphenate for grouped lines.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_hyphenate(['code'])
        >>> [group.code for group in builder.combine()]
        ['AB90-PS03']
        
        If only product has code, then leave it untouched.
        >>> line2.code = None
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_hyphenate(['code'])
        >>> [group.code for group in builder.combine()]
        ['AB90']
        
        If only accessory has code, then use that.
        >>> line1.code = None
        >>> line2.code = 'PS03'
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_hyphenate(['code'])
        >>> [group.code for group in builder.combine()]
        ['PS03']
        
        Can also hyphenate numerical fields.
        >>> line1.revision = 1
        >>> line2.revision = 2
        
        Hyphenate for grouped lines.
        >>> builder = product_builder([line1, line2])
        >>> builder.add_fields_to_hyphenate(['revision'])
        >>> [group.revision for group in builder.combine()]
        ['1-2']
        """
        fields_to_copy = (
            self.fields_to_match +
            self.fields_to_sum + 
            self.fields_to_max + 
            self.fields_to_concatenate +
            self.fields_to_hyphenate +
            ['product_id'])
        sorted = [(line.grouping, line) for line in self.input_lines]
        groups = []
        self.warnings = ''
        for grouping, line in sorted:
            display = display_line(line.name)
            if grouping:
                group, sep, suffix = grouping.partition('.')
            else: 
                group = None
            previous_group, previous_displays = groups[-1] if groups else (None, None)
            if grouping and (previous_group == group):
                previous_displays.append(display)
            else:
                groups.append((group, [display]))
            
            for field_name in fields_to_copy:
                setattr(display, field_name, getattr(line, field_name))
        
        result = []
        for group, displays in groups:
            if len(displays) < 2 or 4 < len(displays):
                result.extend(displays)
                if 4 < len(displays):
                    self.warnings = "More than four lines in group '%s'." % group
            else:
                insertion_point = len(result)
                for i in range(len(displays)-1, 0, -1):
                    product1 = displays[i-1]
                    product2 = displays[i]
                    are_accessories = i > 1
                    self.__combine_products(
                        product1, 
                        product2, 
                        are_accessories,
                        group)
                    if self.warnings:
                        result.insert(insertion_point, product2)
                result.insert(insertion_point, displays[0])
        return result
    
    def __combine_products(self, product1, product2, are_accessories, group):
        """ Combine one product1 into another.
        
        If the products are compatible, then fields from product2 are merged
        into product1. If they are not compatible, then they are left unchanged
        and a warning is added to self.warnings.
        are_accessories is true if both products are accessories.
        """ 
        product1_code, product1_name = self.__split_name(product1.name)
        product2_code, product2_name = self.__split_name(product2.name)
        for field_name in self.fields_to_match:
            product1_value = getattr(product1, field_name)
            product2_value = getattr(product2, field_name)
            if product1_value != product2_value:
                self.warnings = (
                    "Unmatched field in group '%s': %s." %
                    (group, field_name))
                break
        if not product1_code:
            self.warnings = (
                "No [product code] found in '%s'." % product1.name)
        elif not product2_code:
            self.warnings = (
                "No [product code] found in '%s'." % product2.name)
        elif product1.product_id.type != 'product':
            self.warnings = (
                "No stockable product found when grouping '%s'." %
                product1.name)
        elif product2.product_id.type != 'product':
            self.warnings = (
                "No stockable product found when grouping '%s'." %
                product2.name)
        elif not are_accessories and not self.can_append(
            product1.product_id.categ_id.append_with, 
            product2_code):
            self.warnings = (
                "Cannot append product '%s' to product '%s'." %
                (product2_code, product1_code))
        if not self.warnings:
            if are_accessories:
                product1.name = '[%s-%s] %s, %s' % (
                    product1_code,
                    product2_code,
                    product1_name,
                    product2_name)
            else:
                product1.name = '[%s-%s] %s with %s' % (
                    product1_code,
                    product2_code,
                    product1_name,
                    product2_name)
    
            for field_name in self.fields_to_sum:
                product1_value = getattr(product1, field_name)
                product2_value = getattr(product2, field_name)
                setattr(
                    product1, 
                    field_name, 
                    product1_value + product2_value)
            for field_name in self.fields_to_max:
                product1_value = getattr(product1, field_name)
                product2_value = getattr(product2, field_name)
                setattr(
                    product1, 
                    field_name, 
                    max(product1_value, product2_value))
            for field_name in self.fields_to_concatenate:
                product1_value = getattr(product1, field_name)
                product2_value = getattr(product2, field_name)
                if product1_value and product2_value:
                    setattr(
                        product1, 
                        field_name, 
                        product1_value.rstrip() + 
                            '\n\n' + 
                            product2_value)
                elif product2_value:
                    setattr(
                        product1, 
                        field_name, 
                        product2_value)
            for field_name in self.fields_to_hyphenate:
                product1_value = getattr(product1, field_name)
                product2_value = getattr(product2, field_name)
                if product1_value and product2_value:
                    setattr(
                        product1, 
                        field_name, 
                        str(product1_value) + '-' + str(product2_value))
                elif product2_value:
                    setattr(
                        product1, 
                        field_name, 
                        str(product2_value))
                elif product1_value != None:
                    setattr(
                        product1,
                        field_name,
                        str(product1_value))
    
    def can_append(self, append_with, accessory_code):
        """ Check whether an accessory can be appended to a product.
        
        append_with - restrictions for the product category about which
            accessories can be appended.
        accessory_code - the product code for the accessory that will be 
            appended.
        
        Returns True if the accessory can be appended, otherwise False.
        
        If append_with is empty, then any accessory_code is allowed.
        >>> builder = product_builder([])
        >>> builder.can_append('', 'FOO')
        True
        
        Simplest restriction is a single product code.
        >>> builder.can_append('GOOD', 'GOOD')
        True
        >>> builder.can_append('GOOD', 'BAD')
        False
        
        Can use SQL wild cards.
        >>> builder.can_append('GO_D', 'GOAD')
        True
        >>> builder.can_append('GO_D', 'GOAT')
        False
        >>> builder.can_append('GO_D', 'GORED')
        False
        >>> builder.can_append('GO%D', 'GORED')
        True
        >>> builder.can_append('GO%D', 'GOAT')
        False
        
        Regular expression wild cards don't work.
        >>> builder.can_append('GO.D', 'GOAD')
        False
        >>> builder.can_append('GO.D', 'GO.D')
        True
        
        Can specify more than one possible match as comma-separated list.
        >>> builder.can_append('GOOD,GREAT', 'GOOD')
        True
        >>> builder.can_append('GOOD,GREAT', 'GREAT')
        True
        >>> builder.can_append('GOOD,GREAT', 'BAD')
        False
        
        Ignore white space around commas.
        >>> builder.can_append('GOOD, GREAT ,  AMAZING', 'GREAT')
        True
        """
        if not append_with:
            return True
        for option in re.split('\s*,\s*', append_with):
            pattern = re.escape(option).replace('\_', '.').replace('\%', '.*')
            if re.match(pattern, accessory_code):
                return True
        return False

    def add_fields_to_match(self, field_names):
        """ Add fields that must match before lines can combine.
        """
        self.fields_to_match.extend(field_names)
    
    def add_fields_to_sum(self, field_names):
        """ Add fields to sum up on combined lines.
        """
        self.fields_to_sum.extend(field_names)

    def add_fields_to_max(self, field_names):
        """ Add fields to select the maximum from on combined lines.
        """
        self.fields_to_max.extend(field_names)

    def add_fields_to_concatenate(self, field_names):
        """ Add string fields to concatenate on combined lines.
        """
        self.fields_to_concatenate.extend(field_names)

    def add_fields_to_hyphenate(self, field_names):
        """ Add string fields to hyphenate on combined lines.
        """
        self.fields_to_hyphenate.extend(field_names)

class display_line:
    """ Holds the relevant details from a single sales order line, or two
    combined lines.
    
    >>> display_line('AB90')
    display_line('AB90')
    """
    def __init__(self, name):
        self.name = name
        
    def __repr__(self):
        return 'display_line(%r)' % self.name

if __name__ == "__main__":
    import doctest
    failures, tests = doctest.testmod()
    if failures == 0:
        print "Success."
