from osv import fields, osv
from tools.translate import _

class sale_order(osv.osv):
    _inherit = 'sale.order'
    
    def action_wait(self, cr, uid, ids, context=None):
        """ Seems like there's multiple ways that this function gets called.
            One way is when the user has clicked on the "Confirm Order" button. 
            This is used by the workflow system to determine if we should move 
            to the wait_ship or wait_invoice state
            This class adds a check that line items can be properly combined.
        """
    
        for o in self.browse(cr, uid, ids):
            self._combined_lines(o, shouldValidate=True)
        
        return super(sale_order, self).action_wait(cr, uid, ids, context)
            

    def _prepare_line_to_combine(self, line):
        """ Prepare a sales order line before it gets combined.
        
        Copy fields from related tables.
        """
        if line.product_uos:
            line.quantity = line.product_uos_qty
            line.uom_name = line.product_uos.name
        else:
            line.quantity = line.product_uom_qty
            line.uom_name = line.product_uom.name

    def _prepare_product_builder(self, builder):
        """ Configure product builder with fields to match, copy, etc. """
        builder.add_fields_to_match(['tax_id', 
                                     'quantity', 
                                     'uom_name', 
                                     'discount'])
        builder.add_fields_to_max(['delay'])
        builder.add_fields_to_sum(['price_unit', 'price_subtotal'])
        builder.add_fields_to_concatenate(['notes'])

    def _combined_lines(self, order, shouldValidate=False):
        """ Try to combine the lines in a sales order.
        
        order is a sale.order browse object.
        shouldValidate is True if this should throw an exception when lines
            cannot be combined according to the grouping field.
        
        Returns a list of display lines.
        """
        product_obj = self.pool.get('product.product')
        
        # Calculate some fields before grouping
        builder = product_obj._create_builder(order.order_line)
        self._prepare_product_builder(builder)
        for line in builder.input_lines:
            self._prepare_line_to_combine(line)
        
        #Now group
        combined_lines = builder.combine()
        if builder.warnings:
            raise osv.except_osv(_('Invalid grouping!'), builder.warnings)
        return combined_lines
    
    def _prepare_order_picking(self, cr, uid, order, context=None):
        """ Because the order check-list now displays the notes from the sales
        order as well as the notes from the picking, we no longer need to copy
        the notes from the sales order onto the picking.
        """
        data = super(sale_order, self)._prepare_order_picking(cr,
                                                              uid,
                                                              order,
                                                              context)
        del data['note']
        return data
    
    def _prepare_order_line_move(self, 
                                 cr, 
                                 uid, 
                                 order, 
                                 line, 
                                 picking_id, 
                                 date_planned, 
                                 context=None):
        """ Prepare a stock move as part of the stock picking for a sales
        order line.
        
        This version adds the grouping field to the base superclass 
        version. It also stops the note from being copied, because the order
        check-list now displays the stock move note as well as the sales order
        line note.
        
        returns a dictionary of field values.
        """
        data = super(sale_order, self)._prepare_order_line_move(cr, 
                                                                uid, 
                                                                order, 
                                                                line, 
                                                                picking_id, 
                                                                date_planned, 
                                                                context)
        data['grouping'] = line.grouping
        del data['note']
        return data

sale_order()

class sale_order_line(osv.osv):
    _inherit = 'sale.order.line'
    
    _columns = {
        'grouping': fields.char(
            'Grouping',
            size=20,
            help='Used to sort line items alphabetically and group them. '
                'Before the period is the group name, the suffix orders within '
                'a group.'),
    }
    _order = 'grouping, id'

    def _prepare_order_line_invoice_line(self, 
                                         cr, 
                                         uid, 
                                         line, 
                                         account_id=False, 
                                         context=None):
        """ Copy the grouping field from the sales order line to the dict of
           values that's being used to create the new invoice line for a
           sale order line.

           :param browse_record line: sale.order.line record to invoice
           :param int account_id: optional ID of a G/L account to force
               (this is used for returning products including service)
           :return: dict of values to create() the invoice line
        """
        vals = super(sale_order_line, self)._prepare_order_line_invoice_line(
            cr,
            uid,
            line,
            account_id,
            context)
        if vals:
            vals['grouping'] = line.grouping
        return vals

sale_order_line()
