from datetime import date, timedelta, datetime
from osv import fields, osv
import math
import date_math
from sales_set import sales_set
import netsvc
from product_snapshot import Snapshot
from sim_clock import sim_clock
from sim_product import sim_product
from stockout_snapshot import stockout_snapshot

class product_template(osv.osv):
    _inherit = "product.template"

    _columns = {
        'variant_ids':fields.one2many('product.product', 'product_tmpl_id', 'Variants'),
        'is_kit':fields.boolean(
                      'Is Kit',
                      help='Products that are kits are not included in the Likely Product Stockouts view.'
                  ),
        'lead_time_qty' : fields.float(
            string='Lead time quantity',
            help='Amount that we should be able to sell without exceeding ' +
                'customer lead time. Used for predicting likely stock outs.'),
        'categ_name': fields.related('categ_id',
                                     'name',
                                     type='char',
                                     string='Category',
                                     readonly=True)
    }
    _defaults = {
        'is_kit': lambda *a: False,
        'lead_time_qty': lambda *a: 5.0
    }
    
    def copy(self, cr, uid, id, default=None, context=None):
        if default is None:
            default = {}
        default = default.copy()
        default.update({'variant_ids':False,})
        return super(product_template, self).copy(cr, uid, id, default, context)

product_template()

class product_product(osv.osv):
    _inherit = "product.product"
    
    def _get_product_components(self, cr, uid, snapshots, all_product_types=False):
        """Fill in the component details of a dictionary of snapshots
        
        snapshots contains a dictionary of ids and snapshot objects.
        If all_product_types is false, then services and consumables will not be
        included as components, only products of type 'product' will be included.
        On return, the components and qty_produced attributes will
        be set, along with is_slow_to_produce, daily_volume, uom, product_type,
        delay, safety_days, and standard_price.
        """
        
        bom_obj = self.pool.get('mrp.bom')
        uom_obj = self.pool.get('product.uom')
        ending_products = {}
        
        for product in self.browse(cr, uid, snapshots.keys()):
            snapshot = snapshots[product.id]
            snapshot.product_tmpl_id = product.product_tmpl_id
            snapshot.is_slow_to_produce = product.produce_delay > 5.0
            snapshot.can_be_sold = product.sale_ok
            snapshot.daily_sales = product.daily_sales
            snapshot.uom = product.product_tmpl_id.uom_id
            snapshot.product_type = product.type
            snapshot.supply_method = product.product_tmpl_id.supply_method
            snapshot.delay = (product.seller_delay + product.produce_delay)
            snapshot.safety_days = product.safety_days
            snapshot.standard_price = product.product_tmpl_id.standard_price
            
            if product.state == 'end' and product.supply_method == 'buy':
                ending_products[product.default_code] = snapshot

        for product in self.browse(cr, uid, snapshots.keys()):
            if product.state == 'sellable':
                ending_product = ending_products.get(product.default_code)
                if ending_product:
                    ending_product.replacement = snapshots[product.id]
        
        duplicate_bom_product_ids = set()
        bom_line_ids = bom_obj.search(
            cr,
            uid,
            [('bom_id', '!=', None)])
        for bom_line in bom_obj.browse(cr, uid, bom_line_ids):
            bom = bom_line.bom_id
            if not bom.active or not bom.product_id or not bom_line.product_id:
                continue
            product_id = bom.product_id.id
            snapshot = snapshots[product_id]
            if snapshot.supply_method == 'buy':
                continue
            if snapshot.bom_id == None:
                snapshot.bom_id = bom.id
                snapshot.qty_produced = uom_obj._compute_qty_obj(
                    cr,
                    uid,
                    bom.product_uom,
                    bom.product_qty,
                    snapshot.uom)
            else:
                if snapshot.bom_id != bom.id:
                    duplicate_bom_product_ids.add(product_id)
                    continue
            component_id = bom_line.product_id.id
            component_snapshot = snapshots[component_id]
            if component_snapshot.product_type == 'product' or all_product_types:
                snapshot.components.append((
                    component_id, 
                    bom_line.product_qty, 
                    bom_line.product_uom))
        for product_id in duplicate_bom_product_ids:
            snapshot = snapshots[product_id]
            bom_id = bom_obj._bom_find(
                cr, 
                uid, 
                product_id, 
                snapshot.uom.id)
            if bom_id:
                bom = bom_obj.browse(cr, uid, bom_id)
                properties = []
                exploded = bom_obj._bom_explode(cr, uid, bom, 1, properties)
                component_lines = exploded[0]
    
                snapshot.qty_produced = uom_obj._compute_qty_obj(
                    cr,
                    uid,
                    bom.product_uom,
                    bom.product_qty,
                    snapshot.uom)
                snapshot.components = []
                for component_line in component_lines:
                    component_id = component_line['product_id']
                    component_snapshot = snapshots[component_id]
                    if (component_snapshot.product_type == 'product' or 
                        all_product_types):
                        component_uom = uom_obj.browse(
                            cr, 
                            uid, 
                            component_line['product_uom'])
                        snapshot.components.append((
                            component_id, 
                            component_line['product_qty'], 
                            component_uom))
    
    def _calculate_build_capacity(
        self,
        cr,
        uid,
        product_id,
        snapshots,
        committed,
        context):
        """Walk the build hierarchy and calculate build capacity.
        
        If there are no components for the requested product, then the build
        capacity is zero.
        If there are any components listed for the requested product in
        its snapshot, then this method will call itself recursively.
        A product's build capacity is the minimum quantity available among all
        its components. See the code below for details of the calculation.
        Exception: a snapshot that is_slow_to_produce always has a build 
        capacity of zero.
        The committed parameter holds the virtual stock plus the committed stock
        for whichever consumer is calling this method (product_id). That value
        will be passed down the build hierarchy and incremented as it goes.
        Use the snapshot.state to avoid infinite loops. If the snapshot's state
        is CALCULATED, then we repeat the calculation and pass the committed
        stock values down the hierarchy.
        If it's CALCULATING, then there's an infinite loop and we
        just return zero. If it's NEW, then we set the state to CALCULATING,
        do the calculation, and then update the snapshot.
        """
        snapshot = snapshots[product_id]
        if snapshot.state == Snapshot.CALCULATED and committed == 0.0:
            return snapshot.build_capacity
        if snapshot.state == Snapshot.CALCULATING:
            return 0.0
        
        snapshot.state = Snapshot.CALCULATING
        if committed == None:
            product_committed = snapshot.virtual_available
        else:
            product_committed = committed
        uom_obj = self.pool.get('product.uom')
        components_available = []
        for component_id, qty_consumed, uom_consumed in snapshot.components:
            component_snapshot = snapshots[component_id]
            try:
                component_committed = uom_obj._compute_qty_obj(
                    None,
                    None,
                    uom_consumed,
                    product_committed / snapshot.qty_produced * qty_consumed,
                    component_snapshot.uom)
            except:
                product = self.browse(cr, uid, product_id, context=context)
                component = self.browse(cr, uid, component_id, context=context)
                raise osv.except_osv(
                    'Error !', 
                    ("The BOM for product '%s' uses UOM %s for component " +
                     "'%s', which is incompatible with the component's " +
                     "default UOM of %s.") %
                    (product.default_code, 
                     uom_consumed.name, 
                     component.default_code,
                     component_snapshot.uom.name))
            component_build_capacity = self._calculate_build_capacity(
                cr,
                uid,
                component_id,
                snapshots,
                component_committed,
                context)
            
            qty_avail = (
                component_snapshot.virtual_available - 
                component_snapshot.production_qty - 
                component_snapshot.purchase_qty + 
                component_build_capacity)
            adjusted_qty_avail = math.floor(uom_obj._compute_qty_obj(
                None,
                None,
                component_snapshot.uom,
                qty_avail,
                uom_consumed) / qty_consumed) * snapshot.qty_produced
            components_available.append(adjusted_qty_avail)
        if snapshot.is_slow_to_produce:
            snapshot.build_capacity = 0
        else:
            snapshot.build_capacity = (
                components_available and min(components_available)) or 0.0
        snapshot.sales_capacity = (
                snapshot.virtual_available - 
                snapshot.production_qty - 
                snapshot.purchase_qty +
                snapshot.build_capacity)
        if committed != None:
            snapshot.committed += committed

        snapshot.state = Snapshot.CALCULATED
        return snapshot.build_capacity
    
    def _calculate_all_build_capacities(
        self, 
        cr,
        uid, 
        context=None):
        """ Calculate build capacities for all products.
        
        """
        
        all_product_ids = self.search(cr, uid, [])
        snapshots = dict(
            (id, Snapshot(id)) for id in all_product_ids)
        self._get_product_components(cr, uid, snapshots)
        
        for snapshot in snapshots.values():
            self._count_procurements(cr, uid, snapshot)

        product_fields = self._product_available(
            cr, 
            uid, 
            all_product_ids, 
            ['virtual_available'])
        for (product_id, fields) in product_fields.iteritems():
            snapshots[product_id].virtual_available = fields['virtual_available']
        del product_fields
        
        for product_id in all_product_ids:
            self._calculate_build_capacity(
                cr,
                uid,
                product_id,
                snapshots,
                None,
                context)
        
        change_count = 0
        chunk_size = 100
        for i in range(0, len(all_product_ids), chunk_size):
            product_ids = all_product_ids[i:i+chunk_size]
            for product in self.browse(cr, uid, product_ids):
                snapshot = snapshots[product.id]
                if snapshot.has_changed(product):
                    change_count += 1
                    self.write(
                        cr,
                        uid,
                        product.id,
                        snapshot.get_data(),
                        context)
        
        return change_count
    
    def _calculate_all_production_costs(
        self, 
        cr,
        uid, 
        context=None):
        """ Calculate build capacities for all products.
        
        """
        
        template_obj = self.pool.get('product.template')
        uom_obj = self.pool.get('product.uom')
        
        all_product_ids = self.search(cr, uid, [])
        snapshots = dict(
            (id, Snapshot(id)) for id in all_product_ids)
        self._get_product_components(cr, uid, snapshots, all_product_types=True)
        
        change_count = 0
        produced_ids = self.search(cr, uid, [('supply_method', '=', 'produce')])
        for product_id in produced_ids:
            snapshot = snapshots[product_id]
            cost = self._calculate_production_cost(
                cr,
                uid,
                product_id, 
                snapshots, 
                uom_obj)
            if abs(cost - snapshot.standard_price) > 0.0001:
                change_count += 1
                template_obj.write(
                    cr,
                    uid,
                    snapshot.product_tmpl_id.id,
                    {'standard_price': cost},
                    context)
        
        return change_count
    
    def _calculate_production_cost(
        self, 
        cr, 
        uid, 
        product_id, 
        snapshots, 
        uom_obj):
        snapshot = snapshots[product_id]
        if hasattr(snapshot, 'production_cost'):
            return snapshot.production_cost
        if not snapshot.components:
            return snapshot.standard_price
        
        snapshot.production_cost = 0
        for component_id, qty_consumed, consumed_uom in snapshot.components:
            component = snapshots[component_id]
            component_cost = self._calculate_production_cost(
                cr,
                uid,
                component_id, 
                snapshots, 
                uom_obj)
            snapshot.production_cost += uom_obj._compute_qty_obj(
                cr,
                uid,
                consumed_uom,
                qty_consumed,
                component.uom) * component_cost
        return snapshot.production_cost
    
    def _calculate_all_stockouts(
        self, 
        cr,
        uid, 
        context=None):
        """ Calculate potential stockouts for all products.
        
        """
        
        all_product_ids = self.search(cr, uid, [])
        snapshots = dict(
            (id, Snapshot(id)) for id in all_product_ids)
        self._get_product_components(cr, uid, snapshots)
        
        clock = sim_clock(self.pool.get('product.uom'))
        for snapshot in snapshots.itervalues():
            product = sim_product(
                snapshot.product_id,
                daily_sales=snapshot.daily_sales,
                uom=snapshot.uom,
                delay=snapshot.delay,
                safety_days=snapshot.safety_days)
            clock.add_product(product)
        
        for snapshot in snapshots.itervalues():
            product = clock.get_product(snapshot.product_id)
            product.set_production(
                snapshot.bom_id, 
                snapshot.qty_produced, 
                snapshot.uom)
            for component_id, qty_consumed, consumed_uom in snapshot.components:
                component = clock.get_product(component_id)
                product.add_component(component, qty_consumed, consumed_uom)
            if snapshot.replacement:
                product.replacement = clock.get_product(
                    snapshot.replacement.product_id)

        del snapshots
        
        product_fields = self._product_available(
            cr, 
            uid, 
            all_product_ids, 
            ['qty_available'])
        for (product_id, fields) in product_fields.iteritems():
            product = clock.get_product(product_id)
            qty_available = fields['qty_available']
            # postpone intermittent sales so they don't all happen today.
            if product.uom.name == 'PCE':
                qty_available += 0.5
            clock.add_product(product, start_qty=qty_available)

        del product_fields
        
        self._count_expected(cr, uid, clock)
        self._count_sales(cr, uid, clock)
        
        #clock.monitored_product_id = 4

        forecast = clock.run()
        
        if clock.monitored_product_id:
            print clock.monitor_log
        
        component_ids = []
        change_count = 0
        chunk_size = 100
        for i in range(0, len(all_product_ids), chunk_size):
            product_ids = all_product_ids[i:i+chunk_size]
            for product in self.browse(cr, uid, product_ids):
                data = forecast.get_data(product)
                if data:
                    change_count += 1
                    self.write(
                        cr,
                        uid,
                        product.id,
                        data,
                        context)
                if not clock.monitored_product_id:
                    self._update_stockouts(
                        cr, 
                        uid, 
                        product, 
                        clock, 
                        component_ids, 
                        context)
                if len(component_ids) >= 100:
                    self._delete_component_stockouts(
                        cr, 
                        uid, 
                        component_ids, 
                        context)
                    component_ids = []
        self._delete_component_stockouts(
            cr, 
            uid, 
            component_ids, 
            context)
        
        return change_count

    def _count_expected(self, cr, uid, clock):
        move_obj = self.pool.get('stock.move')
        incoming_move_ids = move_obj.search(
                cr,
                uid,
                [('state', 'in', ['waiting', 'confirmed', 'assigned']),
                 '|',
                 ('purchase_line_id', '!=', None),
                 ('production_id', '!=', None)])
        today = date.today()
        for incoming_move in move_obj.browse(cr, uid, incoming_move_ids):
            date_planned = datetime.strptime(
                incoming_move.date, 
                '%Y-%m-%d %H:%M:%S').date()
            ticks = max(1, (date_planned - today).days)
            clock.expect(
                incoming_move.product_id.id,
                incoming_move.product_qty,
                ticks)
    
    def _count_sales(self, cr, uid, clock):
        move_obj = self.pool.get('stock.move')
        outgoing_move_ids = move_obj.search(
                cr,
                uid,
                [('state', 'in', ['waiting', 'confirmed', 'assigned']),
                 ('sale_line_id', '!=', None)])
        today = date.today()
        for outgoing_move in move_obj.browse(cr, uid, outgoing_move_ids):
            date_planned = datetime.strptime(
                outgoing_move.date, 
                '%Y-%m-%d %H:%M:%S').date()
            ticks = max(1, (date_planned - today).days)
            clock.expect(
                outgoing_move.product_id.id,
                -outgoing_move.product_qty,
                ticks)
    
    def _update_stockouts(
        self,
        cr,
        uid, 
        product, 
        clock,
        component_ids,
        context):
        """ Update any product_stockout records that need changes.
        
        If product can be sold, then calculate the stockout levels for all
        its components and update any that have changed.
        Otherwise, add product.id to component_ids.
        """
        if not product.sale_ok:
            component_ids.append(product.id)
            return
        
        # Calculate business days without fluctuation throughout the week.
        ticks = int(product.sale_delay)
        ticks = (ticks / 5) * 7 + ticks % 5
        stockout_snapshots = clock.check_lead_time(
            product.id, 
            product.lead_time_qty, 
            ticks)
        stockouts = {}
        stockout_obj = self.pool.get('product.stockout')
        stockout_ids = stockout_obj.search(
            cr,
            uid,
            [('product_id', '=', product.id)],
            context=context)
        for stockout in stockout_obj.browse(cr, uid, stockout_ids):
            stored = stockouts.setdefault(stockout.component_id.id, stockout)
            if stored != stockout:
                # must have found a duplicate. Delete it.
                stockout_obj.unlink(cr, uid, stockout.id)
        today = datetime.today()
        for component_id, stockout_snapshot in stockout_snapshots.iteritems():
            stockout = stockouts.pop(component_id, None)
            data = stockout_snapshot.get_data()
            if not stockout:
                data.update(
                    product_id=product.id,
                    component_id=component_id,
                    start_date=today)
                stockout_obj.create(
                    cr,
                    uid,
                    data,
                    context)
            elif stockout_snapshot.has_changed(stockout):
                stockout_obj.write(
                    cr,
                    uid,
                    stockout.id,
                    data,
                    context)
        outdated_stockout_ids = [so.id for so in stockouts.values()]
        if outdated_stockout_ids:
            stockout_obj.unlink(cr, uid, outdated_stockout_ids, context)
    
    def _delete_component_stockouts(self, cr, uid, component_ids, context):
        """ Delete any records from the product_stockout table where the
        product_id is in component_ids.
        """
        stockout_obj = self.pool.get('product.stockout')
        stockout_ids = stockout_obj.search(
            cr,
            uid,
            [('product_id', 'in', component_ids)],
            context=context)
        if stockout_ids:
            stockout_obj.unlink(cr, uid, stockout_ids, context)
    
    def _reset_product_sales(
        self, 
        cr, 
        uid, 
        start_date, 
        end_date):
        """ Load existing product_sales records for a date range
        
        The records will be loaded for all months in the range
        [start_date, end_date), plus enough earlier months to cover the maximum
        value of product_category.daily_sales_months and at least 12 months
        to calculate yearly volumes.
        product_sales has the structure 
        {product_id: {month: (id, qty, consumed)}} where
        id is the existing product_sales record and qty and consumed 
        are 0.0 for any months on or after start_date, and the current value for 
        any earlier months. 
        """
        product_sales = {}
        product_sales_obj = self.pool.get('product.sales')
        category_obj = self.pool.get('product.category')
        
        max_months = 12 # need to calculate yearly volumes
        category_ids = category_obj.search(cr, uid, [])
        for category in category_obj.browse(cr, uid, category_ids):
            max_months = max(max_months, category.daily_sales_months)
        history_start = date_math.add_months(end_date, -max_months)
        
        month = min(start_date, history_start)
        while (month < end_date):
            ids = product_sales_obj.search(cr, uid, [('date', '=', month)])
            for row in product_sales_obj.browse(cr, uid, ids):
                months = product_sales.setdefault(row.product_id.id, {})
                if month >= start_date:
                    months[month] = (row.id, 0.0, 0.0)
                else:
                    months[month] = (row.id, row.qty, row.consumed)
            month = date_math.add_months(month, 1)
        return product_sales
    
    def _reset_volumes(
        self, 
        cr, 
        uid):
        """ Load existing product records and starting values.
        
        Returns a dictionary 
        {product_id : [
            is_estimate, 
            history_months,
            daily_sales, 
            yearly_sales, 
            daily_consumption, 
            yearly_consumption]} 
        Dictionary values are None for any products with type != 'product' 
        """
        
        volumes = {}
        ids = self.search(cr, uid, [])
        for product in self.browse(cr, uid, ids):
            if product.product_tmpl_id.type != 'product':
                volumes[product.id] = None
            else:
                is_estimate = product.daily_sales_mode == 'estimated'
                daily_sales = product.daily_sales if is_estimate else 0.0
                volumes[product.id] = [
                    is_estimate,
                    product.product_tmpl_id.categ_id.daily_sales_months,
                    daily_sales,
                    0.0,
                    0.0,
                    0.0]
        return volumes
    
    def _write_product_sales(
        self, 
        cr, 
        uid,
        last_year, 
        start_date,
        end_date,
        product_sales, 
        context):
        """ Write product sales records to the database
        
        product_sales has the structure {product_id: {month: (id, qty)}} where
        id is the existing product_sales record or None for new records.
        """
        product_sales_obj = self.pool.get('product.sales')
        
        blank_entry = (None, 0, 0)
        months_to_write = []
        month = start_date
        while (month < end_date):
            months_to_write.append(month)
            month = date_math.add_months(month, 1)
        
        for product_id, months in product_sales.items():
            for month in months_to_write:
                id, qty, consumed = months.get(month, blank_entry)
                days = (date_math.add_months(month, 1) - month).days
                daily_sales = round(qty/days, 2)
                daily_consumed = round(consumed/days, 2)
                if id == None:
                    product_sales_obj.create(
                        cr,
                        uid,
                        {
                            'product_id': product_id, 
                            'date': month, 
                            'qty': qty,
                            'daily_qty': daily_sales,
                            'consumed': consumed,
                            'daily_consumed': daily_consumed},
                        context)
                else:
                    product_sales_obj.write(
                        cr, 
                        uid, 
                        id,
                        {
                            'qty': qty, 
                            'daily_qty': daily_sales,
                            'consumed': consumed,
                            'daily_consumed': daily_consumed},
                        context)
        
        # Now remove any old zero records.
        finished = False
        while (not finished):
            ids = product_sales_obj.search(
                cr,
                uid,
                [
                    ('qty', '=', 0),
                    ('consumed', '=', 0),
                    ('date', '<', last_year)],
                limit=100)
            if not ids:
                finished = True
            else:
                product_sales_obj.unlink(
                    cr,
                    uid,
                    ids)
    
    def _write_volumes(
        self, 
        cr, 
        uid,
        product_volumes, 
        context):
        """ Write product sales records to the database
        
        product_volumes has the structure: 
        {product_id : [
            is_estimate, 
            history_months,
            daily_sales, 
            yearly_sales, 
            daily_consumption, 
            yearly_consumption]} 
        """
        
        for product_id, fields in product_volumes.items():
            if fields == None:
                continue # Service or consumable
            self.write(
                cr,
                uid,
                product_id,
                {
                    'daily_sales': fields[2],
                    'yearly_sales': fields[3],
                    'daily_consumption': fields[4],
                    'yearly_consumption': fields[5]})
    
    def _calculate_volumes(
        self, 
        cr, 
        uid,
        today,
        product_sales,
        product_volumes):
        """ Calculate sales and consumption volumes from sales records.
        
        product_sales has the structure 
        {product_id: {month: (id, qty, consumed)}}.
        product_volumes has the structure
        {product_id : [
            is_estimate, 
            history_months,
            daily_sales, 
            yearly_sales, 
            daily_consumption, 
            yearly_consumption]} 
        Dictionary values are None for products with type != 'product'.
        """
        product_ids = set(product_volumes.keys())
        snapshots = dict(
            (id, Snapshot(id)) for id in product_ids)
        self._get_product_components(cr, uid, snapshots)
        
        for product_id, months in product_sales.items():
            fields = product_volumes.get(product_id)
            if fields == None:
                continue #Service or consumable
            is_estimate = fields[0]
            history_months = fields[1]
            product_ids.remove(product_id)
            s = sales_set(today, history_months=history_months)
            c = sales_set(today, history_months=history_months)
            for month, row in months.items():
                id, qty, consumed = row
                s.add_sales(month, qty)
                c.add_sales(month, consumed)
            if not is_estimate:
                fields[2] = s.forecast_sales()
            fields[3] = s.yearly_sales
            fields[4] = c.forecast_sales()
            fields[5] = c.yearly_sales
        
    def _calculate_all_sales_volumes(
        self, 
        cr,
        uid, 
        sales_months, 
        context=None):
        """ Calculate daily sales volumes for all products.
        
        Walk through sales orders from the last sales_months months and calculate
        volumes for products and components. Update or create the product_sales
        records for each product. Also do the same for consumption.
        """
        
        today = date.today()
        start_date = date_math.add_months(today, -sales_months)
        end_date = date_math.add_months(today, 0)
        last_year = date_math.add_months(today, -12)

        product_sales = self._reset_product_sales(
            cr, 
            uid, 
            start_date, 
            end_date)
        product_volumes = self._reset_volumes(cr, uid) 
        
        sale_obj = self.pool.get('sale.order')
        sale_ids = sale_obj.search(
            cr,
            uid,
            [
                ('state', '!=', 'draft'), 
                ('state', '!=', 'cancel'), 
                ('date_order', '>=', str(start_date)),
                ('date_order', '<', str(end_date))])
        for sale in sale_obj.browse(cr, uid, sale_ids):
            month = datetime.strptime(sale.date_order, '%Y-%m-%d').date()
            month = month.replace(day=1)
            for line in sale.order_line:
                if line.product_id.id not in product_volumes:
                    continue #must be a service or consumable
                if line.product_uom_qty == 0.0:
                    continue #probably got back ordered in Simply Accounting
                months = product_sales.setdefault(line.product_id.id, {})
                id, qty, consumed = months.get(month, (None, 0.0, 0.0))
                #FIXME: handle uom conversion
                qty += line.product_uom_qty
                months[month] = (id, qty, consumed)
        
        location_ids = self.pool.get('stock.location').search(
            cr,
            uid,
            [('usage', '=', 'production')])
        move_obj = self.pool.get('stock.move')
        move_ids = move_obj.search(
            cr,
            uid,
            [
                ('state', '=', 'done'), 
                ('date', '>=', str(start_date)),
                ('date', '<', str(end_date)),
                ('location_dest_id', 'in', location_ids)])
        for move in move_obj.browse(cr, uid, move_ids):
            month = datetime.strptime(move.date, '%Y-%m-%d %H:%M:%S').date()
            month = month.replace(day=1)
            if move.product_id.id not in product_volumes:
                continue #must be a service or consumable
            if move.product_qty == 0.0:
                continue #some kind of dummy move
            months = product_sales.setdefault(move.product_id.id, {})
            id, qty, consumed = months.get(month, (None, 0.0, 0.0))
            #FIXME: handle uom conversion
            consumed += move.product_qty
            months[month] = (id, qty, consumed)

        self._calculate_volumes(
            cr,
            uid,
            today,
            product_sales, 
            product_volumes)
        self._write_product_sales(
            cr, 
            uid, 
            last_year,
            start_date,
            end_date,
            product_sales, 
            context)
        self._write_volumes(
            cr, 
            uid, 
            product_volumes, 
            context)
    
    def _incoming_qty(self, cr, uid, ids, field_name, arg, context):
        """ Calculate the total of purchase_qty, production_qty, and
        procurement_qty.
        """
        
        result = {}
        for product in self.browse(cr, uid, ids, context):
            result[product.id] = (product.purchase_qty +
                                  product.production_qty +
                                  product.procurement_qty)
        return result

    _columns = {
        'build_capacity' : fields.float(
            string='Build capacity',
            readonly=True,
            help="Amount that could be built with current real stock, " + 
                "assuming existing production orders are filled first"),
        'sales_capacity' : fields.float(
            string='Sales capacity',
            readonly=True,
            help="Amount that could be sold using current real stock, " +
                "assuming existing sales and production orders are filled " +
                "first. Computed as: Virtual Stock - " +
                "Purchase Orders + Build Capacity"),
        'committed' : fields.float(
            string='Committed stock',
            readonly=True,
            help='Amount that has been built into other assemblies or finished ' +
                'products. Computed as sum of Virtual Stock + Committed Stock ' +
                'for all consumers.'),
        'procurement_qty' : fields.float(
            string='Procurements',
            readonly=True,
            help='Amount to be received from procurements that are waiting ' +
            'for purchase order approval. The purchase order amount may be ' +
            'different from the original procurement amount.'),
        'purchase_qty' : fields.float(
            string='Purchase orders',
            readonly=True,
            help='Amount to be received from approved purchase orders'),
        'production_qty' : fields.float(
            string='Slow production orders',
            readonly=True,
            help='Amount to be produced from production orders taking more than five days'),
        'incoming_qty' : fields.function(
            _incoming_qty,
            method=True,
            type='float',
            readonly=True,
            string='Total Incoming',
            help='Amount to be received from procurements, purchase orders, ' +
                'and production orders, combined.'),
        'daily_sales' : fields.float(
            string='Daily Sales',
            digits=(12,3),
            help='Forecast quantity of a product that will be sold ' +
                'each day. Estimated by user or calculated as the average ' +
                'over the last six months. Used for stockout simulation.'),
        'daily_consumption' : fields.float(
            string='Daily Consumption',
            digits=(12,3),
            help='Average consumption over the last six months. For ' +
                'reference only, not used in stockout simulation.',
            readonly=True),
        'yearly_sales' : fields.float(
            string='Yearly Sales',
            help='Total number sold in the last 12 months.',
            readonly=True),
        'yearly_consumption' : fields.float(
            string='Yearly Consumption',
            help='Total number consumed by production in the last 12 months.',
            readonly=True),
        'daily_sales_mode' : fields.selection(
            (('estimated', 'User Estimated'), ('calculated', 'Calculated')),
            'Daily Sales Mode',
            help='Whether to calculate based on sales history or a user estimate.'),
        'days_until_stockout' : fields.float(
            string='Days Until Stockout',
            readonly=True,
            help='Days until all current real stock was consumed in the ' + 
                'production simulation.'),
        'safety_days': fields.float(
            string='Days of Safety Stock',
            help='A buffer to be added to the vendor lead time before reordering'),
        'days_until_reorder' : fields.float(
            string='Days Until Reorder',
            readonly=True,
            help='If no order has been placed or the order will arrive too ' +
                'late, then this is Days Until Stockout minus supplier delay ' +
                'minus safety days. If an order has been placed in time, ' +
                'then this is the days until the next order is needed.'),
        'sales_history' : fields.one2many(
            'product.sales',
            'product_id',
            'Usage History',
            readonly=True),
    }
    
    _defaults = {
        'daily_sales_mode': lambda *a: 'calculated'
    }
    
    def _calc_seller(self, cr, uid, ids, fields, arg, context=None):
        """ We don't like the default delay of one day when no supplier
        is configured. I think it messed up products that we produce ourselves.
        """
        result = super(product_product, self)._calc_seller(cr,
                                                           uid,
                                                           ids,
                                                           fields,
                                                           arg,
                                                           context)
        for row in result.itervalues():
            if not row['seller_info_id']:
                row['seller_delay'] = 0
        return result

    def _count_procurements(self, cr, uid, snapshot):
        quantity = 0.0
        procurement_obj = self.pool.get('procurement.order')
        move_obj = self.pool.get('stock.move')
        proc_ids = procurement_obj.search(
            cr, 
            uid, 
            [
                ('product_id', '=', snapshot.product_id), 
                ('procure_method', '=', 'make_to_order'),
                ('state', 'in', ['confirmed', 'exception', 'running', 'ready', 'waiting'])])
        
        # This will map move_id -> qty for the procurement moves
        procurement_moves = {}
        if proc_ids:
            for proc in procurement_obj.browse(cr, uid, proc_ids):
                procurement_move = proc.move_id
                if procurement_move:
                    procurement_moves[procurement_move.id] = (
                        procurement_move.product_qty)

        incoming_move_ids = move_obj.search(
                cr,
                uid,
                [('product_id', '=', snapshot.product_id),
                 ('state', 'in', ['waiting', 'confirmed', 'assigned']),
                 '|',
                 ('purchase_line_id', '!=', None),
                 ('production_id', '!=', None)])
        incoming_move_id_set = set(incoming_move_ids)
        purchases = {}
        productions = {}
        if incoming_move_ids:
            for incoming_move in move_obj.browse(cr, uid, incoming_move_ids):
                if incoming_move.purchase_line_id:
                    purchases[incoming_move.id] = incoming_move.product_qty
                elif snapshot.is_slow_to_produce:
                    productions[incoming_move.id] = incoming_move.product_qty
                if incoming_move.move_dest_id:
                    # don't count the procurement anymore
                    procurement_moves[incoming_move.move_dest_id.id] = 0
                    # don't count the shipment twice.
                    incoming_move_id_set.discard(incoming_move.move_dest_id.id)
        snapshot.purchase_qty = sum(
            [purchases.get(i, 0) for i in incoming_move_id_set])
        snapshot.production_qty = sum(
            [productions.get(i, 0) for i in incoming_move_id_set])
        
        # count any procurements that haven't been replaced by purchases   
        for qty in procurement_moves.itervalues():
            snapshot.procurement_qty += qty
    def copy(self, cr, uid, id, default=None, context=None):
        if default is None:
            default = {}
        default = default.copy()
        default.update({'variant_ids': False, 'sales_history': False})
        return super(product_product, self).copy(cr, uid, id, default, context)

product_product()

class mrp_bom(osv.osv):
    _inherit = "mrp.bom"
    
    def _get_product_capacities(self, cr, uid, ids, field_names, arg, context):
        product_boms = {} # {product_id: [bom_id]}
        for bom in self.pool.get('mrp.bom').browse(cr, uid, ids):
            product_id = bom.product_id.id
            boms = product_boms.setdefault(product_id, [])
            boms.append(bom.id)
        
        #strip 'product_' off the start of the field names.
        product_field_names = map(lambda name: name[8:], field_names)
        product_field_names.append('id')
        
        product_ids = product_boms.keys()
        product_table = self.pool.get('product.product').read(
            cr, 
            uid, 
            product_ids, 
            fields=product_field_names)
        
        capacities = {}
        for product_row in product_table:
            bom_ids = product_boms[product_row['id']]
            for bom_id in bom_ids:
                fields = {}
                if 'product_sales_capacity' in field_names:
                    fields['product_sales_capacity'] = product_row['sales_capacity']
                if 'product_build_capacity' in field_names:
                    fields['product_build_capacity'] = product_row['build_capacity']
                capacities[bom_id] = fields
        return capacities

    _columns = {
        'product_sales_capacity' : fields.function(
            _get_product_capacities, 
            type="float",
            method=True,
            string="Sales capacity",
            multi='capacities'),                                           
        'product_build_capacity' : fields.function(
            _get_product_capacities, 
            type="float",
            method=True,
            string="Build capacity",
            multi='capacities'),
        'product_revision' : fields.related(
            'product_id',
            'revision',
            type='integer',
            string='Product Revision',
            readonly=True)
    }
mrp_bom()

class stock_move(osv.osv):
    _inherit = "stock.move"
    
    _columns = {
        'product_revision' : fields.related(
            'product_id',
            'revision',
            type='integer',
            string='Product Revision',
            readonly=True)
    }
stock_move()

class mrp_production(osv.osv):
    _inherit = "mrp.production"
    
    _columns = {
        'product_revision' : fields.related(
            'product_id',
            'revision',
            type='integer',
            string='Product Revision',
            readonly=True)
    }
mrp_production()


class product_sales(osv.osv):
    _name = "product.sales"
    _description = "Product Sales History"
    _order = "date desc"
    _rec_name = "product_id"
    _columns = {
        'product_id' : fields.many2one(
            'product.product', 
            'Product', 
            ondelete='cascade',
            required=True),
        'date' : fields.date('Month'),
        'qty' : fields.float(
            'Sold Total',
            help='The number of items sold during the month'),
        'daily_qty' : fields.float(
            'Sold Daily',
            help='The monthly sales divided by the number of days in the month'),
        'consumed' : fields.float(
            'Consumed Total',
            help='The number of items consumed in production during the month'),
        'daily_consumed' : fields.float(
            'Consumed Daily',
            help='The monthly consumption divided by the number of days in the month'),
    }
product_sales()

class product_supplierinfo(osv.osv):
    _inherit = 'product.supplierinfo'
    
    def _get_days_until_reorder(
        self, 
        cr, 
        uid, 
        ids, 
        field_name, 
        arg, 
        context):
        results = {}
        for supplierinfo in self.browse(cr, uid, ids):
            # Unfortunately, the supplier info points to a product template,
            # not a product, so we have to look through all the products
            # under this template. This isn't a problem for us, because we're
            # only using one product under each template.
            min_days = 10000
            for product in supplierinfo.product_id.variant_ids:
                min_days = min(product.days_until_reorder, min_days)
            results[supplierinfo.id] = min_days
        
        return results
    
    _columns = {
        'order_cost' : fields.float(
            'Order Cost',
            help='The cost of placing an order with the supplier (shipping, ' +
                'administration, etc.)'),
        'days_until_reorder' : fields.function(
            _get_days_until_reorder,
            type="float",
            method=True,
            string='Days Until Reorder', 
            help='Virtual stock plus build capacity divided by daily sales ' +
                'and consumption minus delivery delay from first supplier.')
    }
product_supplierinfo()

class pricelist_partnerinfo(osv.osv):
    _inherit = 'pricelist.partnerinfo'
    
    def _get_npc(self, cr, uid, ids, field_names, arg, context):
        results = {}
        company = self.pool.get('res.users').browse(cr, uid, uid).company_id
        holding_cost = company.holding_cost
        rate_of_return = company.rate_of_return
        
        for entry in self.browse(cr, uid, ids, context):
            supplierinfo = entry.suppinfo_id
            product_template = supplierinfo.product_id
            order_cost = supplierinfo.order_cost
            volume = product_template.volume
            daily_volume = self._get_daily_volume(product_template)
            if daily_volume == 0.0:
                days_to_consume = 20000.0
                npc = 0.0
            else:
                min_quantity = max(entry.min_quantity, 1.0)
                days_to_consume = min_quantity / daily_volume
                years_to_consume = days_to_consume / 365.25
                # factor for calculating the NPC of a recurring cost
                recurrence_factor = (
                    (1 + rate_of_return * years_to_consume/0.7) /
                    (rate_of_return * years_to_consume / 0.7))
                lifetime_cost = (
                    (entry.price * min_quantity + order_cost) * recurrence_factor +
                    (0.5 * years_to_consume * min_quantity * holding_cost * 
                        product_template.volume * recurrence_factor) /
                        (1 + rate_of_return) ** (years_to_consume / 4))
                npc = lifetime_cost / 365.25 / daily_volume / 0.7 * rate_of_return
            
            results[entry.id] = {
                'npc' : npc,
                'days_to_consume' : days_to_consume
            }
        return results
    
    def _get_daily_volume(self, product_template):
        volume = 0.0
        for product in product_template.variant_ids:
            volume += product.daily_consumption + product.daily_sales
        return volume
    
    _columns = {
        'npc': fields.function(
            _get_npc,
            type="float",
            method=True,
            multi='npc',
            string='Net Present Cost', 
            help='Net present cost per unit for a recurring purchase of this size'),
        'days_to_consume': fields.function(
            _get_npc,
            type='float',
            method=True,
            multi='npc',
            string='Days to Consume This',
            help='Estimate of how long an order of this size will take to ' +
                'be consumed.')
    }
pricelist_partnerinfo()

class res_company(osv.osv):
    _inherit = 'res.company'
    
    _columns = {
        'holding_cost' : fields.float(
            'Holding Cost',
            help='The cost per cubic meter per year to store products in ' +
                'inventory.'),
        'rate_of_return' : fields.float(
            'Rate of Return',
            help='Internal rate of return per year used to calculate net ' +
                'present value.')
    }
    
    _defaults = {
        'rate_of_return': lambda *a: 0.35
    }
res_company()

class product_stockout(osv.osv):
    _name = "product.stockout"
    _description = "Product Stockout"
    _order = "product_id, component_id"
    
    def _component_fields(self, cr, uid, ids, field_names, arg, context):
        """ Retrieve a set of fields from the related component product.
        
        field_names must all start with 'component_' and the rest of the name
        must match a field on the product table.
        """
        
        regular_field_names = [field_name
                               for field_name in field_names
                               if field_name != 'component_incoming_qty']
        result = {}
        prefix_length = len('component_')
        for stockout in self.browse(cr, uid, ids, context):
            component = stockout.component_id
            row = {}
            for field_name in regular_field_names:
                row[field_name] = getattr(component, field_name[prefix_length:])
            if 'component_incoming_qty' in field_names:
                row['component_incoming_qty'] = (component.purchase_qty +
                                                 component.production_qty +
                                                 component.procurement_qty)
            result[stockout.id] = row
        
        return result
    
    _columns = {
        'product_id' : fields.many2one(
            'product.product', 
            'Product', 
            ondelete='cascade',
            readonly=True,
            required=True),
        'component_id' : fields.many2one(
            'product.product', 
            'Component', 
            ondelete='cascade',
            readonly=True,
            required=True),
        'consumption' : fields.float(
            string='Components per Product',
            readonly=True,
            help='Amount of component required to build one product.'),
        'component_procurement_days' : fields.float(
            string='Procurement Days of Component',
            readonly=True,
            help='Days to produce or purchase the component.'),
        'component_to_product_days' : fields.float(
            string='Component to Product Days',
            readonly=True,
            help='Days to assemble this component into the finished product.'),
        'start_date' : fields.date(
            'Since',
            help='When this last changed.',
            readonly=True),
        'note' : fields.text('Note'),
        'is_kit' : fields.related(
            'product_id',
            'is_kit',
            type='boolean',
            readonly=True,
            string='Is Kit'),
        'product_revision' : fields.related(
            'product_id',
            'revision',
            type='integer',
            readonly=True,
            string='Product Revision'),
        'product_sales_capacity' : fields.related(
            'product_id',
            'sales_capacity',
            type='float',
            readonly=True,
            string='Sales Capacity of Product'),
        'product_sale_delay' : fields.related(
            'product_id',
            'sale_delay',
            type='float',
            readonly=True,
            string='Lead Time for Product Customers'),
        'product_lead_time_qty' : fields.related(
            'product_id',
            'lead_time_qty',
            type='float',
            readonly=True,
            string='Quantity for Customer Lead Time'),
        'product_state' : fields.related(
            'product_id',
            'state',
            type='selection',
            selection=[
                ('',''),
                ('draft', 'In Development'),
                ('sellable','In Production'),
                ('end','End of Lifecycle'),
                ('obsolete','Obsolete')],
            readonly=True,
            string='Product State'),
        'component_state' : fields.related(
            'component_id',
            'state',
            type='selection',
            selection=[
                ('',''),
                ('draft', 'In Development'),
                ('sellable','In Production'),
                ('end','End of Lifecycle'),
                ('obsolete','Obsolete')],
            readonly=True,
            string='Component State'),
        'component_revision' : fields.related(
            'component_id',
            'revision',
            type='integer',
            readonly=True,
            string='Component Revision'),
        'component_sales_capacity' : fields.related(
            'component_id',
            'sales_capacity',
            type='float',
            readonly=True,
            string='Sales Capacity of Component'),
        'component_days_until_stockout' : fields.related(
            'component_id',
            'days_until_stockout',
            type='float',
            string='Days Until Stockout',
            readonly=True,
            help='estimated days until the component stocks out'),
        'component_days_until_reorder' : fields.related(
            'component_id',
            'days_until_reorder',
            type='float',
            string='Days Until Reorder',
            readonly=True,
            help='estimated days until the component should be reordered'),
        'component_qty_available' : fields.related(
            'component_id',
            'qty_available',
            type='float',
            readonly=True,
            string='Component Real Stock'),
        'component_virtual_available' : fields.related(
            'component_id',
            'virtual_available',
            type='float',
            readonly=True,
            string='Component Virtual Stock'),
        'component_procurement_qty' : fields.function(
            _component_fields,
            method=True,
            multi='component_fields',
            type='float',
            readonly=True,
            string='Component Procurements',
            help='Amount to be received from new procurements'),
        'component_purchase_qty' : fields.function(
            _component_fields,
            method=True,
            multi='component_fields',
            type='float',
            readonly=True,
            string='Component Purchases',
            help='Amount to be received from approved purchase orders'),
        'component_production_qty' : fields.function(
            _component_fields,
            method=True,
            multi='component_fields',
            type='float',
            readonly=True,
            string='Component Production',
            help='Amount to be produced from production orders taking more ' +
                'than five days'),
        'component_incoming_qty' : fields.function(
            _component_fields,
            method=True,
            multi='component_fields',
            type='float',
            readonly=True,
            string='Components Incoming',
            help='Amount to be received from procurements, purchase orders, ' +
                'and production orders, combined.')
    }
product_stockout()

class product_category(osv.osv):
    _inherit = 'product.category'
    
    _columns = {
        'daily_sales_months': fields.integer(
            string='Daily Sales Months',
            help='Determines how many months to use when calculating daily ' +
                'sales and consumption.')
    }
    
    _defaults = {
        'daily_sales_months': lambda *a: 6
    }
product_category()
