import datetime
import date_math

class sales_set:
    """ A set of sales data used to calculate a predicted sales volume for a
    single product.
    
    To create a sales_set, just pass it today's date. It will calculate the
    cutoff date for the previous twelve months and start calculating average 
    volumes from the last six months.
    >>> s = sales_set(datetime.date(2010, 3, 15))
    >>> s.year_start
    datetime.date(2009, 3, 1)
    >>> s.history_start
    datetime.date(2009, 9, 1)
    
    You can also specify when to start calculating average volume. 
    >>> s = sales_set(datetime.date(2010, 3, 15), history_months=12)
    >>> s.year_start
    datetime.date(2009, 3, 1)
    >>> s.history_start
    datetime.date(2009, 3, 1)
    
    And finally, it will ignore any sales after the start of the current month.
    >>> s.end_date
    datetime.date(2010, 3, 1)
    """
    def __init__(self, today, history_months=6):
       self.year_start = date_math.add_months(today, -12)
       self.history_start = date_math.add_months(today, -history_months)
       self.history_months = history_months
       self.end_date = date_math.add_months(today, 0)
       self.yearly_sales = 0
       self.__monthly_sales = {} 
    
    def add_sales(self, month, qty):
        """ Add sales data for a given month.
        
        Any sales within the previous twelve months will contribute to the
        yearly_sales quantity.
        >>> s = sales_set(datetime.date(2010, 3, 15))
        >>> s.add_sales(datetime.date(2010, 1, 1), 10)
        >>> s.yearly_sales
        10
        
        Sales outside the previous twelve months are ignored.
        >>> s.add_sales(datetime.date(2009, 1, 1), 2)
        >>> s.yearly_sales
        10
        >>> s.add_sales(datetime.date(2010, 4, 1), 3)
        >>> s.yearly_sales
        10
        
        The exact start date of the previous twelve months will be included in
        yearly sales.
        >>> s.add_sales(datetime.date(2009, 3, 1), 2)
        >>> s.yearly_sales
        12
        
        The current month will be excluded from yearly sales.
        >>> s.add_sales(datetime.date(2010, 3, 1), 3)
        >>> s.yearly_sales
        12
        """
        if self.year_start <= month and month < self.end_date:
            self.yearly_sales += qty
        if self.history_start <= month and month < self.end_date:
            monthly_sales = self.__monthly_sales.get(month, 0.0)
            self.__monthly_sales[month] = monthly_sales + qty
    
    def forecast_sales(self):
        """ Forecast what daily sales volume will be.
        
        Just uses an average of sales since the start of this set's history.
        
        >>> s = sales_set(datetime.date(2010, 1, 1))
        >>> s.add_sales(datetime.date(2009, 6, 1), 184+60)
        >>> s.add_sales(datetime.date(2009, 7, 1), 184)
        
        The sales from last June are more than six months ago, so they will be
        ignored. There are 184 days in the last six months, so the sales average
        one per day.
        
        >>> s.forecast_sales()
        1.0

        If we use seven months of history then the high volume in last June
        will double the sales average per day.
        
        >>> s = sales_set(datetime.date(2010, 1, 1), history_months=7)
        >>> s.add_sales(datetime.date(2009, 6, 1), 184+60)
        >>> s.add_sales(datetime.date(2009, 7, 1), 184)
        
        >>> s.forecast_sales()
        2.0
        
        We can even use more than 12 months of history for calculating average.
        Yearly sales still use 12 months.
        
        >>> s = sales_set(datetime.date(2011, 1, 1), history_months=24)
        >>> s.add_sales(datetime.date(2010, 6, 1), 10)
        >>> s.add_sales(datetime.date(2009, 7, 1), 365*2-10)
        
        >>> s.forecast_sales()
        1.0
        
        >>> s.yearly_sales
        10

        If history_months are zero, just return a forecast of zero.
        
        >>> s = sales_set(datetime.date(2011, 1, 1), history_months=0)
        >>> s.add_sales(datetime.date(2010, 6, 1), 1)
        
        >>> s.forecast_sales()
        0.0
        
        """
        
        if self.history_months == 0:
            return 0.0
        
        months = map(
            lambda m: date_math.add_months(self.end_date, -m-1), 
            range(0, self.history_months))
        month_lengths = map(
            lambda month: (date_math.add_months(month, 1) - month).days,
            months)
        monthly_sales = map(
            lambda month: self.__monthly_sales.get(month, 0.0),
            months)
        return sum(monthly_sales) / sum(month_lengths)

if __name__ == "__main__":
    import doctest
    failcount, testcount = doctest.testmod()
    if failcount == 0:
        print "Success."
