from sim_stock import sim_stock
from sim_forecast import sim_forecast

class sim_clock:
    """
    """
    
    def __init__(self, uom_obj):
        """ Initialize an instance.
        
        uom_obj is the model object that will be used to convert units of measure.
        """
        self.uom_obj = uom_obj
        self.products = {}
        self.start_stock = sim_stock()
        self.prepped_stock = None
        self.verbose = False
        self.monitored_product_id = None
        self.reset()
        
    def reset(self, ignore_safety=False):
        self.stock = sim_stock(self.start_stock).create_next()
        self.maxtick = 0
        for product in self.products.itervalues():
            product.ignore_safety = ignore_safety
            self.maxtick = max(self.maxtick, 2*product.get_total_delay())
        self.maxtick = max(90, self.maxtick)
        self.rewind_count = 0
        self.monitor_log = ''
    
    def add_product(self, product, start_qty=0):
        """ Add a new product to the simulation
        """
        self.products[product.id] = product
        self.start_stock.add_product(product.id, start_qty)
        
    def get_product(self, product_id):
        """ Find a product by id.
        
        >>> product_id = 99
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id)
        >>> clock.add_product(product)
        >>> product2 = clock.get_product(product_id)
        >>> product == product2
        True
        """
        return self.products[product_id]
    
    def expect(self, product_id, qty, tick):
        """ Expect an order to arrive on a given tick.
        
        If an order arrives on the day a product stocks out, then it won't 
        trigger a reorder until the new stock is used up. The stockout date is
        not affected.
        
        This first example shows what happens when no order is expected.
        >>> product_id = 99
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1, delay=4)
        >>> clock.add_product(product, start_qty=5)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        6
        >>> forecast.get_reorder(product_id)
        2
        
        Now, if we expect an order with 10 items, the reorder is pushed back
        by 10 ticks, but the stockout is unchanged.
        >>> clock.expect(product_id, qty=10, tick=6)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        6
        >>> forecast.get_reorder(product_id)
        12

        If the order is going to arrive a day late, then it will still show as 
        needing a reorder.
        >>> product_id = 99
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1, delay=4)
        >>> clock.add_product(product, start_qty=5)
        >>> clock.expect(product_id, qty=10, tick=7)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        6
        >>> forecast.get_reorder(product_id)
        2
        
        A confirmed sale can cause a stockout.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1, delay=10)
        >>> clock.add_product(product, start_qty=5)
        >>> clock.expect(product_id, qty=-2, tick=2)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        4
        
        
        """
        self.start_stock.expect(product_id, qty, tick)
        
    def run(self):
        """ Run the simulation and return the number of days until stockout
        for each product.
        
        Here's a single product stocking out.
        >>> product_id = 99
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.5)
        >>> clock.add_product(product, start_qty=1.0)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        3
        
        If you adjust the product's daily sales, you can rerun the simulation.
        >>> product.daily_sales = 1.0
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        2
        
        If you adjust the product's start quantity, you can rerun the simulation.
        >>> clock.add_product(product, start_qty=2.0)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        3
        
        If you set up a bill of materials, then the component will stock out.
        >>> motor_id = 17
        >>> bom_id = 62
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.5)
        >>> motor = sim_product(motor_id)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        1
        >>> forecast.get_stockout(motor_id)
        3
        
        If the product has more than one component, then they may stock out
        on different days.
        >>> screw_id = 18
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1.0)
        >>> motor = sim_product(motor_id)
        >>> screw = sim_product(screw_id)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=4.0)
        >>> clock.add_product(screw, start_qty=4.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> product.add_component(screw, 2.0)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        1
        >>> forecast.get_stockout(screw_id)
        3
        >>> forecast.get_stockout(motor_id)
        5
        
        By default, the simulation only runs for 90 days.
        >>> clock.add_product(motor, start_qty=100)
        >>> clock.add_product(screw, start_qty=100)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(screw_id)
        51
        >>> forecast.get_stockout(motor_id)
        
        However, it will run for twice the longest delay, if that's longer.
        >>> motor.delay = 75
        >>> forecast = clock.run()
        >>> forecast.get_stockout(screw_id)
        51
        >>> forecast.get_stockout(motor_id)
        101
        
        If we contract out the production of a product, then we need to place
        the order in time to avoid stocking out. In this example, we have
        eight products in stock and sell one per day. That means we stock out
        on day nine. The delay is five days, so we have to place the order
        on day four. That makes both components stock out on day four.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1.0, delay=5.0)
        >>> motor = sim_product(motor_id, delay=2.0)
        >>> screw = sim_product(screw_id)
        >>> clock.add_product(product, start_qty=8.0)
        >>> clock.add_product(motor)
        >>> clock.add_product(screw)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> product.add_component(screw, 2.0)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(motor_id)
        4
        >>> forecast.get_reorder(motor_id)
        2
        >>> forecast.get_stockout(product_id)
        9
        >>> forecast.get_reorder(product_id)
        4
        
        Sometimes, the backtracking can cause a component to stockout earlier
        than it did on a previous pass. In this example, the screw is used in
        assembling the motor, and also in the finished product. We stockout of
        screws on day three by building finished products, but then on day five
        we stock out of motors and realize we should have ordered motors on day
        two. That moves the stockout of screws back to day two.
        >>> motor_bom_id = 37
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=1.0)
        >>> motor = sim_product(motor_id, delay=3.0)
        >>> screw = sim_product(screw_id)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=4.0)
        >>> clock.add_product(screw, start_qty=4.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> product.add_component(screw, 2.0)
        >>> motor.set_production(motor_bom_id)
        True
        >>> motor.add_component(screw)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        1
        >>> forecast.get_stockout(motor_id)
        5
        >>> forecast.get_stockout(screw_id)
        2
        
        As a diagnostic tool, we can ask that a specific product be monitored.
        The clock will then record which products' daily consumption triggered
        that product to be consumed and reordered.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.1)
        >>> motor = sim_product(motor_id)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.monitored_product_id = motor_id
        >>> forecast = clock.run()
        >>> print clock.monitor_log
        Starting quantity of 17: 1.0.
        Producing 99 on tick 1 leaves quantity of 17 at 0.0 / 0.0 at tick 1.
        Producing 99 on tick 11 leaves quantity of 17 at -1.0 / -1.0 at tick 11 with reorder at tick 11.
        
        If components have delays, then they will get reordered before the 
        current tick.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.1)
        >>> motor = sim_product(motor_id, delay=3.0)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.monitored_product_id = motor_id
        >>> forecast = clock.run()
        >>> print clock.monitor_log
        Starting quantity of 17: 1.0.
        Producing 99 on tick 1 leaves quantity of 17 at 0.0 / 0.0 at tick 1.
        Producing 99 on tick 11 leaves quantity of 17 at -1.0 / -1.0 at tick 11 with reorder at tick 8.
        
        If products have delays, then the components will get consumed before
        the current tick.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.1, delay=3.0)
        >>> motor = sim_product(motor_id)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.monitored_product_id = motor_id
        >>> forecast = clock.run()
        >>> print clock.monitor_log
        Starting quantity of 17: 1.0.
        Producing 99 on tick 1 leaves quantity of 17 at 0.0 / 0.0 at tick 0.
        Producing 99 on tick 11 leaves quantity of 17 at -1.0 / -1.0 at tick 8 with reorder at tick 8.
        
        If components have replacements, then the replacements will get 
        reordered, not the old component.
        >>> new_motor_id = 314
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.5)
        >>> motor = sim_product(motor_id)
        >>> new_motor = sim_product(new_motor_id)
        >>> motor.replacement = new_motor
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> clock.add_product(new_motor)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> forecast = clock.run()
        >>> forecast.get_stockout(product_id)
        1
        >>> forecast.get_stockout(motor_id)
        >>> forecast.get_stockout(new_motor_id)
        3
        """
        self.reset()
        self.stock.track_unused()
        self.running = True
        tick = 0
        if self.monitored_product_id:
            monitored_ticks = self.get_monitored_ticks()
            monitored_reorder = None
            self.monitor_log = "Starting quantity of %d: %.1f." % (
                self.monitored_product_id,
                monitored_ticks[-1][0])
        while(self.running):
            tick += 1
            for product_id, product in self.products.iteritems():
                product.consume_daily_sales(self.stock, self.uom_obj)
                if self.monitored_product_id:
                    new_monitored_ticks = self.get_monitored_ticks()
                    new_monitored_reorder = self.get_monitored_reorder(tick)
                    consumption_tick = None
                    for i in range(len(monitored_ticks)):
                        if new_monitored_ticks[i] < monitored_ticks[i]:
                            consumption_tick = i
                            break
                    if (consumption_tick == None and 
                        (new_monitored_ticks[-1] != monitored_ticks[-1] or
                         new_monitored_reorder != monitored_reorder)):
                        consumption_tick = tick
                    monitored_ticks = new_monitored_ticks
                    monitored_reorder = new_monitored_reorder
                    if consumption_tick != None:
                        reorder_text = (
                            ' with reorder at tick %d' % monitored_reorder
                            if monitored_reorder != None
                            else '')
                        virtual_quantity = self.stock.get_quantity(
                            self.monitored_product_id)
                        self.monitor_log += (
                            "\nProducing %d on tick %d leaves quantity of " +
                            "%d at %.1f / %.1f at tick %d%s.") % (
                            product_id,
                            tick,
                            self.monitored_product_id,
                            monitored_ticks[consumption_tick][0],
                            monitored_ticks[consumption_tick][1],
                            consumption_tick,
                            reorder_text)

            self.running = tick < self.maxtick
                
            self.stock = self.stock.create_next()
        
        return self.stock.create_forecast()
    
    def get_monitored_ticks(self):
        ticks = []
        step = self.stock
        while step:
            ticks.insert(
                0, 
                (
                    step.get_real_quantity(self.monitored_product_id), 
                    step.get_quantity(self.monitored_product_id))
                )
            step = step.previous
        return ticks
    
    def get_monitored_reorder(self, current_tick):
        tick = current_tick
        if not self.stock.has_order(self.monitored_product_id):
            return None
        step = self.stock
        while step and step.has_order(self.monitored_product_id):
            tick -= 1
            step = step.previous
        return tick + 1 if step else 0
    
    def check_lead_time(self, product_id, quantity, ticks):
        """ Check whether the given quantity of the given product can be 
        produced in the given ticks.
        
        Returns a list of stockout snapshots for the components that will 
        stockout because of this request, or an empty list if the request can be
        fulfilled.

        >>> product_id = 99
        >>> motor_id = 17
        >>> bom_id = 62
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id, delay=20)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.check_lead_time(product_id, 1.0, 5)
        {}
        
        Now if we request more than can be produced:
        >>> clock.check_lead_time(product_id, 2.0, 5)
        {17: stockout_snapshot(17)}
        
        If we increase the lead time to allow reordering:
        >>> clock.check_lead_time(product_id, 2.0, 20)
        {}
        
        Decreasing the lead time again makes it stockout again.
        >>> clock.check_lead_time(product_id, 2.0, 19)
        {17: stockout_snapshot(17)}
        
        We also need to take confirmed sales into account when looking at 
        stockouts. In this example, we're expecting to ship two products on day
        3, and we have components for four. Adding three more to the two we've
        already confirmed will cause a stockout.
        >>> clock = sim_clock(mock_uom_obj())
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=4.0)
        >>> clock.expect(product_id, -2.0, 3)
        >>> clock.check_lead_time(product_id, 3.0, 5)
        {17: stockout_snapshot(17)}
        
        Products that compete for the same components can cause a stockout.
        >>> gantry_id = 88
        >>> gantry_bom_id = 8
        >>> gantry = sim_product(gantry_id)
        >>> gantry.set_production(gantry_bom_id)
        True
        >>> gantry.add_component(motor)
        >>> clock = sim_clock(mock_uom_obj())
        >>> clock.add_product(product)
        >>> clock.add_product(gantry)
        >>> clock.add_product(motor, start_qty=4.0)
        >>> clock.expect(gantry_id, -2.0, 3)
        >>> clock.check_lead_time(product_id, 3.0, 5)
        {17: stockout_snapshot(17)}
        
        However, we don't want to include stocked out components if they are
        unrelated. In this example, the screw will stockout on day 3, but
        it's not used in the product we're asking about.
        >>> screw_id = 18
        >>> screw = sim_product(screw_id, delay=20)
        >>> clock.add_product(screw)
        >>> clock.expect(screw_id, -1.0, 3)
        >>> clock.check_lead_time(product_id, 5.0, 5)
        {17: stockout_snapshot(17)}
        
        We don't use safety_days when checking lead time.
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id, delay=2.0, safety_days=20)
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.check_lead_time(product_id, 2.0, 5)
        {}
        
        If we buy and sell a product, it should show up as its own component
        when it's likely to stockout.
        >>> clock = sim_clock(mock_uom_obj())
        >>> screw = sim_product(screw_id, delay=20)
        >>> clock.add_product(screw, start_qty=2.0)
        >>> clock.check_lead_time(screw_id, 5.0, 5)
        {18: stockout_snapshot(18)}
        
        If components have replacements, then the replacements can be used.
        >>> new_motor_id = 314
        >>> clock = sim_clock(mock_uom_obj())
        >>> product = sim_product(product_id, daily_sales=0.5)
        >>> motor = sim_product(motor_id, delay=20)
        >>> new_motor = sim_product(new_motor_id, delay=20)
        >>> motor.replacement = new_motor
        >>> clock.add_product(product)
        >>> clock.add_product(motor, start_qty=1.0)
        >>> clock.add_product(new_motor, start_qty=10.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> clock.check_lead_time(product_id, 5.0, 5)
        {}
        >>> clock.check_lead_time(product_id, 20.0, 5)
        {314: stockout_snapshot(314)}
        """
        
        product = self.products[product_id]
        self.reset(ignore_safety=True)
        if self.prepped_stock:
            self.stock = self.prepped_stock.clone()
        while self.stock.tick <= ticks:
            self.stock = self.stock.create_next()
            # trigger any production orders needed to meet expected sales.
            for other_product_id, other_product in self.products.iteritems():
                other_product.consume(self.stock, 0.0, self.uom_obj)
        if not self.prepped_stock or self.prepped_stock.tick < self.stock.tick:
            self.prepped_stock = self.stock.clone()
        while self.stock.tick > ticks + 1:
            self.stock = self.stock.previous
        product.consume(self.stock, quantity, self.uom_obj)
        
        snapshots = product.get_all_snapshots(self.uom_obj)
        reordered_products = set(
            self.stock.create_forecast().get_reorders_by_tick(0))
        return dict([
            (component_id, snapshot)
            for component_id, snapshot in snapshots.iteritems()
            if component_id in reordered_products])

if __name__ == "__main__":
    import doctest
    from mock_uom import mock_uom_obj
    from sim_product import sim_product
    failures, tests = doctest.testmod()
    if failures == 0: print "Success."
