class sim_forecast:
    """ Holds the complete results from a simulation run.
    """
    def __init__(self):
        self.stockouts = {}
        self.reorder = {}
        self.unused = set()
        
    def set_stockout(self, product_id, stockout):
        self.stockouts[product_id] = stockout
    
    def get_stockout(self, product_id):
        """ Get the day on which a product will stock out.
        
        >>> product_id = 99
        >>> forecast = sim_forecast()
        >>> forecast.set_stockout(product_id, 7)
        >>> forecast.get_stockout(product_id)
        7
                
        >>> dummy_id = 98
        >>> forecast.get_stockout(dummy_id)

        """
        return self.stockouts.get(product_id)
    
    def get_reorders_by_tick(self, tick):
        """ Get a list of all product ids that are forecast to stockout on or
        before the requested tick.
        
        >>> product_id = 99
        >>> forecast = sim_forecast()
        >>> forecast.get_reorders_by_tick(7)
        []
        >>> forecast.set_reorder(product_id, 7)
        >>> forecast.get_reorders_by_tick(7)
        [99]
        >>> forecast.get_reorders_by_tick(6)
        []
        
        """
        return [product_id 
                for product_id, reorder in self.reorder.iteritems() 
                if reorder <= tick]

    def set_reorder(self, product_id, reorder):
        self.reorder[product_id] = reorder
    
    def get_reorder(self, product_id):
        """ Get the day on which a product should be reordered.

        >>> product_id = 99
        >>> forecast = sim_forecast()
        >>> forecast.set_reorder(product_id, 4)
        >>> forecast.get_reorder(product_id)
        4
                
        >>> dummy_id = 98
        >>> forecast.get_reorder(dummy_id)
        """
        return self.reorder.get(product_id)
    
    def set_unused(self, product_id, is_unused=True):
        """ Mark a product as completely unused during the simulation.

        >>> product_id = 99
        >>> forecast = sim_forecast()
        >>> forecast.get_unused(product_id)
        False
        >>> forecast.set_unused(product_id)
        >>> forecast.get_unused(product_id)
        True
        >>> forecast.set_unused(product_id, is_unused=False)
        >>> forecast.get_unused(product_id)
        False
        """
        if is_unused:
            self.unused.add(product_id)
        else:
            self.unused.remove(product_id)
    
    def get_unused(self, product_id):
        return product_id in self.unused
    
    def get_data(self, product):
        """ Get data fields that need to be updated.
        
        product is a product browse object that holds all the current fields
        from the database.
        
        Returns a dictionary with entries for each field that needs to be 
        updated.
        
        >>> product = mock_browse()
        >>> product.id = 99
        >>> product.days_until_stockout = 10
        >>> product.days_until_reorder = 8
        >>> forecast = sim_forecast()
        >>> forecast.set_stockout(product.id, 10)
        >>> forecast.set_reorder(product.id, 8)
        >>> forecast.get_data(product)
        {}
        
        >>> forecast.set_stockout(product.id, 9)
        >>> forecast.get_data(product)
        {'days_until_stockout': 9}

        >>> forecast.set_reorder(product.id, 7)
        >>> forecast.get_data(product)
        {'days_until_stockout': 9, 'days_until_reorder': 7}
        
        If the product has no stockout or reorder, then it will return 10000.
        >>> product.id = 101
        >>> forecast.get_data(product)
        {'days_until_stockout': 10000, 'days_until_reorder': 10000}
        
        If the product was completely unused during the simulation, then it will
        return 20000.
        >>> forecast.set_unused(product.id)
        >>> forecast.get_data(product)
        {'days_until_stockout': 20000, 'days_until_reorder': 20000}
        """
        data = {}
        if product.id in self.unused:
            stockout = reorder = 20000
        else:
            stockout = self.stockouts.get(product.id, 10000)
            reorder = self.reorder.get(product.id, 10000)
        if product.days_until_stockout != stockout:
            data['days_until_stockout'] = stockout
        if product.days_until_reorder != reorder:
            data['days_until_reorder'] = reorder
        return data

if __name__ == "__main__":
    class mock_browse():
        """ Dummy class to use in place of a real browse object. """
        
    import doctest

    failures, tests = doctest.testmod()
    if failures == 0: print "Success."
