from stockout_snapshot import stockout_snapshot

class sim_product:
    """ Holds the details of a product used in the simulation.
    
    daily_sales is the quantity we forecast we will sell in an average day.
    uom is a unit of measure browse object for the daily_sales amount.
    delay is the number of days it takes for a supplier to ship or for a
        production order to complete.
    """
    def __init__(self, id, daily_sales=0, uom=None, delay=0, safety_days=0):
        """
        """
        self.id = id
        self.daily_sales = daily_sales
        self.uom = uom
        self.delay = delay
        self.safety_days = safety_days
        self.ignore_safety = False
        self.bom_id = None
        self.components = []
        self.replacement = None
    
    def consume_daily_sales(self, stock, uom_obj):
        """ Adjust the stock level for this product by the daily sales amount.
        
        >>> product_id = 99
        >>> product = sim_product(product_id, 0.25)
        >>> stock = sim_stock({product_id: 1.0})
        >>> product.consume_daily_sales(stock, mock_uom_obj())
        >>> stock
        sim_stock({99: 0.75})
        """
        self.consume(stock, self.daily_sales, uom_obj)
    

    def get_total_delay(self):
        """ Calculate the total delay for procurement and safety buffer. """
        delay = self.delay
        if not self.ignore_safety:
            delay += self.safety_days
        return delay

    def consume(self, stock, quantity, uom_obj):
        """ Adjust the stock level for this product by the requested amount.
        
        >>> product_id = 99
        >>> product = sim_product(product_id)
        >>> stock = sim_stock({product_id: 1.0})
        >>> product.consume(stock, 0.125, mock_uom_obj())
        >>> stock
        sim_stock({99: 0.875})
        
        When a product has components and its balance drops below zero, it will
        produce enough to bring its balance back up and record the reorder.
        >>> motor_id = 17
        >>> bom_id = 42
        >>> motor = sim_product(motor_id)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> stock = sim_stock({product_id: 0.0, motor_id: 1.0})
        >>> product.consume(stock, 0.25, mock_uom_obj())
        >>> stock
        sim_stock({17: 0.0, 99: 0.75})
        >>> stock.has_order(product_id)
        True
        
        A product may need to produce more than one to bring its balance back up.
        >>> stock = sim_stock({product_id: 0.0, motor_id: 5.0})
        >>> product.consume(stock, 2.25, mock_uom_obj())
        >>> stock
        sim_stock({17: 2.0, 99: 0.75})
        
        If a product with components and a delay triggers production, then it 
        will consume the components on a previous day, and produce the product 
        on the current day. It will also record the order on that previous day.
        >>> product.delay = 2.0
        >>> day1 = sim_stock({product_id: 0.0, motor_id: 1.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 0.25, mock_uom_obj())
        >>> day1
        sim_stock({17: 0.0, 99: 0.0})
        >>> day3
        sim_stock({17: 0.0, 99: 0.75})
        >>> day1.has_order(product_id)
        True
        
        Further consumption will continue to trigger orders, until all the 
        components have been stocked out and reordered.
        This example continues from the previous one.
        >>> day2.has_order(motor_id)
        False
        >>> day4 = day3.create_next()
        >>> product.consume(day4, 1.0, mock_uom_obj())
        >>> day2.has_order(motor_id)
        True
        >>> day4.get_quantity(product_id)
        0.75
        
        Now that the product and all its components have been stocked out and
        reordered, we no longer need to trigger production for it, and the
        balance can go below zero.
        This example continues from the previous one.
        >>> day5 = day4.create_next()
        >>> product.consume(day5, 1.0, mock_uom_obj())
        >>> day5.get_quantity(product_id)
        -0.25
        
        If a product has a delay and safety days, then it adds them when
        deciding when to place the order.
        >>> product.delay = 1.0
        >>> product.safety_days = 1.0
        >>> day1 = sim_stock({product_id: 0.0, motor_id: 1.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 0.25, mock_uom_obj())
        >>> day1.has_order(product_id)
        True
        
        For some simulations, safety days can be ignored.
        >>> product.delay = 1.0
        >>> product.safety_days = 1.0
        >>> product.ignore_safety = True
        >>> day1 = sim_stock({product_id: 0.0, motor_id: 1.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 0.25, mock_uom_obj())
        >>> day1.has_order(product_id)
        False
        >>> day2.has_order(product_id)
        True
        
        If the delay would take it back before the first day, then it stops at
        the first day in the simulation. For now, it assumes the production
        gets done quickly to finish today.
        >>> product.delay = 10.0
        >>> day1 = sim_stock({product_id: 0.0, motor_id: 1.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 0.25, mock_uom_obj())
        >>> day1.has_order(product_id)
        True
        >>> day1.has_order(motor_id)
        False
        >>> (day1.get_quantity(product_id), day1.get_quantity(motor_id))
        (0.0, 0.0)
        >>> (day3.get_quantity(product_id), day3.get_quantity(motor_id))
        (0.75, 0.0)
        
        If a product has no components and a delay, then it will record the 
        reorder.
        >>> product = sim_product(product_id)
        >>> product.delay = 2.0
        >>> day1 = sim_stock({product_id: 0.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 0.25, mock_uom_obj())
        >>> day1.has_order(product_id)
        True
        >>> day2.has_order(product_id)
        True
        >>> day3.has_order(product_id)
        True
        >>> day1.get_quantity(product_id)
        0.0
        >>> day3.get_quantity(product_id)
        -0.25
        
        If a product has no components and a delay, then a rewind may cause
        its reorder date to move earlier.
        >>> motor_id = 17
        >>> bom_id = 2
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> motor.delay = 1.0
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> day1 = sim_stock({product_id: 0.0, motor_id: 0.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> product.consume(day3, 1.0, mock_uom_obj())
        >>> day2.has_order(motor_id)
        True
        >>> day1.has_order(motor_id)
        False
        >>> product.consume(day2, 1.0, mock_uom_obj()) # after some other rewind
        >>> day1.has_order(motor_id) # should have moved earlier
        True
        
        If a product has a replacement, then consuming it below zero will
        actually consume the replacement.
        >>> new_motor_id = 314
        >>> old_motor = sim_product(motor_id)
        >>> new_motor = sim_product(new_motor_id)
        >>> old_motor.replacement = new_motor
        >>> day1 = sim_stock({motor_id: 1.0, new_motor_id: 10.0})
        >>> old_motor.consume(day1, 2.0, mock_uom_obj())
        >>> day1.has_order(motor_id)
        False
        >>> day1.has_order(new_motor_id)
        False
        >>> day1.get_quantity(motor_id)
        1.0
        >>> day1.get_quantity(new_motor_id)
        8.0
        
        """
        if quantity > stock.get_quantity(self.id) and self.replacement:
            self.replacement.consume(stock, quantity, uom_obj)
            return
        
        balance = stock.adjust_quantity(self.id, -quantity)
        if balance < 0:
#            print "consuming %f of %d leaves %f with %d delay and components %s." % (
#                quantity, 
#                self.id,
#                balance,
#                self.delay,
#                [c[0].id for c in self.components])
            is_producing = self.components and self.should_produce(stock)
            step = stock
            delay = self.get_total_delay()
            while delay > 0 and step.previous != None:
                delay -= 1
                step = step.previous
            if is_producing:
                qty_produced = 0
                while qty_produced + balance < 0:
                    qty_produced += self.qty_produced
                self.consume_components(step, qty_produced, uom_obj)
                stock.adjust_quantity(self.id, qty_produced)
            step.add_order(self.id)
    
    def should_produce(self, stock):
        """ Gets a flag for whether this product should still trigger production.
        
        Returns True if this product or any of its components have not stocked
        out yet or have not triggered a reorder yet, otherwise False.
        """
        if stock.get_real_quantity(self.id) >= 0:
            return True
        
        step = stock
        delay = self.delay
        while delay > 0 and step.previous != None:
            delay -= 1
            step = step.previous
            
        if not self.components:
            return not step.has_order(self.id)
        
        for component, qty, uom in self.components:
            if component.should_produce(step):
                return True
        return False 
    
    def consume_components(self, stock, quantity_produced, uom_obj):
        """ Consume enough components to produce quantity_produced of this
        product.
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> bom_id = 42
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> stock = sim_stock({product_id: 0.5, motor_id: 3.0})
        >>> product.consume_components(stock, 1.0, mock_uom_obj())
        >>> stock
        sim_stock({17: 2.0, 99: 0.5})
        >>> product.consume_components(stock, 2.0, mock_uom_obj())
        >>> stock
        sim_stock({17: 0.0, 99: 0.5})

        Now try a bom that produces a batch of products
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> product.set_production(bom_id, qty_produced=10.0)
        True
        >>> product.add_component(motor, quantity=10.0)
        >>> stock = sim_stock({product_id: 0.5, motor_id: 30.0})
        >>> product.consume_components(stock, 20.0, mock_uom_obj())
        >>> stock
        sim_stock({17: 10.0, 99: 0.5})
        """
        batch_count = quantity_produced / self.qty_produced
        for component, qty, uom in self.components:
            qty_consumed = batch_count * qty
            if uom == component.uom:
                converted_qty = qty_consumed
            else:
                converted_qty = uom_obj._compute_qty_obj(
                    None,
                    None,
                    uom,
                    qty_consumed,
                    component.uom)
            component.consume(stock, converted_qty, uom_obj)
    
    def set_production(self, bom_id, qty_produced=1.0, uom=None):
        """ Set the header information for producing this product.
        
        Returns True if successful, False if this information conflicts with
        the current information for this product.
        
        >>> bom_id1 = 123
        >>> bom_id2 = 222
        >>> product = sim_product(99, 0.5)
        >>> product.set_production(bom_id1, 1.0)
        True
        >>> product.set_production(bom_id1, 1.0)
        True
        >>> product.set_production(bom_id2, 1.0)
        False
        """
        if self.bom_id != None and self.bom_id != bom_id:
            return False
        self.bom_id = bom_id
        self.qty_produced = qty_produced
        return True
    
    def add_component(self, component, quantity=1.0, uom=None):
        """ Add a component to the production of this product.
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> bom_id = 123
        >>> product = sim_product(product_id, 0.5)
        >>> product.set_production(bom_id, 1.0)
        True
        >>> motor = sim_product(motor_id, 0)
        >>> product.add_component(motor, 1)
        >>> stock = sim_stock({product_id: 0, motor_id: 10})
        >>> product.consume_daily_sales(stock, mock_uom_obj())
        >>> stock
        sim_stock({17: 9.0, 99: 0.5})
        
        We can also handle uom conversion.
        >>> screw_id = 18
        >>> unit = 1
        >>> pack = 0.1
        >>> product = sim_product(product_id, 0.5, unit)
        >>> product.set_production(bom_id, 1.0, unit)
        True
        >>> screw = sim_product(screw_id, 0, unit)
        >>> product.add_component(screw, 1, pack) #consumes a pack of 10 screws
        >>> stock = sim_stock({product_id: 0, screw_id: 100})
        >>> product.consume_daily_sales(stock, mock_uom_obj())
        >>> stock
        sim_stock({18: 90.0, 99: 0.5})
        """
        
        self.components.append((component, quantity, uom))
        
    def get_all_snapshots(self, uom_obj=None, ancestors=None):
        """ Find all components, including nested ones and replacements. 
        Calculate delays and consumption ratios. Also include a snapshot for the
        product itself.
        
        uom_obj is used to convert units of measure when a BOM's unit of measure
        is different from the component's inventory unit of measure.
        ancestors is a list of product ids that form the chain in the BOM 
        structure from the root product down to this one.
        Returns a dictionary with component id as the key and snapshot as the
        value.
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> bom_id = 123
        >>> product = sim_product(product_id, 0.5)
        >>> product.set_production(bom_id, 1.0)
        True
        >>> motor = sim_product(motor_id, 0)
        >>> product.add_component(motor, 1)
        >>> product.get_all_snapshots()
        {17: stockout_snapshot(17), 99: stockout_snapshot(99)}
        
        Obviously, more than one component may be included.
        >>> screw_id = 18
        >>> screw = sim_product(screw_id)
        >>> product.add_component(screw)
        >>> product.get_all_snapshots()
        {17: stockout_snapshot(17), 18: stockout_snapshot(18), 99: stockout_snapshot(99)}
        
        Nested components should also be included.
        >>> wire_id = 19
        >>> nested_bom_id = 456
        >>> wire = sim_product(wire_id)
        >>> motor.set_production(nested_bom_id, 1.0)
        True
        >>> motor.add_component(wire)
        >>> product.get_all_snapshots()
        {19: stockout_snapshot(19), 17: stockout_snapshot(17), 18: stockout_snapshot(18), 99: stockout_snapshot(99)}
        
        Avoid infinite loop when boms have a cycle.
        >>> wire_id = 19
        >>> cycle_bom_id = 999
        >>> wire.set_production(cycle_bom_id, 1.0)
        True
        >>> wire.add_component(product)
        >>> product.get_all_snapshots()
        {19: stockout_snapshot(19), 17: stockout_snapshot(17), 18: stockout_snapshot(18), 99: stockout_snapshot(99)}

        Consumption ratio is calculated for all snapshots.
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> wire = sim_product(wire_id)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor, 2.0)
        >>> motor.set_production(nested_bom_id)
        True
        >>> motor.add_component(wire, 3.0)
        >>> snapshots = product.get_all_snapshots()
        >>> snapshots[product_id].consumption
        1.0
        >>> snapshots[motor_id].consumption
        2.0
        >>> snapshots[wire_id].consumption
        6.0
        
        Take BOM quantity into account
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> product.set_production(bom_id, qty_produced=10.0)
        True
        >>> product.add_component(motor, 20.0)
        >>> snapshots = product.get_all_snapshots()
        >>> snapshots[product_id].consumption
        1.0
        >>> snapshots[motor_id].consumption
        2.0
        
        Convert units of measure.
        >>> uom_obj = mock_uom_obj() 
        >>> unit = 1
        >>> pack = 0.1
        >>> product = sim_product(product_id, uom=unit)
        >>> product.set_production(bom_id, uom=unit)
        True
        >>> screw = sim_product(screw_id, uom=unit)
        >>> product.add_component(screw, quantity=1.0, uom=pack)
        >>> snapshots = product.get_all_snapshots(uom_obj)
        >>> snapshots[product_id].consumption
        1.0
        >>> snapshots[screw_id].consumption
        10.0

        Delays are calculated for all snapshots.
        >>> product = sim_product(product_id, delay=2.0)
        >>> motor = sim_product(motor_id, delay=4.0)
        >>> wire = sim_product(wire_id, delay=8.0)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor, 2.0)
        >>> motor.set_production(nested_bom_id)
        True
        >>> motor.add_component(wire, 3.0)
        >>> snapshots = product.get_all_snapshots()
        >>> snapshots[product_id].component_procurement_days
        2.0
        >>> snapshots[motor_id].component_procurement_days
        4.0
        >>> snapshots[wire_id].component_procurement_days
        8.0
        >>> snapshots[product_id].component_to_product_days
        0.0
        >>> snapshots[motor_id].component_to_product_days
        2.0
        >>> snapshots[wire_id].component_to_product_days
        6.0
        
        If subcomponent is used by more than one component, the amounts should
        be summed. In this example, both the motor and the screw
        contain wires. (Strange screws!)
        >>> product = sim_product(product_id)
        >>> motor = sim_product(motor_id)
        >>> screw = sim_product(screw_id)
        >>> wire = sim_product(wire_id)
        >>> product.set_production(bom_id)
        True
        >>> product.add_component(motor)
        >>> product.add_component(screw)
        >>> motor.set_production(nested_bom_id)
        True
        >>> motor.add_component(wire, quantity=2.0)
        >>> screw_bom_id = 777
        >>> screw.set_production(screw_bom_id)
        True
        >>> screw.add_component(wire, quantity=3.0)
        >>> snapshots = product.get_all_snapshots()
        >>> snapshots[wire_id].consumption
        5.0
        
        If any products have replacements, they have to be included as well.
        >>> new_motor_id = 314
        >>> product = sim_product(product_id, 0.5)
        >>> product.set_production(bom_id, 1.0)
        True
        >>> motor = sim_product(motor_id, 0)
        >>> product.add_component(motor, 1)
        >>> new_motor = sim_product(new_motor_id, 0)
        >>> motor.replacement = new_motor
        >>> product.get_all_snapshots()
        {17: stockout_snapshot(17), 314: stockout_snapshot(314), 99: stockout_snapshot(99)}
        
        """
        
        if ancestors == None:
            ancestors = []
        current = ancestors + [self.id]
        product_snapshot = stockout_snapshot(self.id)
        product_snapshot.component_procurement_days = self.delay
        snapshots = {self.id: product_snapshot}
        replacement = self.replacement
        if replacement and replacement.id not in ancestors:
            snapshots.update(self.replacement.get_all_snapshots(uom_obj, 
                                                                current))
        for component, qty, uom in self.components:
            if component.id not in ancestors:
                if component.uom == uom:
                    qty_consumed = qty
                else:
                    qty_consumed = uom_obj._compute_qty_obj(
                        None,
                        None,
                        uom,
                        qty,
                        component.uom)
                component_snapshots = component.get_all_snapshots(
                    uom_obj, 
                    current)
                for component_snapshot in component_snapshots.itervalues():
                    component_snapshot.component_to_product_days += (
                        product_snapshot.component_procurement_days)
                    component_snapshot.consumption *= (
                        qty_consumed / self.qty_produced)

                    duplicate_snapshot = snapshots.get(
                        component_snapshot.component_id)
                    if not duplicate_snapshot:
                        snapshots[component_snapshot.component_id] = (
                            component_snapshot)
                    else:
                        duplicate_snapshot.component_to_product_days = max(
                            duplicate_snapshot.component_to_product_days,
                            component_snapshot.component_to_product_days)
                        duplicate_snapshot.consumption += (
                            component_snapshot.consumption)
        return snapshots

if __name__ == "__main__":
    from sim_stock import sim_stock
    from mock_uom import mock_uom_obj
    import doctest
    
    failures, tests = doctest.testmod()
    if failures == 0: print "Success."
