from sim_forecast import sim_forecast

class sim_stock(object):
    """ Simulates stock levels for all products on a single day.
    """
    def __init__(self, source=None):
        """ Copy stock levels from a source day, or initialize as empty.
        
        >>> day1 = sim_stock()
        >>> day1
        sim_stock({})
        >>> day1.previous
        >>> day2 = sim_stock({99: 1.0})
        >>> day3 = sim_stock(day2)
        >>> day3
        sim_stock({99: 1.0})
        
        Even when you pass in a source, it doesn't become the previous day.
        >>> day3.previous
        
        You have to set the previous day explicitly.
        >>> day3.previous = day2
        
        When the source is another stock day, copy all attributes.
        >>> day3b = sim_stock(day3)
        >>> day3b.previous == day2
        True
        """
        
        self.unused = None
        if isinstance(source, sim_stock):
            self.previous = source.previous
            self.next = source.next
            self.tick = source.tick
            self.monitored_product_id = source.monitored_product_id
            self.monitored_quantity = source.monitored_quantity
            self.orders = set(source.orders)
            self.future = [dict(d) for d in source.future]
            self.stock = dict(source.stock)
            self.real_stock = dict(source.real_stock)
        else:
            self.previous = None
            self.next = None
            self.tick = 0
            self.monitored_product_id = None
            self.monitored_quantity = None
            self.orders = set()
            self.future = []
            if source == None:
                self.stock = {}
                self.real_stock = {}
            else:
                self.stock =  dict(source)
                self.real_stock = dict(source)
                
    def clone(self):
        """ Copy a stock object, along with all its future and past relatives.

        Changes to the copy will not affect the original.
        >>> product_id = 99
        >>> stock = sim_stock({product_id: 1.0})
        >>> clone = stock.clone()
        >>> clone.adjust_quantity(product_id, 5.0)
        6.0
        >>> clone.get_quantity(product_id)
        6.0
        >>> stock.get_quantity(product_id)
        1.0
        
        Future days are also duplicated.
        >>> stock = sim_stock({product_id: 1.0})
        >>> stock2 = stock.create_next()
        >>> stock2.adjust_quantity(product_id, 2.0)
        3.0
        >>> clone = stock.clone()
        >>> clone2 = clone.next
        >>> clone2.adjust_quantity(product_id, 5.0)
        8.0
        >>> clone2.get_quantity(product_id)
        8.0
        >>> stock2.get_quantity(product_id)
        3.0

        Past days are also duplicated.
        >>> stock = sim_stock({product_id: 1.0})
        >>> stock2 = stock.create_next()
        >>> stock2.adjust_quantity(product_id, 2.0)
        3.0
        >>> clone2 = stock2.clone()
        >>> clone = clone2.previous
        >>> clone.adjust_quantity(product_id, 5.0)
        6.0
        >>> clone.get_quantity(product_id)
        6.0
        >>> stock.get_quantity(product_id)
        1.0
        """
        clone = sim_stock(self)
        if self.unused != None:
            clone.unused = set(self.unused)
        step = clone
        while step.next:
            step.next = sim_stock(step.next)
            step.next.unused = step.unused
            step = step.next
        step = clone
        while step.previous:
            step.previous = sim_stock(step.previous)
            step.previous.unused = step.unused
            step = step.previous
        return clone

    def add_product(self, product_id, quantity):
        """ Track a new product, starting with the given quantity.
        """
        self.stock[product_id] = quantity
        self.real_stock[product_id] = quantity
    
    def get_quantity(self, product_id):
        """ Read the stock level for a product.
        
        >>> product_id = 99
        >>> stock = sim_stock({product_id: 1.0})
        >>> stock.get_quantity(product_id)
        1.0
        """
        
        return self.stock[product_id]
    
    def get_real_quantity(self, product_id):
        """ Read the stock level for a product, assuming no new ones are 
        purchased or produced.
        
        >>> product_id = 99
        >>> stock = sim_stock({product_id: 1.0})
        >>> stock.get_quantity(product_id)
        1.0
        >>> stock.get_real_quantity(product_id)
        1.0
        >>> stock.adjust_quantity(product_id, -0.75)
        0.25
        >>> stock.get_quantity(product_id)
        0.25
        >>> stock.get_real_quantity(product_id)
        0.25
        >>> stock.adjust_quantity(product_id, 1.0)
        1.25
        >>> stock.get_quantity(product_id)
        1.25
        >>> stock.get_real_quantity(product_id)
        0.25
        
        Adding a product with a start quantity sets both real and virtual stock.
        >>> screw_id = 18
        >>> stock.add_product(screw_id, 99.0)
        >>> stock.get_quantity(screw_id)
        99.0
        >>> stock.get_real_quantity(screw_id)
        99.0
        """
        return self.real_stock[product_id]
    
    def adjust_quantity(self, product_id, quantity):
        """ Change the quantity of the given product by the given amount.
        
        >>> product_id = 99
        >>> stock = sim_stock({product_id: 1.0})
        >>> stock.adjust_quantity(product_id, -0.75)
        0.25
        >>> stock.get_quantity(product_id)
        0.25
        
        Adjusting the quantity also makes the same adjustment in all future
        ticks.
        >>> day1 = sim_stock({product_id: 1.0})
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> day2.adjust_quantity(product_id, -1)
        0.0
        >>> day3.get_quantity(product_id)
        0.0
        """
        if quantity != 0.0 and self.unused != None:
            self.unused.discard(product_id)

        balance = self.stock[product_id] + quantity
        self.stock[product_id] = balance
        if product_id == self.monitored_product_id:
            self.monitored_quantity = balance
        if quantity < 0:
            real_balance = self.real_stock[product_id] + quantity
            self.real_stock[product_id] = real_balance
            if product_id == self.monitored_product_id:
                self.monitored_real_quantity = real_balance
        if self.next:
            self.next.adjust_quantity(product_id, quantity)
        return balance
    
    def monitor_product(self, product_id):
        """ Set up one special product whose stock level is quick to look up.
        
        Starting the monitor records the current quantity.
        >>> product_id = 99
        >>> motor_id = 17
        >>> day1 = sim_stock({product_id: 5.0, motor_id: 2.0})
        >>> day1.monitor_product(product_id)
        >>> day1.monitored_quantity
        5.0
        >>> day1.monitor_product(motor_id)
        >>> day1.monitored_quantity
        2.0
        
        You can also get the real quantity
        >>> day1.monitored_real_quantity
        2.0
        
        Adjusting the quantity of the monitored product updates the monitored
        quantity.
        >>> day1.adjust_quantity(motor_id, -1.0)
        1.0
        >>> day1.monitored_quantity
        1.0
        
        Adjusting the quantity of other products doesn't update the monitored
        quantity.
        >>> day1.adjust_quantity(product_id, -1.0)
        4.0
        >>> day1.monitored_quantity
        1.0
        
        Monitoring carries into copies of the day.
        >>> day2 = sim_stock(day1)
        >>> day2.monitored_quantity
        1.0
        >>> day2.adjust_quantity(motor_id, 2.0)
        3.0
        >>> day2.monitored_quantity
        3.0
        """
        self.monitored_product_id = product_id
        self.monitored_quantity = self.stock[product_id]
        self.monitored_real_quantity = self.real_stock[product_id]
    
    def expect(self, product_id, quantity, ticks):
        """ Add a shipment that will arrive in the future. The quantity will
        be applied when calling create_next().
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> day1 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day1.expect(product_id, 2.0, 1)
        >>> day2 = day1.create_next()
        >>> day2.get_quantity(product_id)
        3.0
        >>> day2.get_quantity(motor_id)
        0.0
        
        The stock does not increase until the day arrives.
        >>> day1 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day1.expect(motor_id, 1.0, 2)
        >>> day2 = day1.create_next()
        >>> day3 = day2.create_next()
        >>> day2.get_quantity(motor_id)
        0.0
        >>> day3.get_quantity(motor_id)
        1.0
        
        If the future day has already been created, then the quantity will
        be applied there.
        >>> day1 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day2 = day1.create_next()
        >>> day1.expect(product_id, 2.0, 1)
        >>> day2.get_quantity(product_id)
        3.0
        >>> day2.get_quantity(motor_id)
        0.0
        
        You can have multiple expectations on the same day, and also negative
        quantities (outgoing shipments).
        >>> day1 = sim_stock({product_id: 1.0})
        >>> day1.expect(product_id, 3.0, 1)
        >>> day1.expect(product_id, -1.0, 1)
        >>> day2 = day1.create_next()
        >>> day2.get_quantity(product_id)
        3.0
        """
        ticks = int(ticks)
        step = self
        while ticks > 0 and step.next:
            step = step.next
            ticks -= 1
        if ticks <= 0:
            step.adjust_quantity(product_id, quantity)
        else:
            while len(step.future) < ticks:
                step.future.append({})
            balance = step.future[ticks-1].get(product_id, 0.0) + quantity
            step.future[ticks-1][product_id] = balance
    
    def create_next(self):
        """ Create a new set of stock levels to represent the next tick in the
        simulation.
        
        >>> product_id = 99
        >>> day1 = sim_stock({product_id: 1.0})
        >>> day1.adjust_quantity(product_id, 1.0)
        2.0
        >>> day2 = day1.create_next()
        >>> day2.get_quantity(product_id)
        2.0
        >>> day2.get_real_quantity(product_id)
        1.0
        >>> day2.previous == day1
        True
        >>> day1.next == day2
        True
        """
        self.next = sim_stock(self)
        self.next.previous = self
        self.next.tick += 1
        self.next.unused = self.unused
        self.future = []
        if self.next.future:
            today = self.next.future.pop(0)
            for id, qty in today.iteritems():
                self.next.adjust_quantity(id, qty)
        return self.next
    
    def create_forecast(self):
        """ Walk through all previous days to the beginning and find on which
        tick each product's real stock went below zero, and when the first
        reorder was needed.
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> day0 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day1 = day0.create_next()
        >>> day1.adjust_quantity(motor_id, -1)
        -1.0
        >>> forecast = day1.create_forecast()
        >>> forecast.get_stockout(motor_id)
        1
        >>> forecast.get_stockout(product_id)
        
        It's based off real stock, not predicted increases.
        >>> day1.adjust_quantity(motor_id, 10)
        9.0
        >>> forecast = day1.create_forecast()
        >>> forecast.get_stockout(motor_id)
        1
        
        Obviously, different products can stock out on different ticks.
        >>> day2 = day1.create_next()
        >>> day2.adjust_quantity(product_id, -2)
        -1.0
        >>> forecast = day2.create_forecast()
        >>> forecast.get_stockout(motor_id)
        1
        >>> forecast.get_stockout(product_id)
        2
        
        The reorder ticks are filled in by calling add_order().
        >>> day1.add_order(product_id)
        >>> day2.add_order(motor_id)
        >>> forecast = day2.create_forecast()
        >>> forecast.get_reorder(product_id)
        1
        >>> forecast.get_reorder(motor_id)
        2
        
        Use the first reorder in the forecast.
        >>> day2.add_order(product_id)
        >>> forecast = day2.create_forecast()
        >>> forecast.get_reorder(product_id)
        1

        """
        if self.previous:
            forecast = self.previous.create_forecast()
            forecast.tick_count += 1
        else:
            forecast = sim_forecast()
            forecast.tick_count = 0
            
        for id, qty in self.real_stock.iteritems():
            if qty < 0 and forecast.get_stockout(id) == None:
                forecast.set_stockout(id, forecast.tick_count)
        for id in self.orders:
            if forecast.get_reorder(id) == None:
                forecast.set_reorder(id, forecast.tick_count)
        if self.unused != None:
            for id in self.unused:
                forecast.set_unused(id)
        return forecast
    
    def track_unused(self):
        """ Track which products were unused.
        
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> day0 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day0.track_unused()
        >>> day1 = day0.create_next()
        >>> day1.adjust_quantity(motor_id, -1)
        -1.0
        >>> forecast = day1.create_forecast()
        >>> forecast.get_unused(product_id)
        True
        >>> forecast.get_unused(motor_id)
        False
        
        Adjusting by quantity zero is not considered "using" a product.
        >>> day1.adjust_quantity(product_id, 0.0)
        1.0
        >>> forecast = day1.create_forecast()
        >>> forecast.get_unused(product_id)
        True
        >>> forecast.get_unused(motor_id)
        False

        If you don't call track_unused(), the forecast doesn't notice they were
        unused.
        >>> day0 = sim_stock({product_id: 1.0, motor_id: 0.0})
        >>> day1 = day0.create_next()
        >>> day1.adjust_quantity(motor_id, -1)
        -1.0
        >>> forecast = day1.create_forecast()
        >>> forecast.get_unused(product_id)
        False
        >>> forecast.get_unused(motor_id)
        False

        """
        self.unused = set(self.stock.keys())
    
    def add_order(self, product_id):
        """ Place an order on this tick for the given product. This doesn't 
        affect any stock levels, it's just used to fill in the forecast when
        create_forecast is called.
        
        >>> product_id = 99
        >>> motor_id = 17
        >>> stock = sim_stock()
        >>> stock.add_order(product_id)
        >>> stock.has_order(product_id)
        True
        >>> stock.has_order(motor_id)
        False
        
        Orders carry forward to future days.
        >>> stock.create_next().has_order(product_id)
        True
        
        Orders also carry forward if the future days are already created.
        >>> day1 = sim_stock()
        >>> day2 = day1.create_next()
        >>> day1.add_order(product_id)
        >>> day2.has_order(product_id)
        True
        """
        step = self
        while step != None:
            step.orders.add(product_id)
            step = step.next
    
    def has_order(self, product_id):
        return product_id in self.orders
    
    def __repr__(self):
        """ Format a stock simulation for creation.
        
        >>> sim_stock({99: 1.0})
        sim_stock({99: 1.0})
        """
        
        return "sim_stock(%s)" % self.stock
    
    def __iter__(self):
        """ Iterate through the stock levels for this day. Each item is a tuple
        of (product_id, quantity).
        
        """
        return self.stock.iteritems()

if __name__ == "__main__":
    import doctest
    failures, tests = doctest.testmod()
    if failures == 0:
        print "Success."
