from datetime import datetime
from datetime import timedelta
import logging

from osv import osv, fields

class inventory_processing(osv.osv_memory):
    _name = 'inventory.processing'
    _description = 'Inventory Processing'

    # the following data fields are not used for data keeping
    # they exist so the fields can be displayed in the view
    #   date_procurements, date_minimum_stock_rules, date_sales_volumes,
    #   date_stockouts, date_build_capacities, date_production_costs
    _columns = {'is_finished': fields.boolean('Is Finished'),
                'run_procurements': fields.boolean(
                    'Run Procurements',
                    help='Trigger all active procurements to see if they ' +
                        'are ready to complete.'),
                'date_procurements': fields.date('Last Run: ', readonly=True),
                'minimum_stock_rules': fields.selection(
                    (('run', 'Run'),
                     ('auto', 'Run with automatic minimum')),
                    'Minimum Stock Rules',
                    help='Check all minimum stock rules and generate any ' +
                        'needed procurements.'),
                'date_minimum_stock_rules': fields.date('Last Run: ', readonly=True),
                'run_sales_volumes': fields.boolean(
                    'Run Sales Volumes',
                    help='Update the sales history for all products, as ' +
                        'well as the daily and yearly sales and consumption.'),
                'date_sales_volumes': fields.date('Last Run: ', readonly=True),
                'sales_months': fields.integer(
                    'Sales months', 
                    help='Number of months of sales orders to include in ' +
                        'the calculation'), 
                'run_stockouts': fields.boolean(
                    'Run Stockout Simulation',
                    help='Calculate days until stockout and days until ' +
                        'reorder for all products, as well as likely ' +
                        'stockouts.'),
                'date_stockouts': fields.date('Last Run: ', readonly=True),
                'run_build_capacities': fields.boolean(
                    'Run Build Capacities',
                    help='Calculate how many of each product we can build ' +
                        'with current stock. Also update quantities in ' +
                        'outstanding procurements (purchase orders or slow ' +
                        'production orders) and already committed to ' +
                        'assemblies and finished products. These are ' +
                        'independent of the stockout simulation.'),
                'date_build_capacities': fields.date('Last Run: ', readonly=True),
                'run_production_costs': fields.boolean(
                    'Run Production Costs',
                    help="Sum up the cost of each product's components."),
                'date_production_costs': fields.date('Last Run: ', readonly=True),
                'products_changed': fields.integer(
                    'Number of Products Updated', 
                    readonly=True)}
    
    def default_get(self, cr, uid, fields, context=None):
                                                           
        tags = {'date_procurements': ['run_procurements','run_scheduler'], 
                'date_minimum_stock_rules': ['minimum_stock_rules', ''],
                'date_sales_volumes': ['run_sales_volumes', '_calculate_all_sales_volumes'],
                'date_stockouts': ['run_stockouts', '_calculate_all_stockouts'],
                'date_build_capacities': ['run_build_capacities', '_calculate_all_build_capacities'],
                'date_production_costs': ['run_production_costs', '_calculate_all_production_costs']}
        defaults = {}        
        defaults['sales_months'] = 3 # before: _defaults = {'sales_month': lambda *a: 3 }
        for tag in tags:
        # searching for the newest run of each from the scheduler and the current (manual) window
            if tag in fields:
                if tag == "date_minimum_stock_rules":
                    param = "%s IS NOT NULL" % tags[tag][0]
                else:
                    param = "%s = 't'" % tags[tag][0]
                cr.execute('''
                   SELECT * FROM inventory_processing
                   WHERE %s AND
                   is_finished = 't'
                   ORDER BY create_date DESC
                   LIMIT 1''' % param)
                newest = cr.fetchone()
                if newest:
                    newest_date = datetime.strptime(newest[3], '%Y-%m-%d %H:%M:%S.%f')
                else:
                    newest_date = None
                    
                if tags[tag][1]:
                    cr.execute('''
                        SELECT * FROM ir_cron
                        WHERE function = '%s'
                    ''' % tags[tag][1])
                    auto_newest = cr.fetchone()
                    auto_next_call = datetime.strptime(auto_newest[11], '%Y-%m-%d %H:%M:%S')
                    auto_interval = eval('timedelta('+str(auto_newest[6])+'='+str(auto_newest[15])+')')
                    # FIXME work_days and weeks interval types are not taken care of
                    #   auto newest date is calculated by subtracting the interval from the next run date
                    #   the interval with units work_dates and weeks are ignored

                    auto_newest_date = auto_next_call - auto_interval
                    
                    if newest_date is None or auto_newest_date > newest_date: 
                        defaults[tag] = auto_newest_date.strftime('%Y-%m-%d %H:%M:%S')
                    else:
                        defaults[tag] = newest_date.strftime('%Y-%m-%d %H:%M:%S')
                else:
                    if newest_date is None: continue
                    else: defaults[tag] = newest_date.strftime('%Y-%m-%d %H:%M:%S')
        return defaults
                
    def run(self, cr, uid, ids, context=None):
        products_changed = 0
        for request in self.browse(cr, uid, ids):
            start_time = datetime.now()
            proc_obj = self.pool.get('procurement.order')
            prod_obj = self.pool.get('product.product')
            
            
            if request.run_procurements:
                proc_obj._procure_confirm(cr, 
                                          uid, 
                                          context=context)
            if request.minimum_stock_rules:
                automatic = request.minimum_stock_rules == 'auto'
                proc_obj._procure_orderpoint_confirm(cr, 
                                                     uid, 
                                                     automatic=automatic,
                                                     context=context)
            if request.run_sales_volumes:
                prod_obj._calculate_all_sales_volumes(
                    cr, 
                    uid, 
                    request.sales_months,
                    context)
            if request.run_stockouts:
                products_changed += prod_obj._calculate_all_stockouts(
                    cr, 
                    uid, 
                    context)
            if request.run_build_capacities:
                products_changed += prod_obj._calculate_all_build_capacities(
                    cr, 
                    uid, 
                    context)
            if request.run_production_costs:
                products_changed += prod_obj._calculate_all_production_costs(
                    cr, 
                    uid, 
                    context)
                
            duration = datetime.now() - start_time
            logger = logging.getLogger('inventory.processing')
            logger.info('Inventory processing took %ss.' % duration.seconds)
                
        self.write(cr, uid, ids, {'is_finished': True, 
                                  'products_changed': products_changed})
        return {} # no further action required
