from datetime import date, timedelta

def add_business_days(start_date, days):
    """ Add a number of business days to a date, skipping weekends.

    returns the new date

    1 Nov 2010 is a Monday. Here's a calendar to think about the test cases:
    M  Tu W  Th F  Sa Su
     1  2  3  4  5  6  7
     8  9 10 11 12 13 14
    15 16 17 18 19 20 21
    22 23 24 25 26 27 28
    29 30

    Within the week, it's easy.
    >>> add_business_days(date(2010, 11, 1), 1)
    datetime.date(2010, 11, 2)
    >>> add_business_days(date(2010, 11, 1), 4)
    datetime.date(2010, 11, 5)

    Skip a weekend.
    >>> add_business_days(date(2010, 11, 1), 5)
    datetime.date(2010, 11, 8)

    Start on a weekend.
    >>> add_business_days(date(2010, 11, 6), 1)
    datetime.date(2010, 11, 8)
    >>> add_business_days(date(2010, 11, 7), 1)
    datetime.date(2010, 11, 8)
    >>> add_business_days(date(2010, 11, 7), 6)
    datetime.date(2010, 11, 15)

    Floating point days values also work, and fractions are truncated.
    >>> add_business_days(date(2010, 11, 1), 4.0)
    datetime.date(2010, 11, 5)
    >>> add_business_days(date(2010, 11, 1), 4.999)
    datetime.date(2010, 11, 5)

    Zero days makes no change.
    >>> add_business_days(date(2010, 11, 1), 0)
    datetime.date(2010, 11, 1)

    Negative days would require different logic, so we haven't implemented them.
    >>> add_business_days(date(2010, 11, 1), -1)
    datetime.date(2010, 10, 29)


    Deal with non-leap year dates

    February 2010
    Su Mo Tu We Th Fr Sa
       1  2  3  4  5  6
    7  8  9 10 11 12 13
    14 15 16 17 18 19 20
    21 22 23 24 25 26 27
    28

    >>> add_business_days(date(2010, 02, 27), 1)
    datetime.date(2010, 3, 1)

    Deal with leap-year dates

       February 2012
    Su Mo Tu We Th Fr Sa
              1  2  3  4
     5  6  7  8  9 10 11
    12 13 14 15 16 17 18
    19 20 21 22 23 24 25
    26 27 28 29

    >>> add_business_days(date(2012, 02, 28), 1)
    datetime.date(2012, 2, 29)

    >>> add_business_days(date(2012, 02, 28), 2)
    datetime.date(2012, 3, 1)

    Deal with leap-year dates where the leap day is on a Monday

       February 2016
    Su Mo Tu We Th Fr Sa
        1  2  3  4  5  6
     7  8  9 10 11 12 13
    14 15 16 17 18 19 20
    21 22 23 24 25 26 27
    28 29

    >>> add_business_days(date(2016, 02, 26), 1)
    datetime.date(2016, 2, 29)

    >>> add_business_days(date(2016, 02, 26), 2)
    datetime.date(2016, 3, 1)
    
    Also works with datetime input.
    >>> add_business_days(datetime(2010, 11, 1, 10, 52), 1)
    datetime.datetime(2010, 11, 2, 10, 52)
    >>> add_business_days(datetime(2010, 11, 1, 17, 31), 5)
    datetime.datetime(2010, 11, 8, 17, 31)

    """

    if days < 0:
        return sub_business_days(start_date,-days)

    # To make it easier to deal with weekends, we adjust the start date
    # back to the closest Monday.
    monday = start_date - timedelta(days=start_date.weekday())
    # Add a corresponding extra few days to the time period. (Not for weekends.)
    days_from_monday = int(days) + min(4, start_date.weekday())

    # Now add extra days for weekends. Divide by 5 to calculate number of whole
    # weeks and multiply by 7 to get number of days including weekends.
    calendar_days = (days_from_monday / 5) * 7 + days_from_monday % 5
    end_date = monday + timedelta(days=calendar_days)
    return end_date

def sub_business_days(start_date, days):
    """ Subtract a number of business days to a date, skipping weekends.

    returns the new date

    1 Nov 2010 is a Monday. Here's a calendar to think about the test cases:
    M  Tu W  Th F  Sa Su
     1  2  3  4  5  6  7
     8  9 10 11 12 13 14
    15 16 17 18 19 20 21
    22 23 24 25 26 27 28
    29 30

    Within the week, it's easy.
    >>> sub_business_days(date(2010, 11, 3), 1)
    datetime.date(2010, 11, 2)
    >>> sub_business_days(date(2010, 11, 5), 4)
    datetime.date(2010, 11, 1)

    Skip a weekend.
    >>> sub_business_days(date(2010, 11, 10), 5)
    datetime.date(2010, 11, 3)

    Start on a weekend.
    >>> sub_business_days(date(2010, 11, 13), 1)
    datetime.date(2010, 11, 12)
    >>> sub_business_days(date(2010, 11, 14), 1)
    datetime.date(2010, 11, 12)
    >>> sub_business_days(date(2010, 11, 21),6)
    datetime.date(2010, 11, 12)

    Floating point days values also work, and fractions are truncated.
    >>> sub_business_days(date(2010, 11, 5), 4.0)
    datetime.date(2010, 11, 1)
    >>> sub_business_days(date(2010, 11, 5), 4.999)
    datetime.date(2010, 11, 1)

    Zero days makes no change.
    >>> sub_business_days(date(2010, 11, 1), 0)
    datetime.date(2010, 11, 1)

    Negative days would require different logic, so we haven't implemented them.
    >>> sub_business_days(date(2010, 11, 1), -1)
    datetime.date(2010, 11, 2)


    Deal with non-leap year dates

    February 2010
    Su Mo Tu We Th Fr Sa
       1  2  3  4  5  6
    7  8  9 10 11 12 13
    14 15 16 17 18 19 20
    21 22 23 24 25 26 27
    28

    >>> sub_business_days(date(2010, 03, 1), 1)
    datetime.date(2010, 2, 26)

    Deal with leap-year dates

       February 2012
    Su Mo Tu We Th Fr Sa
              1  2  3  4
     5  6  7  8  9 10 11
    12 13 14 15 16 17 18
    19 20 21 22 23 24 25
    26 27 28 29

    >>> sub_business_days(date(2012, 02, 29), 1)
    datetime.date(2012, 2, 28)

    >>> sub_business_days(date(2012, 03, 01), 2)
    datetime.date(2012, 2, 28)

    Deal with leap-year dates where the leap day is on a Monday

       February 2016
    Su Mo Tu We Th Fr Sa
        1  2  3  4  5  6
     7  8  9 10 11 12 13
    14 15 16 17 18 19 20
    21 22 23 24 25 26 27
    28 29

    >>> sub_business_days(date(2016, 02, 29), 1)
    datetime.date(2016, 2, 26)

    >>> sub_business_days(date(2016, 03, 01), 2)
    datetime.date(2016, 2, 26)

    """

    if days < 0:
        return add_business_days(start_date,-days)

    # To make it easier to deal with weekends, we adjust the start date
    # ahead to the closest friday
#    print start_date.weekday()
    day_delta = 4 - start_date.weekday()
    friday = start_date + timedelta(days=day_delta)
    # Add a corresponding extra few days to the time period. (Not for weekends.)
    days_from_friday = int(days) + max(day_delta,-1)
#    print days_from_friday

    # Now add extra days for weekends. Divide by 5 to calculate number of whole
    # weeks and multiply by 7 to get number of days including weekends.
    calendar_days = (days_from_friday / 5) * 7 + days_from_friday % 5
    end_date = friday - timedelta(days=calendar_days)
    return end_date

if __name__ == "__main__":
    import doctest
    from datetime import datetime
    failures, tests = doctest.testmod()
    if failures == 0:
        print "Success."
