from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import pytz

from osv import fields, osv
from tools import DEFAULT_SERVER_DATETIME_FORMAT

class sale_order(osv.osv):
    _inherit = "sale.order"
    
    _columns = {
        'days_valid': fields.integer('Days Quote is Valid', 
                                     readonly=True, 
                                     states={'draft': [('readonly', False)]}),
    }
    
    def onchange_shop_id(self, cr, uid, ids, shop_id):
        result = super(sale_order, self).onchange_shop_id(cr, uid, ids, shop_id)
        if shop_id:
            shop = self.pool.get('sale.shop').browse(cr, uid, shop_id)
            result['value']['days_valid'] = shop.quote_days_valid_default
        return result

    def _get_date_planned(self, cr, uid, order, line, start_date, context=None):
        """ If there is no current context, as in a workflow, we have to look 
        up which time zone to use based on the sales order's sales person's 
        user preferences."""
        if context is None and order.user_id:
            context = {'tz':order.user_id.context_tz}
        start_utc = fields.datetime.context_timestamp_from_business_date(
            cr, 
            uid, 
            start_date, 
            context)
        
        date_planned = start_utc + relativedelta(days=line.delay or 0.0)    
        date_planned = (date_planned - 
                        timedelta(days=order.company_id.security_lead))
        result = date_planned.strftime(DEFAULT_SERVER_DATETIME_FORMAT)
        return result
    
    def _prepare_order_picking(self, cr, uid, order, context=None):
        """ If there is no current context, as in a workflow, we have to look 
        up which time zone to use based on the sales order's sales person's 
        user preferences."""
        if context is None and order.user_id:
            context = {'tz':order.user_id.context_tz}
        data = super(sale_order, self)._prepare_order_picking(cr,
                                                              uid,
                                                              order,
                                                              context=context)
        data['date'] = fields.datetime.context_timestamp_from_business_date(
            cr, 
            uid, 
            data['date'], 
            context)
        return data
sale_order()

class sale_order_line(osv.osv):
    _inherit = "sale.order.line"

    def _notes_shortened (self, cr, uid, ids, field_name, arg, context):
        """ Return the first line of the notes field for display in the tree.
        
        Also truncates the first line if it's too long.
        """
        PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH = 80
        result = {}
        for product in self.browse(cr, uid, ids, context=context):
            notes = product.notes
            if not notes:
                result[product.id] = notes
            else:
                shortened = notes.partition('\n')[0]
                if ( len(shortened) < PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH ): 
                    result[product.id] = shortened 
                else:
                    result[product.id] = shortened[:PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH] + " ..."
        return result

    def _name_tagged (self, cr, uid, ids, field_name, arg, context):
        """ Return the name field with an extra tag if notes field isn't empty.
        
        """
        result = {}
        for product in self.browse(cr, uid, ids, context=context):
            if not product.notes:
                result[product.id] = product.name
            else:
                result[product.id] = product.name + " *HAS NOTES*"
        return result

    _columns = {
        'date_order': fields.related(
            'order_id',
            'date_order', 
            string='Order Date', 
            type='date',
            readonly=True),
        'notes_shortened': fields.function(
              _notes_shortened,
              type = 'char',
              method = True,
              string = 'Note'),
        'name_tagged': fields.function(
              _name_tagged,
              type = 'char',
              method = True,
              string = 'Name')
    }

sale_order_line()

class sale_shop(osv.osv):
    _inherit = "sale.shop"
    
    _columns = {
        'quote_days_valid_default': fields.integer('Default Quote Valid Days'),
    }
    
    _defaults = {
        'quote_days_valid_default': 90,
    }

sale_shop()
