import datetime
from osv import fields,osv,orm
from tools import config
from tools.translate import _

class hr_employee(osv.osv):
    _inherit = 'hr.employee'
    _order = 'work_email'

    _columns = {
        'code': fields.char('Code', 
                            size=32, 
                            help='Used for insurance providers and others'),
    }

hr_employee()

class hr_timesheet_sheet(osv.osv):
    _inherit = 'hr_timesheet_sheet.sheet'

    def read(self, cr, user, ids, fields=None, context=None, load='_classic_read'):
        """ Work around a client bug by removing some expensive fields
        when they're not needed. """
        trimmed_fields = fields
        if not isinstance(ids, (int, long)) and len(ids) > 1 and fields:
            extra_fields = ['timesheet_ids',
                            # Colin Ligertwood <colin@zaber.com> 2013-08-13
                            #
                            # Commented lines represent fields needed by
                            # zerp_timesheet_alert that are clobbered by this
                            # function
                            #
                            # 'user_id',
                            # 'name',
                            'attendances_ids',
                            'state_attendance',
                            # 'company_id',
                            'period_ids',
                            'total_timesheet_day',
                            # 'date_current',
                            'total_difference_day',
                            'total_attendance_day']
            trimmed_fields = [field 
                              for field in fields 
                              if field not in extra_fields]
        return super(hr_timesheet_sheet, self).read(cr,
                                                    user,
                                                    ids,
                                                    trimmed_fields,
                                                    context=context,
                                                    load=load)

    def create(self, cr, uid, vals, context=None):
        """ Call the regular create() and then check that the sign in/out
        attendances alternate as expected.
        """
        id = super(hr_timesheet_sheet, self).create(cr, 
                                                    uid, 
                                                    vals, 
                                                    context=context)
        sheet = self.browse(cr, uid, id, context)
        self._validate_attendance_actions(cr, uid, sheet, context)
        return id

    def write(self, cr, uid, ids, vals, context=None):
        """ Call the regular write() and then check that the sign in/out
        attendances alternate as expected.
        """
        result = super(hr_timesheet_sheet, self).write(cr, 
                                                       uid, 
                                                       ids, 
                                                       vals, 
                                                       context=context)
        if 'attendances_ids' in vals:
            for sheet in self.browse(cr, uid, ids, context):
                self._validate_attendance_actions(cr, uid, sheet, context)
        return result
    
    def _validate_attendance_actions(self, cr, uid, sheet, context=None):
        """ Validate that the attendance actions occur in a legal order:
        alternating between 'sign_in' and 'sign_out', with 'action' ignored.
        """
        attendances = list(sheet.attendances_ids)
        if not attendances:
            return
        attendances.sort(key=lambda att: att.name)
        
        interesting_actions = ('sign_in', 'sign_out')
        att_obj = self.pool.get('hr.attendance')
        prev_att_ids = att_obj.search(
            cr, 
            uid, 
            [('employee_id', '=', sheet.employee_id.id), 
             ('name', '<', attendances[0].name), 
             ('action', 'in', interesting_actions)],
            limit=1, 
            order='name DESC')
        next_att_ids = att_obj.search(
            cr, 
            uid, 
            [('employee_id', '=', sheet.employee_id.id), 
             ('name', '>', attendances[-1].name), 
             ('action', 'in', interesting_actions)],
            limit=1, 
            order='name ASC')
        neighbour_ids = list(prev_att_ids)
        neighbour_ids.extend(next_att_ids)
        neighbours = att_obj.browse(cr, uid, neighbour_ids, context=context)
        attendances.extend(neighbours)
        attendances.sort(key=lambda att: att.name)

        prev_att = None
        for att in attendances:
            if att.action not in interesting_actions:
                continue
            if prev_att and prev_att.action == att.action:
                datetime_format = '%Y-%m-%d %H:%M:%S'
                att_time = datetime.datetime.strptime(att.name,
                                                      datetime_format)
                prev_att_time = datetime.datetime.strptime(prev_att.name,
                                                           datetime_format)
                att_local = fields.datetime.context_timestamp(
                    cr,
                    uid,
                    att_time,
                    context=context)
                prev_att_local = fields.datetime.context_timestamp(
                    cr,
                    uid,
                    prev_att_time,
                    context=context)
                raise osv.except_osv(_('Error!'), 
                                     'Action %s at %s cannot follow %s at %s.' %
                                     (att.action,
                                      att_local.strftime(datetime_format),
                                      prev_att.action,
                                      prev_att_local.strftime(datetime_format)))
            prev_att = att
        
hr_timesheet_sheet()

class hr_attendance(osv.osv):
    _inherit = 'hr.attendance'
    
    def _altern_si_so(self, cr, uid, ids):
        """ Implementing this logic at the attendance level doesn't work, so
        we skip it, and check at the whole time sheet level. 's all good!"""
        return True

    _constraints = [(_altern_si_so, 
                     'Error: You should never see this message.', 
                     ['action'])]
    
hr_attendance()

class hr_timesheet_sheet_sheet_day(osv.osv):
    _inherit = "hr_timesheet_sheet.sheet.day"

    def read(self, cr, user, ids, fields=None, context=None, load='_classic_read'):
        """ Take a shortcut to avoid an expensive, pointless query. """
        if fields != ['__last_update']:
            return super(hr_timesheet_sheet_sheet_day, self).read(cr,
                                                                  user,
                                                                  ids,
                                                                  fields,
                                                                  context,
                                                                  load)
        else:
            now = datetime.datetime.today().strftime('%Y-%m-%d %H:%M:%S')
            if isinstance(ids, (int, long)):
                ids = [ids]
            return [dict(id=id, __last_update=now)
                    for id in ids]
    
hr_timesheet_sheet_sheet_day()
