import logging
from osv import fields,osv

from property_filter import exclude_properties

class product_product(osv.osv):
    ''' If the product has a document_id associated with it, let's include that
        as part of the reference code. Useful especially on POs
        Create a field that shows the default partner/supplier name and that 
        partner's code for the part
        Add a field for the name of the default supplier
    '''

    _inherit = 'product.product'

    def _notes_shortened (self, cr, uid, ids, field_name, arg, context):
    # --------------------------------------------------
        PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH = 15
        result = {}
        for product in self.browse(cr, uid, ids, context=context):
            notes = product.notes
            if not notes:
                result[product.id] = notes
            else:
                shortened = notes.partition('\n')[0]
                if ( len(shortened) < PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH ): 
                    result[product.id] = shortened 
                else:
                    result[product.id] = (
                        shortened[:PRODUCT_NOTES_TREE_VIEW_MAX_LENGTH] + " ...")
        return result


    def _product_partner_ref(self, cr, uid, ids, name, arg, context=None):
        res = super(product_product, self)._product_partner_ref(cr, 
                                                                uid, 
                                                                ids, 
                                                                name, 
                                                                arg, 
                                                                context)

        for p in self.browse(cr, uid, ids, context):
            data = self._get_partner_code_name(cr, 
                                               uid, 
                                               [], 
                                               p, 
                                               context.get('partner_id', None), 
                                               context)

            # If we can find a document_id we want to add it to the description.
            # Otherwise we leave it alone
            if 'document_id' not in data:
                data['document_id'] = p.product_tmpl_id.document_id
            if not data['document_id']:
                continue

            if not data['code']:
                data['code'] = p.default_code
            if data['code']:
                data['code'] += ', '

            if not data['name']:
                data['name'] = p.name
            res[p.id] = ('['+data['code']+data['document_id']+'] ') + \
                        (data['name'] or '')

        return res

    def _product_supplier_name(self, cr, uid, ids, name, arg, context=None):
        res = {}
        for product in self.browse(cr, uid, ids, context):
            supplier_name = ""
            if len(product.seller_ids) > 0:
                supplier_name = product.seller_ids[0].name.name
            res[product.id] = supplier_name
        return res

    def _product_supplier_name_search(self, cr, uid, obj, name, args, context):

        # How are we filtering?
        op = args[0][1]

        # We only know one way of search
        if op != "ilike":
            logger = logging.getLogger('zaber_custom')
            logger.warning(("Do not know how to handle operation '%s' " +
                            "on product.product.default_partner_name " +
                            "field") % (op,))
            return [('id','=','0')]

        # Now let's see which products have any items which match the partners
        supply_obj = self.pool.get('product.supplierinfo') 
        supply_ids = supply_obj.search(cr,uid,[('name','ilike',args[0][2])])
        product_ids = []
        for supply_rec in supply_obj.browse(cr,uid,supply_ids,context):

            # This makes sure that the current supplier is the default supplier
            if supply_rec.product_id.seller_ids[0].id == supply_rec.id:
                product_ids.append(supply_rec.product_id.id)

        return [('id','in',product_ids)]

    def name_search(self, 
                    cr, 
                    user, 
                    name='', 
                    args=None, 
                    operator='ilike', 
                    context=None, 
                    limit=80):
        """ The offical product.product module searches using the following 
        rules:
            1. Search for an exact match on default_code
            2. No matches on #1, then search for an exact match on ean13
            3. No matches on #1, or #2, then do substring matches on
                default_code and product_names

            That was slightly confusing for users so we're going to go 
            immediately to #3.
        """
        if not args:
            args=[]
        if not context:
            context={}
        if name:
            ids = self.search(cr, 
                              user, 
                              [('default_code',operator,name)]+ args, 
                              limit=limit, 
                              context=context)
            ids += self.search(cr, 
                               user, 
                               [('name',operator,name)]+ args, 
                               limit=limit, 
                               context=context)
        else:
            ids = self.search(cr, user, args, limit=limit, context=context)
        result = self.name_get(cr, user, ids, context)
        return result

    def _dirname(self, cr, uid, ids, name, arg, context=None):
        """ The directory name to use for the FTP server. """
        
        return dict([
            (product.id, '%s.REV%d' % (product.default_code, product.revision))
            for product in self.browse(cr, uid, ids, context)])
        
    def _dirsearch(self, cr, uid, obj, name, args, context):
        result = []
        delimiter = '.REV'
        for field, operator, value in args:
            if result:
                result.insert(0, '&')
            parts = value.split(delimiter)
            if len(parts) > 1:
                code, revision = delimiter.join(parts[0:-1]), parts[-1]
                result.append('&')
                result.append(self.__build_condition('revision', revision))
            else:
                code = value 
            result.append(self.__build_condition('default_code', code))
        return result
    
    def __build_condition(self, field, target):
        if '*' not in target and '?' not in target:
            operator = '='
            comparator = target
        else:
            operator = 'like'
            comparator = (
                target
                .replace('\\', '\\\\')
                .replace('%', '\\%')
                .replace('_', '\\_')
                .replace('*', '%')
                .replace('?', '_'))
        return (field, operator, comparator)

    _columns = {
        # Replace default_code to give it a new name.
        'default_code' : fields.char('Code', size=64, select=True),
        'notes': fields.text('Notes'),
        'notes_shortened': fields.function(
              _notes_shortened,
              type = 'char',
              method = True,
              string = 'Note'),
        'partner_ref': fields.function(_product_partner_ref, 
                                       method=True, 
                                       type='char', 
                                       string='Customer ref'),
        'default_partner_name': fields.function(
                                    _product_supplier_name, 
                                    method=True, 
                                    type='char', 
                                    readonly=True,
                                    fnct_search=_product_supplier_name_search,
                                    string='Supplier Name'),
        'dirname': fields.function(
            _dirname, 
            method=True,
            fnct_search=_dirsearch, 
            type='char',
            string='Directory Name')
    }
    
    def _generate_order_by(self, order_spec, query):
        """ Calculate the order by clause to use in SQL based on a set of
        model fields. """
        if order_spec != 'default_partner_name':
            return super(product_product, self)._generate_order_by(order_spec,
                                                                   query)
        return """
            ORDER BY
            (
            select  min(supp.name)
            from    product_supplierinfo supinf
            join    res_partner supp
            on      supinf.name = supp.id
            where   supinf.product_id = product_product.id
            ),
            product_product.default_code
            """

    _order = 'default_code'
    
    def read(self, cr, uid, ids, fields=None, context=None, load='_classic_read'):
        filtered_fields = exclude_properties(self, fields, ids)
        return super(product_product, self).read(cr, 
                                                 uid, 
                                                 ids, 
                                                 fields=filtered_fields, 
                                                 context=context, 
                                                 load=load)

product_product()

class product_template(osv.osv):
    _inherit = 'product.template'
    
    _columns = {'loc_rack': fields.char('Rack', size=255),
                'loc_row': fields.char('Row', size=255),
                'loc_case': fields.char('Case', size=255),
                'document_id': fields.char('Document ID', size=100)}
    
    _defaults = {'sale_delay': lambda *a: 5,
                 'produce_delay': lambda *a: 0,
                 'cost_method': lambda *a: 'average'}
    
product_template()

class product_category(osv.osv):
    _inherit = "product.category"

    _columns = {
        'description': fields.text('Category Description'),

        'show': fields.boolean('Show on NAFTA'),
        'producer': fields.boolean('Producer'),
        'net_cost': fields.boolean('Net Cost'),

        'harmonized_code': fields.char('Harmonized Code',size=100),
        'ship_description': fields.char('Ship Description',size=100),
        'preference_criterion': fields.char('Preference Criterion',size=100),
        'origin_country_id': fields.many2one('res.country', 
                                             'Country of Origin', 
                                             change_default=True),
    }

    _defaults = {
        'show': lambda *a: True,
        'producer': lambda *a: False,
        'net_cost': lambda *a: True,
    }

product_category()
