# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution	
#    Zaber Technologies Inc. All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
import datetime

from osv import osv
from report import report_sxw
import pooler
import re

def fix_zero_rounding( x ):
    if abs(x) < 1.0e-11: return 0
    return x


class AccountStatement(report_sxw.rml_parse):

    # Setting debug to True will cause the code to generate
    # a file /tmp/account_statement.txt that will list all
    # move_lines and invoices that were used in calculation
    debug = False

    def __init__(self, cr, uid, name, context):
        super(AccountStatement, self).__init__(cr, uid, name, context=context)

        # Find the default company address
        company = self.localcontext['company']
        addresses = company.partner_id.address
        addr_preferences = {
            'invoice': 1,
            'default': 2,
            'delivery': 3,
            'contact': 4,
            'other': 5,
        }
        address = sorted(
                      addresses,
                      key=lambda a:addr_preferences.get(a.type,'10') 
                  )[0]

        # Load up the system with values our rml expects
        self.localcontext.update( {
            'time' : time,
            'adr_get' : self._adr_get,
            'account_statement' : self._account_statement,
            'tel_get' : self._tel_get,
            'invoice_type': {
                                'out_invoice': 'Invoice',
                                'out_refund': 'Credit Memo',
                                'in_invoice': 'Invoice',
                                'in_refund': 'Credit Memo',
                            },
            'company_address': address,
        })

        self.context = context

    def set_context(self, objects, data, ids, report_type = None):
        ''' To avoid having to use repeatIn to set the invoice lines, we want to 
            precalculate the lines. Unfortunately, unless we hook into the set_context,
            we don't actually know what partner is being fetched. So we do the work here
            to figure it out.
        '''

        if 'form' in data:
            number_paid_invoices = data['form']['number_paid_invoices']
            new_ids = [data['form']['partner_id']]
        else:
            number_paid_invoices = 10000
            new_ids = ids

        res_partner = pooler.get_pool(self.cr.dbname).get('res.partner')
        objects = res_partner.browse(self.cr,
                                     self.uid,
                                     new_ids,
                                     context=self.context.copy())
        partner_rec = objects[0]
        statement = self._account_statement(partner_rec,number_paid_invoices)
        self.localcontext.update({'statement':statement})

        return super(AccountStatement, self).set_context(objects, 
                                                         data, 
                                                         new_ids, 
                                                         report_type)

    def _adr_get(self, partner, type):
        res_partner = pooler.get_pool(self.cr.dbname).get('res.partner')
        res_partner_address = pooler.get_pool(self.cr.dbname).get('res.partner.address')
        addresses = res_partner.address_get(self.cr, self.uid, [partner.id], [type])
        adr_id = addresses and addresses[type] or False
        if adr_id:
            result = res_partner_address.read(self.cr, self.uid, [adr_id],context=self.context.copy())
            result[0]['state_name'] = result[0]['state_id'] and result[0]['state_id'][1] or False
            result[0]['country_name'] = result[0]['country_id'] and result[0]['country_id'][1] or False
            return result   
         
        result = {
                  'name': False,
                  'street': False,
                  'city' : False,
                  'zip' : False,
                  'state_name': False,
                  'country_name' : False,
                 }
        return [result]

    def _tel_get(self,partner):
        if not partner:
            return False
        res_partner_address = pooler.get_pool(self.cr.dbname).get('res.partner.address')
        res_partner = pooler.get_pool(self.cr.dbname).get('res.partner')
        addresses = res_partner.address_get(self.cr, self.uid, [partner.id], ['invoice'])
        adr_id = addresses and addresses['invoice'] or False
        if adr_id:
            adr=res_partner_address.read(self.cr, self.uid, [adr_id])[0]
            return adr['phone']
        else:
            return partner.address and partner.address[0].phone or False
        return False

    def _account_statement(self, partner, number_paid_invoices=4 ):

        # Get a list of all the invoices associated with the active user
        inv_obj = pooler.get_pool(self.cr.dbname).get('account.invoice')
        inv_list = inv_obj.search(self.cr, self.uid,
                [
                  ( 'partner_id', '=', partner.id ),
                  ( 'state', 'not in', [ 'draft', 'cancel' ] ),
                ])

        # We don't want anything from the write-offs account
        journal_obj = pooler.get_pool(self.cr.dbname).get('account.journal')
        journal_ids = journal_obj.search(self.cr, 
                                         self.uid, 
                                         [('code', '=', 'WOJ')])
        # If no journal matches code, nothing to exclude.
        writeoffs_journal_id = journal_ids[0] if journal_ids else -1

        # Get the appropriate moves for the payments
        invoice_lines = []

        if self.debug:
            f = open('/tmp/account_statement.txt','w')

        total_calculated_balance = 0
        total_calculated_balance_by_date_range = {
           '0-30': 0.0,
          '31-60': 0.0,
          '61-up': 0.0,
        }
        currencies_log = {}

        for inv_rec in inv_obj.browse(self.cr,self.uid,inv_list,self.context):

            if self.debug:
                f.write( "\n\nInvoice #"+str(inv_rec.id)+" Name:"+str(inv_rec.name)
                          +" Number:"+inv_rec.number
                          +" State["+inv_rec.state+"] amounttotal: "+str(inv_rec.amount_total)+"\n" )
                f.write( "Type: "+inv_rec.type+"\n" )
                f.write( "State: "+inv_rec.state+"\n" )
                f.write( "Currency: "+inv_rec.currency_id.name+"\n" )
                f.write( "Residual: "+str(inv_rec.residual)+"\n" )

            # The report only works well if a single currency is in use. Let's log the currencies
            # we encounter here
            currency_name = inv_rec.currency_id.name
            currencies_log.setdefault(currency_name,[])
            currencies_log[currency_name].append(inv_rec.number)

            # Throw an error if we find that we're dealing with more than one currency
            if len(currencies_log) >= 2:
                error_message = "Report build aborted as invoices span more than one currency:\n\n"
                for currency_name, invoice_ids in currencies_log.iteritems():
                    error_message += "Invoices with "+currency_name+": "+", ".join(invoice_ids)+"\n"
                if self.debug:
                    f.write(error_message)
                return {
                    'error': error_message,
                    'invoice_lines': [],
                    'calculated_balance': 0.0,
                    'calculated_balance_by_daterange': total_calculated_balance_by_date_range,
                    'statement_currency': 'N/A',
                }


            # Some prep...
            is_refund           = inv_rec.type == 'out_refund' or inv_rec.type == 'in_refund'
            is_outgoing_invoice = inv_rec.type == 'out_refund' or inv_rec.type == 'in_invoice'


            calculated_payments = []
            calculated_balance  = 0
            running_total = 0

            if self.debug:
                f.write("%10s %40s %10s %10s %10s %-5s %10s %20s-%-10s %10s\n"%( 
                        "Date",
                        "Move Name",
                        "Move ID",
                        "Ref",
                        "Amt Curr",
                        "CUR",
                        "Acc Type",
                        "Jrnl Name",
                        "Type",
                        "Subtotal",
                      ))

            # Then go through the invoice in sorted order
            sorted_move_lines = sorted(inv_rec.payment_ids,cmp=lambda x,y:cmp(x.date,y.date) or cmp(y.amount_currency,x.amount_currency))

            # Filter out any move we're sure isn't going to 
            # show up on the account statement
            filtered_sorted_move_lines = []
            for move_line in sorted_move_lines:
                if move_line.journal_id.id != writeoffs_journal_id:
                    filtered_sorted_move_lines.append(move_line)


            for move_line in filtered_sorted_move_lines:

                # If the move line is the same currency as the company, we're not using
                # an alternate currency. thus, we will pull from the credit/debit
                # fields
                line_amount = move_line.amount_currency
                if not move_line.currency_id:
                      line_amount = move_line.credit or move_line.debit

                # Force the correct sign on the amount value
                amount_currency_vec = abs(line_amount)
                if is_outgoing_invoice:
                    amount_currency_vec*= -1
                if move_line.credit: 
                    amount_currency_vec*= -1

                running_total   += amount_currency_vec

                # We only record a "payment" when a move line moves us closer to the
                # 0 value. This complicated logic captures when multiple payments have 
                # been placed. It doesn't quite handle credit memos properly yet.
                if  running_total < 0                                                        \
                    and not ( move_line.invoice and move_line.invoice.type == "out_refund" ) \
                    and move_line.account_id.type == 'receivable'                            \
                    and move_line.journal_id.id != writeoffs_journal_id                      \
                    and running_total < calculated_balance                                   \
                    and move_line.credit:

                          match = re.search('([^;,$|]*)',move_line.name)
                          move_name_simple = match.group(0).strip()
                          calculated_amount = running_total - calculated_balance
                          if is_refund:
                              calculated_amount *= -1
                          calculated_payments.append({
                            'move_line': move_line,
                            'calculated_amount': calculated_amount,
                            'move_name': move_name_simple,
                            'last_entry': False,
                          })
                          calculated_balance = running_total

                if self.debug:
                    f.write("%10s %40s %10i %10s %10.2f %-5s %10s %20s-%-10s %f\n"%( 
                            str(move_line.date),
                            move_line.name,
                            move_line.id,
                            str(move_line.ref),
                            amount_currency_vec,
                            move_line.currency_id.name,
                            move_line.account_id.type,
                            move_line.journal_id.name,
                            move_line.journal_id.type,
                            running_total,
                            ))

            # We know how much is left to pay on this invoice
            # The test for <1.0e-11 is to handle -0.00 (sometimes due to
            # rounding error, calculated_residual becaomes X.Ye-13!)
            calculated_residual = fix_zero_rounding(
                inv_rec._amount_residual("residual",
                                         {},
                                         context=self.context)[inv_rec.id])

            # Mark down the last entry in the payment list so we can
            # show the residual on the same line
            if len(calculated_payments) > 0:
                calculated_payments[-1]['last_entry'] = True

            # We don't have any matched payments for some reason
            # This can happen when a credit memo has been paid
            # directly by wire.
            elif len(filtered_sorted_move_lines) == 1:
                move_line = filtered_sorted_move_lines[0]

                amount_total = inv_rec.amount_total
                if is_refund:
                    amount_total *= -1
                calculated_amount = calculated_residual - amount_total

                match = re.search('([^;,$|]*)',move_line.name)
                move_name_simple = match.group(0).strip()

                calculated_payments.append({
                    'move_line': move_line,
                    'calculated_amount': calculated_amount,
                    'move_name': move_name_simple,
                    'last_entry': True,
                })

            # Figure out which direction this invoice is going. Is it going 
            # out to a customer?  (in which case it's a credit) or is it 
            # coming in from a supplier? (in which case it's a debit)
            residual_direction = 1;
            if is_outgoing_invoice:
                residual_direction = -1

            total_calculated_balance += calculated_residual * residual_direction

            # Is this invoice overdue
            now = datetime.datetime.today()
            is_overdue = False
            if inv_rec.date_due \
                and inv_rec.state == 'open' \
                and calculated_residual > 0 \
                and now.strftime("%Y-%m-%d") > inv_rec.date_due:
                    is_overdue = True 

            # If there is a residual on this invoice, we want to put
            # the amount due in buckets
            if inv_rec.state == 'open' \
                and calculated_residual != 0:
                  inv_date = datetime.datetime.strptime( 
                                      inv_rec.date_invoice or inv_rec.create_date, 
                                      "%Y-%m-%d" 
                                  )
                  amount_age = now - inv_date
                  if amount_age.days < 31:
                      total_calculated_balance_by_date_range[ '0-30'] += calculated_residual * residual_direction
                  elif amount_age.days < 61:
                      total_calculated_balance_by_date_range['31-60'] += calculated_residual * residual_direction
                  else:
                      total_calculated_balance_by_date_range['61-up'] += calculated_residual * residual_direction

            invoice_amount_total = inv_rec.amount_total
            if is_refund:
                invoice_amount_total *= -1
                calculated_residual  *= -1

            invoice_lines.append({
              'invoice':    inv_rec,
              'invoice_amount_total': invoice_amount_total,
              'is_overdue': is_overdue,
              'calculated_payment_lines': calculated_payments,
              'calculated_residual': fix_zero_rounding(calculated_residual),
            })

            if self.debug:
                f.write( "Inferred Residual: "+str(calculated_residual)+"\n" )

        # Were there any invoices?
        if len(currencies_log) == 0:
            return {
                'error': "No invoices found.",
                'invoice_lines': [],
                'calculated_balance': 0.0,
                'calculated_balance_by_daterange': total_calculated_balance_by_date_range,
                'statement_currency': '',
            }


        # Now we sort and flag which invoices we actually want to show
        sorted_invoice_lines = sorted(invoice_lines,
                      cmp=lambda x,y:cmp(y['invoice'].date_invoice,x['invoice'].date_invoice))
        filtered_invoice_lines = []
        for invoice_line in sorted_invoice_lines:
            if invoice_line['calculated_residual']:
                filtered_invoice_lines.insert(0,invoice_line)
                continue
            if number_paid_invoices <= 0:
                continue
            number_paid_invoices -= 1
            filtered_invoice_lines.insert(0,invoice_line)

        return {
            'invoice_lines': filtered_invoice_lines,
            'calculated_balance': total_calculated_balance,
            'calculated_balance_by_daterange': total_calculated_balance_by_date_range,
            'statement_currency': currencies_log.keys()[0], # there should be one key/currency
            'error': False,
        }

report_sxw.report_sxw(
        'report.zerp.account.statement', 
        'res.partner',
        'addons/zerp_account_statement/report/account_statement.rml', 
        parser=AccountStatement)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:

