from osv import fields, osv


class zerp_bom_component_replace(osv.osv_memory):
    """New style wizard for updating a set of BoMs to use a new component
    in place of one being discontinued.

    This is ready to ship but there are several non-critical issues which
    should be addressed.

    TODO: There's no feedback about the status of the action here.
    Implement a popup which says it went well or it didn't.

    TODO: It would be nice to target specific BoMs with this tool.
    Create another view after selecting the target component listing
    available BoMs to choose. Perhaps another many2one widget.

    TODO: Check to make sure the user triggering this view is in group
    admins, otherwise alert and close. Better yet, don't show the view
    at all.
    """

    _name = "zerp.bom_component_replace"
    _description = "ZERP BoM Component Replace"
    _columns = {
        "product_id": fields.many2one(
            'product.product',
            'Target Component Product ID',
            help="The Product ID of the component to be replaced with the" \
                 "currently selected product",
            required=True
        )
    }
    _defaults = {}

    def _bom_rec_with_target_list(self, cr, uid, ids, context=None,
                                  target_product_id=None):
        """Return a list of BoM row ids with the target component to be
        replaced.

        TODO: Remove old and now redundent second browse for bom_res. The
        info is available in bom_ids.

        :param cr:
        :param uid:
        :param ids:
        :param context:
        :param target_product_id:   product_id of product being replaced
        """

        bom_obj = self.pool.get("mrp.bom")
        # Get list of BoM rows with our target product ID
        bom_ids = bom_obj.search(cr, uid,
                                 [("product_id.id", "=", target_product_id)])
        bom_res = bom_obj.browse(cr, uid, bom_ids)

        bom_rec_list = [rec.id for rec in bom_res]
        return bom_rec_list

    def _target_do_replace(self, cr, uid, ids, context=None, bom_rec_list=None,
                           new_product_id=None):
        """Do the work of replacing the target components in all of the
        target BoM rows.

        TODO: Possibly change the bom name as well as the product_id. I'm
        not sure about this as often the bom name is arbitrary and not
        necessarily the same as the name of the component product. A possible
        solution is to include a checkbox to confirm a name change.
        - Brian would prefer to copy the product name to the bom name or remove
        the bom name column all together.

        :param cr:
        :param uid:
        :param ids:
        :param context:
        :param bom_rec_list:
        :param new_product_id:   product_id of replacement product
        """

        if bom_rec_list:
            bom_obj = self.pool.get("mrp.bom")
            res = bom_obj.browse(cr, uid, bom_rec_list)

            # Useful debug info
            #print "_target_do_replace(new_product_id:{} old_product_id:{}
            # bom_rec_list:{" \
            #"})".format(new_product_id, old_product_id, bom_rec_list)

            # FIXME The following loop of single record writes could be
            # done with
            # a one write on a list of rec.ids, but doing so triggers a known
            # bug in OpenERP:
            #     https://bugs.launchpad.net/openobject-addons/+bug/985525
            for rec in res:
                bom_obj.write(cr, uid, [rec.id],
                              {"product_id": new_product_id})


    def fields_view_get(self, cr, uid, view_id=None, view_type='form',
                        context=None, toolbar=False, submenu=False):
        """Update the first label in the view to include the name of the
        product being replaced.
        """
        if context == None:
            context = {}

        result = super(zerp_bom_component_replace, self).fields_view_get(
            cr, uid, view_id, view_type, context, toolbar, submenu)

        # fetch the currently selected product id from context
        old_product_id = context.get('active_id', [])

        # get the name of the currently selected product from the database
        # so we can display it in the view
        prod_obj = self.pool.get("product.product")
        prod_name = prod_obj.browse(cr, uid, [old_product_id])
        prod_name = prod_name[0].product_tmpl_id.name

        # replace the {product} token in the view with the actual product name
        result['arch'] = result['arch'].replace('{product}',
                                                str(prod_name))

        return result


    def action_display_component_view(self, cr, uid, ids, context=None):
        """Open the first view in response to a click of our registered action
        button. Ask for the Product ID of the component replacing the
        currently selected one in all bills of materials.

        :param cr:
        :param uid:
        :param ids:
        :param context:
        """
        if context is None:
            context = {}

        # The following values are validated by the client and the model
        # before we get our hands on them.
        old_product_id = context.get('active_id', [])

        new_product_id = self.browse(cr, uid, ids)[0].product_id.id

        # request a list of mrp.bom row ids which we want to change
        bom_rec_list = self._bom_rec_with_target_list(cr, uid, ids, context,
                                                      old_product_id)

        # An empty list is falsey in Python.
        if bom_rec_list:
            self._target_do_replace(cr, uid, ids, context, bom_rec_list,
                                    new_product_id)

        # Only return what's necessary to close the window
        return {
            'type': 'ir.actions.act_window_close',
        }
