# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution	
#    Zaber Technologies Inc. All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from lxml import etree
import time
from datetime import date,datetime
import math
from report import report_sxw
from osv import osv
from tools import amount_to_text_en
import locale

class Cheque(report_sxw.rml_parse):

    def _add_header( self, rml_dom, header = 1 ):
        ''' This prevents any header from being attached to the
            the cheque report.
        '''
        return True

    def _dollars_str (self, val, full_length):
        ''' Converts a number into a string representation. As this is
            used for the dollar portion of the cheque, we need to also
            include dashes (for extra paranoia's sake)
        '''
        dollars_str = amount_to_text_en.english_number(int(val)).title()
        # ; n2w.to_currency(int(val)*100).title()
        dollars_str += " " + "-" * ( full_length - len(dollars_str) - 1 )
        return dollars_str

    def _cents_str (self, val):
        ''' Returns the number of cents to be on the cheque in XXX/100
            format
        '''
        cents_str = "%02i/100" % (round(val * 100) % 100 );
        return cents_str

    def _amount_str(self, val, full_length=0):
        ''' Returns the amount owed in string format with commas
        '''
        amount_str = locale.format('%0.02f', val, True)
        if full_length:
            amount_str = ("*"*( full_length - len(amount_str) )) + amount_str
        return amount_str

    def __init__(self, cr, uid, name, context):
        super(Cheque, self).__init__(cr, uid, name, context=context)
        self.localcontext.update({
            'time':        time,
            'dollars_str': self._dollars_str,
            'cents_str':   self._cents_str,
            'amount_str':  self._amount_str,
        })
        self.context = context

    def _address_to_str(self, address):
        ''' Turn a res.partner.address record into a simple 
            string address suitable for putting into a report
            layout
        '''
        if not address:
            return None

        address_str = ""

        # For cheques we do not include the recipient name,
        # only the company name
        #if address.title:
        #    address_str = address.title
        #if  address.name:
        #    address_str += address.name + "\n"

        if address.street:
            address_str += address.street + "\n"
        if address.street2:
            address_str += address.street2 + "\n"
        
        city_state = ""
        if address.city:
            city_state += address.city + ", "
        if address.state_id:
            city_state += address.state_id.name
        if city_state:
            address_str += city_state + "\n"

        if address.country_id:
            address_str += address.country_id.name + " "
        if address.zip:
            address_str += address.zip

        return address_str

    def set_context(self, objects, data, ids, report_type = None):
        ''' We want to be able to capture and process parameter/wizard 
            requests so we intercept in this function that has access
            to the data[] values
        '''

        # Double check that all the invoices listed have the same
        # invoice address
        error_message = None
        address_invoice_id = None
        currency_id = None

        for invoice in objects:
            if invoice.state != 'open':
                error_message = "Can't create a cheque for a non-open invoice"
            if invoice.type not in [ 'in_invoice', 'out_refund' ]:
                error_message = "Can only create cheques for incoming invoices and outgoing refunds"
            if not address_invoice_id:
                address_invoice_id = invoice.address_invoice_id.id
                currency_id = invoice.currency_id.id
                continue
            if address_invoice_id != invoice.address_invoice_id.id:
                error_message = "Invoices with different invoice addresses selected"
            if currency_id != invoice.currency_id.id:
                error_message = "Invoices with different currencies selected"

        recipient_name = ""
        recipient_address = ""
        cheque_amount = 0
        cheque_currency = None
        cheque_note_table = [
                            #  ['77636',self._amount_str(70.56),'note 3'],
                            ]

        if not error_message and objects:
            for invoice in objects:
                residual = invoice.residual
                cheque_amount   += residual
                cheque_note_table.append([
                                      invoice.name,
                                      self._amount_str(residual),
                                      ''
                                  ])
            recipient_name = invoice.partner_id.name
            if invoice.partner_id.title:
                recipient_name += " " + invoice.partner_id.title.name
            recipient_address = self._address_to_str(invoice.address_invoice_id)
            cheque_currency  = invoice.currency_id.name

        cheque_date = date.today().strftime("%Y %m %d")
        if data['form'].get('cheque_date'):
            cheque_date = datetime.strptime(data['form']\
                            .get('cheque_date'),"%Y-%m-%d")\
                            .strftime("%Y %m %d")

        cheque_data = {
          'cheque_number':            data['form'].get('cheque_number') or None,
          'cheque_amount':            cheque_amount,
          'cheque_currency':          cheque_currency,
          'cheque_recipient':         recipient_name,
          'cheque_recipient_address': recipient_address,
          'cheque_note':              data['form'].get('cheque_note'),
          'error':                    error_message,
          'cheque_note_table':        cheque_note_table,
          'cheque_date':              cheque_date,
        }
        self.localcontext.update(cheque_data)
        return super(Cheque, self).set_context(objects, data, ids, report_type)

report_sxw.report_sxw(
        'report.zerp.cheque', 
        'account.invoice',
        'addons/zerp_cheques/report/cheque.rml', 
        parser=Cheque)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:

