from tools.translate import _
import wizard
import pooler
import netsvc
import time
import re

class wizard_zerp_disassembly(wizard.interface):

    def _warehouse_locations(self, cr, uid, context=None):
        """ Get the list of locations ids associated with the company/shop/warehouse
        """

        pool = pooler.get_pool(cr.dbname)

        location_ids = []
        wids = pool.get('stock.warehouse').search(cr, uid, [], context=context)
        for w in pool.get('stock.warehouse').browse(cr, uid, wids, context=context):
            location_ids.append(w.lot_stock_id.id)

        child_location_ids = pool.get('stock.location').search(cr, uid, [
                                        ('location_id', 'child_of', location_ids)
                                      ])
        location_ids = child_location_ids or location_ids

        return location_ids

    def _get_product_packing_sources(self, cr, uid, product_id, context=None):
        """ Gets the list of packing lists that are available (to steal
            from)
        """
        pool = pooler.get_pool(cr.dbname)
        move_obj = pool.get('stock.move')
        move_ids = move_obj.search(cr,uid,[
                    ('state','=','assigned'),
                    ('product_id','=',product_id),
                ],context=context)
        packing_quantities = {}
        for move in move_obj.browse(cr,uid,move_ids,context=context):
            if not (move.picking_id and move.picking_id.id):
                continue
            if not move.picking_id.id in packing_quantities:
                packing_quantities[move.picking_id.id] = [move.picking_id.name,0]
            packing_quantities[move.picking_id.id][1] += move.product_qty

        return packing_quantities

    def __get_product_location_sources( self, cr, uid, product_id, direction, states=['done'], context=None ):
        """ Try and find the quantities of products based upon the
            warehouse locations
        """
        location_ids = self._warehouse_locations(cr,uid,context)

        location_field = "location_dest_id" if direction == "in" \
                                            else "location_id"

        states_str = ','.join(map(lambda s: "'%s'" % s, states))

        search_query = """
          select 
              %s, sum(product_qty)
          from
              stock_move
          where
              %s in(%s)
              and state in(%s)
              and product_id = %s
          group by 
              %s
        """ % (
                location_field, 
                location_field,
                ",".join(map(str, location_ids)) ,
                states_str,
                product_id,
                location_field
              )

        cr.execute(search_query)
        location_counts = cr.fetchall()
        results = {}
        for location_count in location_counts:
            results[location_count[0]] = location_count[1]

        return results

    def part_sources_data(self,cr,uid,product_id,context):
        """ Figure out where a product can be ganked from. 
            Pretty much crawls through the inventory and the 
            packing lists to see where products are and 
            where the products are allocated.
        """

        pool = pooler.get_pool(cr.dbname)

        source_choices = []
        select_options = []

        # Are there any items that are unallocated that we can grab 
        # the products from?
        location_sources = self._get_product_location_sources(
                                        cr,uid,product_id,
                                        ['done'],
                                        ['done','confirmed','assigned'],
                                        context)
        if location_sources:
            loc_obj = pool.get('stock.location')
            for loc in loc_obj.browse(cr,uid,location_sources.keys()):
                field_name = "stock_location_"+str(loc.id)
                select_options.append((
                    field_name, 
                    str(location_sources[loc.id])+" in "+loc.name
                ))

        # Are there packing lists that can be canceled to acquire the product?
        packing_sources = self._get_product_packing_sources(
                                        cr,uid,product_id,context
                                    )
        if packing_sources:
            for packing_id,packing in packing_sources.iteritems():
                field_name = "stock_picking_"+str(packing_id)
                select_options.append((
                    field_name,
                    str(packing[1])+" in "+str(packing[0])
                ))

        # Are there any prodution orders that we need to cancel to be able
        # to grab a product?
        # We make one assumption, any move marked as assigned has a product
        # allocated to it. So we can potentially grab it from them.
        production_sources = self._get_product_production_sources(
                                        cr,uid,product_id,context
                                    )
        if production_sources:
            for move_id,move_data in production_sources.iteritems():
                field_name = "stock_production_move_"+str(move_id)
                production_names = ",".join(move_data['production_names'])
                select_options.append((
                    field_name,
                    str(move_data['product_qty'])+" in "+production_names
                ))

        return {
            'select_options':     select_options,
            'packing_sources':    packing_sources,
            'production_sources': production_sources,
            'location_sources':   location_sources,
        }

    def _check_bom_state(self,cr,uid,product_id,context=None):
        """ Returns None if the product cannot be disassembled
            in any way
        """

        pool     = pooler.get_pool(cr.dbname)
        bom_obj  = pool.get('mrp.bom')
        base_bom_id = bom_obj._bom_find(cr, uid, product_id, None)
        if not base_bom_id:
            raise wizard.except_wizard(
                _('Error'),
                _('No BoM for product. Thus, nothing to disassemble.'))

        # Load the first descendant members. We're loading early
        # so we can check if this part can even be disassembled
        (bom_entries,res2) = bom_obj._bom_explode(
                                cr,
                                uid,
                                bom_obj.browse(cr,uid,base_bom_id,context),
                                1,
                                [])

        if not bom_entries:
            raise wizard.except_wizard(_('Error'), _('BoM has no components!'))

        return bom_entries

    def _create_disassembly_moves(self,cr,uid, location_id, product_qty, 
                reason, bom_entries, context=None):
        """ Do the work of disassembling the product. Using the production location
            id of the parent product, move the components back into the stock.
        """

        pool = pooler.get_pool(cr.dbname)
        move_obj = pool.get('stock.move')

        # * Create disassembly moves
        imd_obj  = pool.get('ir.model.data')
        imd_id   = imd_obj._get_id(cr, uid, 'stock', 'stock_location_stock')
        stock_loc_id = imd_obj.browse(cr, uid, imd_id, context=context).res_id

        move_ids_outof_disassembly = []
        for bom_entry in bom_entries:
            move_rec = {
                'name':             "DIS:"+reason,
                'product_id':       bom_entry['product_id'],
                'product_uom':      bom_entry['product_uom'],
                'date_planned':     time.strftime('%Y-%m-%d'),
                'state':            'waiting',
                'product_qty':      bom_entry['product_qty'] * product_qty,
                'location_id':      location_id,
                'location_dest_id': stock_loc_id,
            }
            move_id = move_obj.create(cr,uid,move_rec)
            move_ids_outof_disassembly.append(move_id)

        # And then we can move stuff out of disassembly where the automatic moves
        # will shift them into their proper locations in stock.
        move_obj.action_confirm(cr,uid,move_ids_outof_disassembly,context=context)
        move_obj.check_assign(cr,uid,move_ids_outof_disassembly,context=context)
        move_obj.action_done(cr,uid,move_ids_outof_disassembly,context=context)


    def _disassemble_from_packing(self, cr, uid, product_id, picking_id, reason="DISASSEMBLY", context=None):
        """ This will take a product id and a packing list id and disassemble 
            just the product. This does the heavy lifting so that the scheduler
            and sales orders don't get messed up.

            Here's how it works.

            * Create a move taking the finished product from production to stores 
              with a very early date. The date ensures the moved product won't be
              grabbed by the mrp scheduler

            * Cancel the customer's packing list. 

            * Cancel the customer's sales order

            * Check availablity on the move created in #1. This will assign
              the products to that move.

            * Reconfirm the customer's sales order. This will cause the
              mrp scheduler to create a new set of moves and procurements for
              the customer. (that doesn't grab the products we want to 
              disassemble)

            * Create moves from production into stock. The moves will need to be
               confirmed and forced.

        """

        pool = pooler.get_pool(cr.dbname)
        bom_obj = pool.get('mrp.bom')
        move_obj = pool.get('stock.move')
        pick_obj = pool.get('stock.picking')
        prod_obj = pool.get('product.product')

        # Sanity Checks
        bom_entries = self._check_bom_state(cr,uid,product_id,context)

        pick_rec = pick_obj.browse(cr,uid,picking_id,context=context)
        if not pick_rec:
            raise wizard.except_wizard(_('Error'), _('Could not find picking record!'))

        prod_rec = prod_obj.browse(cr,uid,product_id,context=context)

        # * Create reverse moves from prod to stores
        move_ids = move_obj.search(cr,uid,[
                    ('state','=','assigned'),
                    ('product_id','=',product_id),
                    ('picking_id','=',picking_id)
                ],context=context)

        capture_move_ids = []
        for move in move_obj.browse(cr,uid,move_ids,context=context):
            move_rec = {
                'name': 'DIS:' + pick_rec.name,
                'product_id': product_id,
                'product_uom': move.product_uom.id,
                'date_planned': '2000-01-01',
                'state': 'waiting',
                'product_qty': move.product_qty,
                'location_id': move.location_id.id,
                'location_dest_id': prod_rec.property_stock_production.id,
            }

            move_id = move_obj.create(cr,uid,move_rec)
            capture_move_ids.append(move_id)

        # * Cancel customer's packing list
        wf_service = netsvc.LocalService("workflow")
        wf_service.trg_validate(uid,'stock.picking',picking_id,'button_cancel',cr)

        # * Cancel customer's sales order if one is associated
        sale_rec = pick_rec.sale_id
        if sale_rec:
            sale_rec.action_cancel()
            wf_service.trg_validate(uid,'sale.order',sale_rec.id,'cancel',cr)
            sale_rec.action_cancel_draft()

        # * Check availablity on the stock.move. This grabs the, now,
        #   unallocated device to go back to stocks
        move_obj.action_confirm(cr,uid,capture_move_ids)
        move_obj.check_assign(cr,uid,capture_move_ids)
        move_obj.action_done(cr,uid,capture_move_ids)

        # * Recreate Customer's sales order if required
        if sale_rec:
            wf_service.trg_validate(uid,'sale.order',sale_rec.id,'order_confirm',cr)

        # * Create disassembly moves
        for move in move_obj.browse(cr,uid,capture_move_ids,context=context):
            self._create_disassembly_moves(cr,uid,
                  prod_rec.property_stock_production.id,
                  move.product_qty,
                  reason,
                  bom_entries,
                  context)

    def _disassemble_from_stock(self, cr, uid, product_id, location_id, product_qty, reason="DISASSEMBLY", context=None):
        """ Disassemble product_qty items of product product_id from 
            the location_id
        """

        if product_qty <= 0:
            raise wizard.except_wizard(_('Error'), _('Please include the number of items to disassemble.'))

        location_sources = self._get_product_location_sources(
                                        cr,uid,product_id,
                                        ['done','waiting','confirmed','assigned'],
                                        ['done','waiting','confirmed','assigned'],
                                        context)


        # Sanity
        if location_id not in location_sources:
            raise wizard.except_wizard(_('Error'), _("Disassembly request location doesn't have any items."))
        if location_sources[location_id] < product_qty:
            raise wizard.except_wizard(_('Error'), _("Disassembly request has asked more items than exist in location."))

        pool     = pooler.get_pool(cr.dbname)
        bom_obj  = pool.get('mrp.bom')

        bom_entries = self._check_bom_state(cr,uid,product_id,context)

        prod_obj = pool.get('product.product')
        move_obj = pool.get('stock.move')
        loc_obj  = pool.get('stock.location')

        prod_rec = prod_obj.browse(cr,uid,product_id,context=context)

        # Create a move from that location into production
        move_rec = {
            'name':             reason,
            'product_id':       product_id,
            'product_uom':      prod_rec.uom_id.id,
            'date_planned':     time.strftime('%Y-%m-%d'),
            'state':            'waiting',
            'product_qty':      product_qty,
            'location_id':      location_id,
            'location_dest_id': prod_rec.property_stock_production.id,
        }
        move_id = move_obj.create(cr,uid,move_rec)

        # Check availablity on the stock.move. This grabs the
        # device to go into production
        move_obj.action_confirm(cr,uid,[move_id])
        move_obj.check_assign(cr,uid,[move_id])
        move_obj.action_done(cr,uid,[move_id])

        # Create moves of components parts from production into disassembly location
        # then into stock
        self._create_disassembly_moves(cr,uid,
              prod_rec.property_stock_production.id,
              product_qty,
              reason,
              bom_entries,
              context)

        return

    def _disassemble_from_production_order_move (self, cr, uid, product_id, move_id, reason, context=None):
        """ Takes a move associated with a production order and 
            cancels it to free up a product.

            We do this in the following steps.

            * Cancel the production order
            * Which frees up the product in the originating location
              * Then proceed like a standard stock disassembly
            
        """

        # Figure out what move and production order(s) we're playing with
        production_sources = self._get_product_production_sources(cr,uid,product_id,context)
        if move_id not in production_sources:
            raise wizard.except_wizard(_('Error'), _('Has the production order already been cancelled or completed? Can\'t find the link to it.'))
        move_data = production_sources[move_id]

        # Sanity check on the product
        pool     = pooler.get_pool(cr.dbname)
        bom_obj  = pool.get('mrp.bom')

        bom_entries = self._check_bom_state(cr,uid,product_id,context)

        prod_obj = pool.get('product.product')
        move_obj = pool.get('stock.move')
        move_rec = move_obj.browse(cr,uid,move_id)

        # Load and cancel the production order
        mrp_obj   = pool.get('mrp.production')

        wf_service = netsvc.LocalService("workflow")

        # Capture where the original set of moves went
        initial_move_ids = move_obj.search(cr,uid,[
                              ('move_dest_id','=',move_id)
                            ])

        for production_id in move_data['production_ids']:
            # mrp_obj.action_cancel(cr,uid,[production_id])
            wf_service.trg_validate(uid, 'mrp.production', production_id,
                    'button_cancel', cr)

        # How many of the product ids are we trying to disassemble?
        product_qty = move_rec.product_qty

        # Create a procurement for the products that were just
        # freed up
        imd_obj  = pool.get('ir.model.data')
        imd_id   = imd_obj._get_id(cr, uid, 'stock', 'stock_location_stock')
        stock_loc_id = imd_obj.browse(cr, uid, imd_id, context=context).res_id
        dis_loc_id = move_rec.product_id.product_tmpl_id.property_stock_production.id
        proc_obj = pool.get('mrp.procurement')

        move_id = move_obj.create(cr, uid, {
            'name':'DIS:'+reason,
            'product_id': product_id,
            'product_qty': product_qty,
            'product_uom': move_rec.product_uom.id,
            'product_uos_qty': move_rec.product_uos and move_rec.product_uos_qty or False,
            'product_uos': move_rec.product_uos and move_rec.product_uos.id or False,
            'date_planned': time.strftime('%Y-%m-%d'),
            'location_id': stock_loc_id,
            'location_dest_id': dis_loc_id,
            'state': 'waiting',
        })
        proc_id = proc_obj.create(cr, uid, {
            'name': reason,
            'origin': move_rec.name+"-DIS",
            'date_planned': time.strftime('%Y-%m-%d %H:%M:%S'),
            'product_id': product_id,
            'product_qty': product_qty,
            'product_uom': move_rec.product_uom and move_rec.product_uom.id or False,
            'product_uos_qty': move_rec.product_uos_qty,
            'product_uos': move_rec.product_uos and move_rec.product_uos.id or False,
            'location_id': stock_loc_id,
            'procure_method':'make_to_stock',
            'move_id': move_id,
        })
        wf_service = netsvc.LocalService("workflow")
        wf_service.trg_validate(uid, 'mrp.procurement', proc_id, 'button_confirm', cr)
        move_obj.action_done(cr,uid,[move_id])
        wf_service.trg_validate(uid, 'mrp.procurement', proc_id, 'button_restart', cr)

        self._create_disassembly_moves(cr,uid,dis_loc_id,product_qty,reason,bom_entries,context)

        return

    def _disassemble(self, cr, uid, data, context=None):
        """ Now do the work of disassembling a product into its 
            constituent pieces. 
        """

        form_data = data['form']
        match_obj = re.search( "(stock_location|stock_picking|stock_production_move)_(\d+)", 
                               form_data['source_choice'] )
        if not match_obj:
            raise wizard.except_wizard(_('Error'), _('Unrecognized option chosen.'))

        # Are we going backwards from a ready picking order?
        product_id = int(data['id'])
        if match_obj.group(1) == 'stock_picking':
            picking_id = int(match_obj.group(2))
            self._disassemble_from_packing(cr,uid,product_id,
                    picking_id,form_data['disassembly_note'],context)

        # Nope, are we pulling the product out of stock and 
        # disassembling it?
        elif match_obj.group(1) == 'stock_location':
            location_id = int(match_obj.group(2))
            self._disassemble_from_stock(cr,uid,product_id,location_id,
                    float(form_data['stock_qty']),form_data['disassembly_note'],context)

        # Are we trying to pull the product from a production
        # order?
        elif match_obj.group(1) == 'stock_production_move':
           move_id = int(match_obj.group(2))
           self._disassemble_from_production_order_move(cr,uid,product_id,
                    move_id,form_data['disassembly_note'],context)

        return {}

    def _get_product_location_sources( self, cr, uid, product_id, states_in=['done'], states_out=['done'], context=None ):
        location_ids = self._warehouse_locations(cr,uid,context)

        # Find out where all the items are incoming
        res = self.__get_product_location_sources(cr,uid,product_id,'in',
                          states_in,context)
        res2 = self.__get_product_location_sources(cr,uid,product_id,'out',
                          states_out,context)
        location_qty = {}
        for location_id in location_ids:
            qty =  res.get(location_id,0.0) - res2.get(location_id,0.0)
            if qty <= 0:
                continue
            location_qty[location_id] = qty
        return location_qty

    def _get_product_production_sources(self,cr,uid,product_id,context=None):
        """ Find all production sources that we can steal products from
        """
        location_ids = self._warehouse_locations(cr,uid,context)
        location_ids_str = ','.join(map(str, location_ids))
        link_to_prod_order_sql = """
          select
              sm.id, 
              sm.name,
              sm.product_qty,
              mpmi.production_id,
              mp.name
          from
              stock_move sm,
              mrp_production_move_ids mpmi,
              mrp_production mp
          where
              sm.product_id = %d 
              and sm.state = 'assigned'
              and (
                  sm.location_id in(%s)
              )
              and sm.location_id <> sm.location_dest_id
              and sm.id = mpmi.move_id
              and mp.id = mpmi.production_id
        """ % (product_id, location_ids_str)

        cr.execute(link_to_prod_order_sql)
        production_counts = cr.fetchall()
        results = {}
        for production_count in production_counts:
            if production_count[0] not in results:
                results[production_count[0]] = {
                        'name': production_count[1],
                        'product_qty': production_count[2],
                        'production_ids': [],
                        'production_names': [] 
                    }
            results[production_count[0]]['production_ids'].append(production_count[3])
            results[production_count[0]]['production_names'].append(production_count[4])

        return results

    def execute_cr(self, cr, uid, data, state='init', context=None):
        """ Intercept requests before the handling routines of a 
            wizard so we can create a dynamic interface
        """

        # Build the dynamic interface if required
        state_def = self.states[state]
        if 'init' in state_def and callable(state_def['init']):
            init_def = state_def['init'](self,cr,uid,data,state,context)
            if 'arch' in init_def:
                self.states[state]['result']['arch'] = init_def['arch'] 
            if 'fields' in init_def:
                self.states[state]['result']['fields'] = init_def['fields'] 
            if 'state' in init_def:
                self.states[state]['result']['state'] = init_def['state'] 

        return super(wizard_zerp_disassembly,self).execute_cr(cr, uid, data, state, context)


    def _prep_init(self,cr,uid,data,state,context):
        """ Create a dynamic form that preps the interface with a list
            of possible records
        """

        source_choices = self.part_sources_data(cr,uid,data['id'],context)
        parameter_fields = {
            'disassembly_note': {
              'string': 'Note', 
              'help': 'Why is this being disassembled?',
              'type': 'char', 
              'required': True,
              'size': 64,
            },
            'stock_qty': {
              'string': 'Stock Quantity', 
              'help': 'The number to be disassembled from stock. Unused when picking list has been chosen.',
              'type': 'char', 
              'required': False,
            },
            'source_choice': {
                'string': 'Disassemble From',
                'type': 'selection',
                'selection': source_choices['select_options'],
                'required': True,
            }
        }

        parameter_states = [('end','Cancel'),('disassemble','Disassemble')]

        parameter_form = '''<?xml version="1.0"?>'''


        # Are there any products floating around that we can grab?
        if source_choices['select_options']:
            parameter_form += '''
            <form string="Disassemble products into subcomponents" colspan="4">
                <label string="This wizard will allow a product to be disassembled into its constituent components. Not a full disassembly but first level of the BoM only." colspan="4" align="0.0" />
                <field name="source_choice" colspan="4"/>'''
            if source_choices['location_sources']:
                parameter_form += ''' <field name="stock_qty" colspan="4"/>'''

            parameter_form += '''
                <field name="disassembly_note" colspan="4"/>
            </form>
            '''
        # No products floating around, give error
        else:
            parameter_form += '''
            <form string="Disassemble products into subcomponents" colspan="4">
              <label string="Sorry, couldn't find any the product anywhere in stock." colspan="4" align="0.0" />
            </form>
            '''
            parameter_states = [('end','Cancel')]

        return {
            'arch':   parameter_form,
            'fields': parameter_fields,
            'state':  parameter_states,
        }

    def _get_default(self, cr, uid, data, context=None): 
        self._check_bom_state(cr,uid,data['id'],context)
        return {
            'disassembly_quantity': 1,
        } 

    states = {
        'init': {
            'actions': [_get_default],
            'init': _prep_init,
            'result': {
                        'type':'form', 
                        'state':[('end','Cancel'),('disassemble','Disassemble')]
                      }
        },
        'disassemble': {
            'actions': [_disassemble],
            'result': {'type':'state', 'state':'end'}
        }
    }

wizard_zerp_disassembly('zerp.disassembly')

