from osv import osv, fields
from openerp import pooler
import netsvc
import threading


class mrp_production(osv.osv):
    """ Override mrp_production to add a new state: 'Calculating' which
        performs all product_produce operations on a background thread
        and changes the 'Produce' button to an unclickable 'Calculating'
    """
    _inherit = "mrp.production"

    _columns = {
        'state': fields.selection(
            [('draft','New'),
             ('picking_except', 'Picking Exception'),
             ('confirmed','Waiting Goods'),
             ('ready','Ready to Produce'),
             ('in_production','Production Started'),
             ('in_calculation','Final Calculation'),
             ('cancel','Cancelled'),
             ('done','Done')],
            'State',
            readonly=True,
            help="When the production order is created the state is set to "
                 "'Draft'.\n If the order is confirmed the state is set to "
                 "'Waiting Goods'.\n If any exceptions are there, the state "
                 "is set to 'Picking Exception'.\nIf the stock is available "
                 "then the state is set to 'Ready to Produce'.\n When the "
                 "production gets started then the state is set to 'In "
                 "Production'.\n When the production is over, the state is "
                 "set to 'Done'.'"
        )
    }

    def dummy_produce_all(self, cr, uid, ids, context=None):
        """ A method spun off on a new thread which grabs a new database
            cursor and then executes action_produce_all from production_qty.
            This does everything involved in transitioning an in_production
            manuf-order to the done state in the background so people in
            production don't have to wait for it.

            TODO: Maybe put extra concurrent execution testing in this bit
            just because.
        """
        # grab a new db cursor as we've lost the previous one when we spawned
        # this new thread. since this is a new cursor, we are responsible for
        # commits and finally closing the connection
        cr = pooler.get_db(cr.dbname).cursor()
        mrp_prod = pooler.get_pool(cr.dbname).get('mrp.production')

        # run the calculacions by calling action_produce_all from
        # production_qty. this will set the state to done when execution
        # completes. If this fails, rollback the cursor, otherwise commit.
        try:
            self.action_produce_all(cr, uid, ids, context)
        except Exception:
            # roll back all of the damage and change the state back to
            # in_production. This is necessary because we commited the state
            # flag change in action_produce_calculate.
            cr.rollback()
            mrp_prod.write(cr, uid, ids, {'state': 'in_production'})
        cr.commit()
        cr.close()

        return True

    def action_produce_calculate(self, cr, uid, ids, context=None):
        """ Short circuit the produce button to fire our own action_produce
            method which simply does all of that stuff in the background
            instead of on the main thread. This frees the client up to do
            other more interesting things than watch it frozed.

            There is some concurrency checking in this to make sure it can
            only be fired once from a client, and only one client can fire it.
            This is accomplished by creating a new state in the production
            workflow called in_calculation which only lives for when the
            action_produce is doing it's thing.
        """
        # Since this button can be accessed by many clients concurrently under
        # very special and unlikely circumstances, let's make sure we're
        # working on a manuf-order that's currently in_production.
        state = self.browse(cr, uid, ids, context)[0].state

        if state == 'in_production':
            # changing the state to in_calculation so nothing can mess up this
            # manufacturing order while we're 'thinking'
            self.write(cr, uid, ids, {'state': 'in_calculation'})

            # committing this cursor since we need to see this state change
            # in other cursors. Usually this is a no-no since openobject
            # server handles commits for all rpc calls.
            cr.commit()

            # trigger the transition in our workflow to hide the produce
            # button which will only show on in_production state, and not
            # in_calculation.
            wf_service = netsvc.LocalService("workflow")
            wf_service.trg_validate(
                uid,
                'mrp.production',
                ids[0],
                'button_produce_calculate',
                cr)

            # finally it's safe to spin off the calculation thread
            thr = threading.Thread(
                target=self.dummy_produce_all,
                args=(cr, uid, ids, context))
            thr.start()

        return True
