# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution    
#    Copyright (C) 2011 Zaber Technologies Inc.
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from osv import osv, fields
from tools import config
from SimpleCalc import evalExpression
import pooler
import csv
import string
import re
import json

##############################################################################
# Request/Receive the CSV data
##############################################################################
class kits_data (osv.osv):
    _name = 'zerp.product.kits.data'
    _rec_name = 'id'
    _columns = {
        'csv_data': fields.text('CSV Data'),
        'state': fields.selection([
          ('draft','Draft State'),
          ('check_csv','Check CSV Sanity State'),
          ('debug_csv','Debug CSV State'),
          ('check_lines','Check CSV Line Sanity State'),
          ('debug_lines','Debug CSV Lines State'),
          ('review','Review State'),
          ('confirmed','Confirmed State')],
          'State'),
        'entry_ids': fields.one2many('zerp.product.kits.entries','kit_data_id','BoM Entries'),
        'errors': fields.text('Error'),
        'next_state': fields.char("Next State Hint", size=64),
        'line_ids': fields.one2many('zerp.product.kits.lines','kit_data_id','BoM Entries'),
    }
    _defaults = {
        'state': lambda *a: 'draft',
        'csv_data': lambda *a: '''# Please search iWiki for the keyword 'zerp_product_kits' to find documentation on this screen
# Hint: The minimal CSV fields to be populated are
''' + "# NewCode\tSource\tSourceRev\tNewName\tNewRev\tNewCategory\tNewBOM\tPrice\tRouting"

    }

    # The following fields must be found in the CSV file
    expected_headers = [
        "newcode",
        "source",
        "sourcerev",
        "newname",
        "newrev",
        "newcategory",
        "newbom",
        "price",
        "routing",
    ]


    ###################################################################
    ## Little functions (Parsers)
    ###################################################################


    def bom_str_parse ( self, bom_str ):
        '''
        Return an array of products parsed out of 
        a bom_string
        '''

        bom_array = []

        bom_entries = string.split(bom_str, "|")
        for bom_entry_str in bom_entries:

            # First parse the elements into variables
            m = re.search( "(.+?)(:r(\d+))?(:(\d+))?$", bom_entry_str )
            pn = m.group(1).strip()
            quantity = m.group(5)
            if quantity == None:
                quantity = "1"
            revision = m.group(3)

            # Create the new record
            bom_array.append({
                'pn': pn,
                'quantity': quantity.strip(),
                'revision': revision,
            })

        return bom_array

    def price_parse_calculate(self, price_eqn, parent_list_price=None):
        '''
        Returns the error message with the price equation if there's an error
        '''

        # Make sure we're not empty
        if price_eqn == None or price_eqn == '':
            raise Exception("No Price Defined")

        # Explicit price?
        if re.match( "^\d+(\.\d+)?$", price_eqn ):
            return price_eqn

        # If the parent has no price, then we want to make that clear
        if parent_list_price == None or parent_list_price == 0:
            raise Exception("No Parent Price")

        # If the price_eqn is "inherit" we just copy the price over and
        # we're done
        if price_eqn == 'inherit':
            return parent_list_price

        # If we're here, we have a much more complicated equation to
        # fullfill
        myvars = {
            'x': parent_list_price,
            'x2': parent_list_price*2,
            'x3': parent_list_price*3,
            'x4': parent_list_price*4,
        }

        try: 
            return evalExpression( price_eqn, myvars )
        except Exception:
            raise Exception("Equation Parse Error")


    ###################################################################
    ## Buttons
    ###################################################################


    def button_draft (self, cr, uid, ids):
        self.write(cr, uid, ids, { 'state' : 'draft' })
        return True

    def button_check_csv (self, cr, uid, ids):
        '''
        Check to see if the CSV isn't completed broken and
        parsing can't even start
        '''
        for data in self.browse(cr,uid,ids):
            self.check_csv_data(cr,uid,data)
        self.write(cr, uid, ids, { 'state' : 'check_csv' })

        # We do one sneaky thing here: Clear out the old CSV load 
        # attempts
        self.delete_confirmed(cr,uid)

        return True

    def button_debug_csv (self, cr, uid, ids):
        self.write(cr, uid, ids, { 'state' : 'debug_csv' })
        return True

    def button_check_lines (self, cr, uid, ids):
        '''
        If we can parse the CSV, this will check for brokeness
        on individual lines
        '''
        for data in self.browse(cr,uid,ids):
            self.check_lines_data(cr,uid,data)
        self.write(cr, uid, ids, { 'state' : 'check_csv' })
        return True

    def button_debug_lines (self, cr, uid, ids):
        self.write(cr, uid, ids, { 'state' : 'debug_lines' })
        return True

    def button_review (self, cr, uid, ids):
        '''
        This function is invoked when there is a CSV file that has
        been updated and validated. What we do here is to parse
        the CSV file into their individual segments then store 
        the successfully parsed entries into zerp.product.kits.entries
        for review by the user (then commit to the database)
        '''
        self.write(cr, uid, ids, { 'state' : 'review' })
        return True

    def button_confirmed (self, cr, uid, ids):
        '''
        Commit the record to the database in its final form
        '''
        for data in self.browse(cr,uid,ids):
            self.confirm_csv_data(cr,uid,data)
        self.write(cr, uid, ids, { 'state' : 'confirmed' })
        return True

    ###################################################################
    ## Actions
    ###################################################################

    def delete_confirmed (self, cr, uid):
        '''
        Remove the old "confirmed" status record. This just makes sure 
        that we don't leave data lying around
        '''
        delete_ids = self.search(cr,uid,[('state','=','confirmed')])
        self.unlink(cr,uid,delete_ids)

    def check_csv_data (self, cr, uid, data):
        '''
        Parse the CSV file into their individual segments then store 
        the successfully parsed entries into zerp.product.kits.entries
        for review by the user (then commit to the database)
        '''
        # Pre-init some vars we'll be used repeatedly
        csv_data    = data.csv_data
        csv_id      = data.id
        entries_obj = self.pool.get('zerp.product.kits.entries')
        lines_obj   = self.pool.get('zerp.product.kits.lines')
        product_obj = self.pool.get('product.product')
        bom_obj     = self.pool.get('zerp.product.kits.bom')

        # Sanity Check: Make sure we have CSV data
        if csv_data == None or csv_data == '' or csv_data == False:
            self.write(cr, uid, [csv_id], { 
                'errors': 'No CSV data found!',
                'next_state': 'debug_csv',
            })
            return csv_id

        # Start getting ready to part (convert text blob 
        # into array of lines)
        csv_lines = string.split(csv_data,"\n")

        # If nothing goes wrong the entry will end up
        # in the 'review' state. If something does go wrong
        # it will end up in either the:
        # 'debug_csv' = csv is too broken to parse
        # 'debug_lines' = only certain lines are broken

        csv_headers = None
        csv_entry_count = 0
        for csv_line in csv_lines:

            # Any line starting with a  '#' is considered a comment
            if (re.match("^\s*#",csv_line)):  
                continue

            csv_entries = csv.reader([csv_line], delimiter='\t')
            csv_entry = csv_entries.next()

            # Don't bother with empty entries
            if (len(csv_entry) == 0):  
                continue

            # If this is the first entry, it's the header
            if (csv_headers == None): 
                csv_headers = map( string.strip, map(string.lower, csv_entry))

                # Check that all the desired fields are in the
                # headers
                csv_lookup = {}
                for header in csv_headers:
                    csv_lookup[header] = 1
                missing_headers = []
                for header in self.expected_headers:
                    if header not in csv_lookup:
                        missing_headers.append(header)
                if len(missing_headers) > 0:
                    error_msg = "Missing header"
                    error_msg += ":" if len(missing_headers) else "s:"
                    self.write(cr,uid,[csv_id],{ 
                        'errors': "Missing header(s): " + (", ".join(missing_headers)),
                        'next_state': 'debug_csv',
                    })
                    return csv_id

                continue

            csv_entry_count += 1

            # Load each line into the csv lines table for the
            # check_line state to review them
            # This route does cause more work for the system, however
            # the resultant code has less duplication in it
            line_id = lines_obj.create(cr,uid,{
                'csv_line': csv_line,
                'errors': '',
                'csv_headers': json.dumps(csv_headers),
                'kit_data_id': csv_id,
            })

        # Sanity Check: Make sure we have CSV data
        if csv_entry_count == 0:
            self.write(cr, uid, [csv_id], { 
                'errors': 'No CSV entries found!',
                'next_state': 'debug_csv',
            } )
            return csv_id

        # Drop out. We're done
        self.write(cr, uid, [csv_id], { 
            'errors': '',
            'next_state': 'check_lines',
        } )
        return csv_id

    def check_lines_data (self, cr, uid, data ):
        '''
        For each CSV upload attempt, check to see if all the lines
        in the file are sane. If not, note down what the problems are
        and let the user resolve the issues
        '''

        # pre-init some vars we'll be used repeatedly
        csv_data    = data.csv_data
        csv_id      = data.id
        entries_obj = self.pool.get('zerp.product.kits.entries')
        lines_obj   = self.pool.get('zerp.product.kits.lines')
        bom_obj     = self.pool.get('zerp.product.kits.bom')

        next_state = 'review'
        for line in data.line_ids:
            csv_headers = json.loads(line.csv_headers)
            csv_entries = csv.reader([line.csv_line], delimiter='\t')
            csv_entry = csv_entries.next()
            line_rec = self.check_line(cr,uid,csv_headers,csv_entry)

            # Note any new errors, move to the next
            if line_rec['errors']:
                lines_obj.write(cr,uid,line.id,{
                    'errors': line_rec['errors'],
                    'product_id': line_rec['product_id'],
                })
                next_state = 'debug_lines'
                continue

            # Now place the records into a target table
            entry_rec = line_rec['entry_rec']
            entry_rec['kit_data_id'] = csv_id
            entry_id = entries_obj.create(cr,uid,entry_rec)

            # And since we now have an entry id, we can tie-in the BoM
            for bom_entry in line_rec['bom']:
                bom_obj.create(cr,uid,{
                    'entry_id': entry_id,
                    'product_id': bom_entry['product_id'],
                    'quantity': bom_entry['quantity'],
                })

            # We no longer need this record
            lines_obj.unlink(cr,uid,line.id)

        # Drop out. We're done
        self.write(cr, uid, [csv_id], { 
            'errors': '',
            'next_state': next_state,
        })

        return csv_id


    def check_line (self, cr, uid, csv_headers, csv_entry ):
        '''
        Check and return a partially populated zerp.product.kits.lines
        record that can be used to update the parent record or create one
        '''

        line_rec = {
            'errors': '',
            'bom': [],
            'entry_rec': {},
            'product_id': None,
        }

        # Convert the csv_entry into a hash
        entry_rec = {}
        for i in range(len(csv_entry)):
            entry_rec[csv_headers[i]] = csv_entry[i]
        line_rec['entry_rec'] = entry_rec

        # Do we have a target name?
        if entry_rec['newcode'] == '':
            line_rec['errors'] = 'Pleave provide a new partnumber'
            return line_rec

        # Does this target name already exist?
        product_obj = self.pool.get('product.product')
        target_name_tests = [('default_code','=',entry_rec['newcode'])]
        if entry_rec['newrev']:
            target_name_tests.append(('revision','=',entry_rec['newrev']))
        product_ids = product_obj.search(cr,uid,target_name_tests)
        if len(product_ids) != 0:
            line_rec['errors'] = 'Partnumber "'+entry_rec['newcode']+'" already exists in database'
            return line_rec

        # Make sure we have a source (aka: parent object) named
        source_name = entry_rec['source']
        del entry_rec['source']
        if source_name == '':
            line_rec['errors'] = 'Require a parent product'
            return line_rec

        # Then make sure that the named source exists!
        source_name_tests = [('default_code','=',source_name)]
        if entry_rec['sourcerev']:
            source_name_tests.append(('revision','=',entry_rec['sourcerev']))
        product_ids = product_obj.search(cr,uid,source_name_tests)
        if len(product_ids) == 0:
            line_rec['errors'] = 'Could not find parent product "'+source_name+'" in database'
            return line_rec
        elif len(product_ids) > 1:
            line_rec['errors'] = 'More than one product "'+source_name+'" in database'
            return line_rec
        line_rec['entry_rec']['product_id'] = product_ids[0]
        line_rec['product_id'] = product_ids[0]

        # Let's see of the defined price makes sense
        product_rec = product_obj.browse(cr,uid,product_ids)[0]
        price_eqn = entry_rec['price']
        try:
            price_eqn = self.price_parse_calculate(price_eqn,product_rec.list_price)
        except Exception as err:
            line_rec['errors'] = 'Price Equation Failed: '+str(err)
            return line_rec
        entry_rec['price'] = float(price_eqn)

        # Check if the new BoM has a routing set up
        if 'routing' not in entry_rec:
            line_rec['errors'] = 'Routing has not been defined'
            return line_rec
        routing_name = entry_rec['routing']
        del entry_rec['routing']
        if routing_name == None or routing_name == '':
            line_rec['errors'] = 'Routing has not been defined'
            return line_rec
        mrp_routing_obj = self.pool.get('mrp.routing')
        routing_ids = mrp_routing_obj.search(cr,uid,[('name','ilike',routing_name)])
        if len(routing_ids) == 0:
            line_rec['errors'] = 'Routing name does not match any known entries'
            return line_rec
        if len(routing_ids) > 1:
            routing_names = []
            for routing in mrp_routing_obj.browse(cr,uid,routing_ids):
                routing_names.append(routing.name)
            line_rec['errors'] = 'Routing name matches more than one entry: '
            line_rec['errors'] += ', '.join(routing_names)
            return line_rec
        entry_rec['routing_id'] = routing_ids[0]

        # Does the new category already exist?
        new_category = entry_rec['newcategory']
        del entry_rec['newcategory']
        if new_category == '' or new_category == None:
            line_rec['errors'] = 'Category has not been defined'
            return line_rec
        if new_category == 'inherit':
            entry_rec['categ_id'] = product_rec.categ_id.id
        else:
            category_obj = self.pool.get('product.category')
            category_ids = category_obj.search(cr,uid,[('name','=',new_category)])
            if len(category_ids) == 0:
                line_rec['errors'] = 'Could not find Category "'+new_category+'"'
                return line_rec
            entry_rec['categ_id'] = category_ids[0]

        # Copy over the newname if it's inherited
        if entry_rec['newname'] == 'inherit':
            entry_rec['newname'] = product_rec.name

        # Now, figure out what the BOM contains and 
        # figure out if the record even exists
        bom_str = entry_rec['newbom']
        del entry_rec['newbom']
        if bom_str == '':
            line_rec['errors'] = 'BoM is required'
            return line_rec

        # check to see if the bom has entries
        bom_array = self.bom_str_parse(bom_str)
        for bom_entry in bom_array:
            component_search_constraints = [('default_code','=',bom_entry['pn'])]
            if bom_entry['revision']:
                component_search_constraints.append(('revision','=',bom_entry['revision']))
                
            product_ids = product_obj.search(cr,uid,component_search_constraints)
            if len(product_ids) == 1:
                bom_entry['product_id'] = product_ids[0]
            elif len(product_ids) > 1:
                line_rec['errors'] = 'Could not find bom item "'+bom_entry['pn']+'" has more than one result!'
                return line_rec
            else:
                line_rec['errors'] = 'Could not find bom item "'+bom_entry['pn']+'" in database'
                return line_rec

        # And since we now have an entry id, we can tie-in the BoM
        for bom_entry in bom_array:
            line_rec['bom'].append({
                'product_id': bom_entry['product_id'],
                'quantity': bom_entry['quantity'],
                'revision': bom_entry['revision'],
            })

        return line_rec

    def confirm_csv_data (self, cr, uid, data):
        '''
        The user has clicked commit. In that case, we want to load our
        target table witih all the information found in the CSV file that
        we've allowed
        '''
        # pre-init some vars we'll be used repeatedly
        product_obj = self.pool.get('product.product')
        mrp_bom_obj = self.pool.get('mrp.bom')
        uom_obj     = self.pool.get('product.uom')
        op_obj      = self.pool.get('stock.warehouse.orderpoint')

        # To create the new kits, we're going to:
        #   1. Generate the product.product entry
        #   2. Generate the BoM for the product.product entry
        for entry in data.entry_ids:

            # Create the product
            product_id = product_obj.copy(cr,uid,entry.product_id.id,{
                              'default_code': entry.newcode,
                              'categ_id': entry.categ_id.id,
                              'list_price': entry.price,
                              'name': entry.newname,
                              'revision': entry.newrev,
                          },{'suppress_copy_suffix': True})
            product_rec = product_obj.browse(cr,uid,[product_id])[0]

            # Create the parent BoM
            parent_bom_id = mrp_bom_obj.create(cr,uid,{
                                  'product_id':  product_id,
                                  'code':        entry.newcode,
                                  'name':        entry.newname,
                                  'routing_id':  entry.routing_id.id,
                                  'product_uom': product_rec.uom_id.id,
                                  'product_qty': 1,
                                  'bom_id':      None,
                              })

            # Attach the BoM
            for bom_entry in entry.bom_ids:
                mrp_bom_obj.create(cr,uid,{
                    'product_id':  bom_entry.product_id.id,
                    'product_qty': bom_entry.quantity,
                    'product_uom': bom_entry.product_id.uom_id.id,
                    'code':        bom_entry.product_id.default_code,
                    'name':        bom_entry.product_id.name,
                    'bom_id':      parent_bom_id,
                })


        return data.id

kits_data()

##############################################################################
# This holds the pool of lines in the CSV data file that are yet unresolved
# (unparsed or broken)
##############################################################################

class kits_lines ( osv.osv ):
    _name = 'zerp.product.kits.lines'
    _rec_name = 'csv_line'

    _columns = {
        'kit_data_id': fields.many2one('zerp.product.kits.data','Base CSV', ondelete='cascade'),
        'csv_line':    fields.text('CSV Line'),
        'product_id':  fields.many2one('product.product','Source Product', ondelete='cascade'),
        'csv_headers': fields.text('CSV Headers'),
        'errors':      fields.char('Error',size=255),
    }
kits_lines()

##############################################################################
# Links the CSV file to a series of Product entries 
##############################################################################

class kits_entries ( osv.osv ):
    _name = 'zerp.product.kits.entries'
    _rec_name = 'newcode'

    def _bom_string(self, cr, uid, ids, field_name, arg, context):
        '''
        Used by the fields.function for bom_string. Converts the 
        one2many relationship into a usable string in the format 
        familar to the user. eg:

          T-LSM1234:3, KT04

        '''
        result = {}

        # Go through each kit entry's bom and create a string representation
        # of it. eg: product1, product2:5
        for kit_entry in self.browse(cr, uid, ids, context=context):
            bom_entry_list = []
            for bom_entry in kit_entry.bom_ids:
                bom_entry_list.append(bom_entry.product_id.default_code)

                # add the :X to create a partnumber:X if we need to include
                # more than one of the partnumber on the BoM
                if bom_entry.quantity > 1:
                    bom_entry_list[-1] += ":"+str(bom_entry.quantity)

            result[kit_entry.id] = "| ".join(bom_entry_list)

        return result

    _columns = {
        'kit_data_id': fields.many2one('zerp.product.kits.data','Base CSV', ondelete='cascade'),
        'product_id':  fields.many2one('product.product','Source Product', ondelete='cascade'),
        'sourcerev':   fields.integer('Source Rev Number'),

        'categ_id':    fields.many2one('product.category','New Category', required=True, change_default=True),

        'newcode':     fields.char('Code', size=64),
        'newrev':      fields.integer('New Rev Number'),
        'newname':     fields.char('New Name', size=128),

        'price':       fields.float('Sale Price', digits=(16, int(config['price_accuracy'])), help="Base price for computing the customer price. Sometimes called the catalog price."),

        'routing_id':  fields.many2one('mrp.routing', 'Routing', help="The list of operations (list of workcenters) to produce the finished product. The routing is mainly used to compute workcenter costs during operations and to plan futur loads on workcenters based on production plannification."),

        'bom_ids':     fields.one2many('zerp.product.kits.bom','entry_id','New BoM'),
        'bom_string':  fields.function(
                                        _bom_string,
                                        type   = 'char',
                                        method = True,
                                        string = "New BoM"
                                    ),
    }
kits_entries()


##############################################################################
# Links the CSV file to a series of BoM entries 
##############################################################################

class kits_bom ( osv.osv ):
    _name = 'zerp.product.kits.bom'
    _rec_name = 'code'
    _columns = {
        'entry_id':   fields.many2one('zerp.product.kits.entries','New Product', ondelete='cascade'),
        'product_id': fields.many2one('product.product','Bom Product', ondelete='cascade'),
        'revision':   fields.integer('Revision Number'),
        'quantity':   fields.float('Quantity'),
    }
kits_bom()

