import datetime

from osv import fields, osv

class account_invoice(osv.osv):
    _inherit = 'account.invoice'
#    
    _columns = {
        'create_date': fields.datetime('Date Created', readonly=True),
    }
    
    def sanity_check(self, cr, uid, context=None):
        self.sanity_check_fast(cr, uid, context)
        return 0

    def sanity_check_fast(self, cr, uid, context):
        ''' A set of fast sanity checks
        '''
        self.residual_when_paid(cr,uid,context)
        self.stuck_supplier_refunds(cr,uid,context)
        self.account_date_period_large_offset(cr,uid,context)
    
    def account_date_period_large_offset(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(
            cr,
            uid,
            context,
            "Account Move with large difference between date and period",
            """\
This is a stub to track incidents where account move lines have been registered to dates that are greater than 60 days out of the associated period. 
""",
            'account_date_period_large_offset')

        cr.execute('''\
        SELECT
                    aml.date as "Date",
                    aml.id,
                    ap.name as "Period",
                    rp.name as "Partner Name",
                    aml.move_id as "Move Line ID",
                    ru.login "Creation User"
        FROM        account_move_line aml
        LEFT JOIN   account_period ap
              ON    aml.period_id = ap.id
        LEFT JOIN   res_partner rp
              ON    aml.partner_id = rp.id
        LEFT JOIN   res_users ru
              ON    aml.create_uid = ru.id
        WHERE
                    abs(ap.date_start - aml.date) > 90 or
                    abs(ap.date_stop  - aml.date) > 90
        ''')
        move_lines = cr.fetchall()
        if not move_lines:
            return # hurrah, no weirdness

        problem_writeup = "The following move lines are 60 days outside of the associated period:\n\n"
        
        for ml in move_lines:
            problem_writeup += (
                "%s (%s) period %s - %s line %s by %s\n" % ml)

        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(cr, 
                                                  uid,
                                                  context,
                                                  case_id,
                                                  problem_writeup)        

    def residual_when_paid(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(
            cr,
            uid,
            context,
            "Residual on a paid invoice",
            """\
This is a stub to track incidents where paid invoices still have a residual. 
""",
            'account_invoice_residual')

        invoice_ids = self.search(cr,
                                  uid,
                                  [('state', '=', 'paid'),
                                   ('residual', '!=', 0)])
        if not invoice_ids: 
            return # okay, no problems

        problem_writeup = "The following paid invoices have a residual:\n\n"
        for invoice in self.browse(cr, uid, invoice_ids):
            problem_writeup += (
                "%s (id %i) on %s for $%.2f with residual $%.2f to %s\n" % (
                    invoice.number,
                    invoice.id,
                    invoice.date_invoice,
                    invoice.amount_total,
                    invoice.residual,
                    invoice.partner_id.name))
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(cr, 
                                                  uid,
                                                  context,
                                                  case_id,
                                                  problem_writeup)

    def stuck_supplier_refunds(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(
            cr,
            uid,
            context,
            "Stuck supplier refunds",
            """\
This is a stub to track incidents where a supplier refund gets left in draft. 
""",
            'account_invoice_stuck_refund')

        now = datetime.datetime.now()
        stale_timestamp = now - datetime.timedelta(60)
        stale_date = stale_timestamp.strftime('%Y-%m-%d')
        
        # Find old supplier refunds, or supplier invoices with negative amount.
        invoice_ids = self.search(cr,
                                  uid,
                                  [('state', '=', 'draft'),
                                   ('create_date', '<', stale_date),
                                   '|',
                                   ('type', '=', 'in_refund'),
                                   '&',
                                   ('type', '=', 'in_invoice'),
                                   ('amount_total', '<', 0.0)])

        if not invoice_ids:
            return # okay, no problems
        
        invoice_types = dict(self._columns['type'].selection)

        problem_writeup = (
            "The following supplier refunds are stuck in draft:\n\n")
        for invoice in self.browse(cr, uid, invoice_ids):
            problem_writeup += (
                "%s (id %i) on %s for $%.2f to %s\n" % (
                    invoice_types[invoice.type],
                    invoice.id,
                    invoice.create_date,
                    invoice.amount_total,
                    invoice.partner_id.name))
#            problem_writeup += (
#                "%s (id %i) on %s for $%.2f with residual $%.2f to %s\n" % (
#                    invoice.number,
#                    invoice.id,
#                    invoice.date_invoice,
#                    invoice.amount_total,
#                    invoice.residual,
#                    invoice.partner_id.name))
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(cr, 
                                                  uid,
                                                  context,
                                                  case_id,
                                                  problem_writeup)

account_invoice()

class account_move_line(osv.osv):
    _inherit = 'account.move.line'
    
    def sanity_check(self, cr, uid, context=None):
        self.account_unbalanced_debit_credit(cr,uid,context)
        return 0

    def account_unbalanced_debit_credit(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(
            cr,
            uid,
            context,
            "Account Move with a difference between debit and credit",
            "This is a stub to track incidents where account move lines have unbalanced debit and credit",
            'account_unbalanced_debit_credit')

        move_line_ids = self.search(cr, uid, [("state", "!=", "valid")])
        
        if not move_line_ids:
            return # hurrah, no weirdness
        
        problem_writeup = "The following account moves have unblanced debit and credit:\n\n"
        
        moves = dict()

        for move_line in self.browse(cr,uid,move_line_ids):
            if move_line.move_id.name not in moves:
                moves[move_line.move_id.name] = {
                                'date': move_line.date,
                                'partner': move_line.partner_id.name,
                                'tot_debit': float(move_line.debit),
                                'tot_credit': float(move_line.credit)}
            else:
                moves[move_line.move_id.name]['tot_debit'] += float(move_line.debit)
                moves[move_line.move_id.name]['tot_credit'] += float(move_line.credit)
        
        for move_name in moves.iterkeys():
            problem_writeup += "%s (move %s) - %s - credit: %.2f - debit: %.2f\n" % (
                moves[move_name]['partner'],
                move_name,
                moves[move_name]['date'],
                moves[move_name]['tot_credit'], 
                moves[move_name]['tot_debit'])
        
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                cr, uid,
                context,
                case_id,
                problem_writeup)
        return
        
account_move_line()
