from osv import fields
from osv import osv

class mrp_bom(osv.osv):
    _inherit = 'mrp.bom'

    def sanity_check(self, cr, uid, context=None):
        ''' Figure out if there are any top level bom entries that do not 
            have a routing id set
        '''

        self.bom_cycle_check(cr,uid,context)
        self.bom_missing_routings(cr,uid,context)
        self.bom_inactive_products(cr,uid,context)
        return 0

    def bom_missing_routings(self, cr, uid, context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(cr,uid,context,
        'BoM without a Routing Found',
        """ This is a stub to track incidents of BoMs without a routing ID. Without a 
            routing ID the stock moves gets borked.
        """,
        'mrp_bom_missing_routings')

        bom_ids = self.search(cr,uid,[
                        ('bom_id','=',None),
                        ('routing_id','=',None),
                        ])

        if not bom_ids:
            return # hurrah, no weirdness

        problem_writeup = "The following BoMs are missing a routing:\n\n"
        for bom_data in self.browse(cr,uid,bom_ids):
            problem_writeup += "%i: %s\n\n" % ( bom_data.id, bom_data.name )
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                  cr, uid, context,
                  case_id,
                  problem_writeup
              )


    def bom_cycle_check(self, cr, uid, context):
        ''' Hunt for any cycles that might of shown up in the database
        '''

        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(cr,uid,context,
        'BoM Cycles Detected',
        """ This is a stub to track incidents of BoMs with circular references.
            (eg. where the following the BoM will cause it to require itself)
        """,
        'mrp_bom_cycle_check')


        # Build a list of all the from and to bom items
        bom_lookup_by_id = {}
        bom_lookup_by_product_id = {}

        all_ids = self.search(cr, uid, [])
        for bom in self.browse(cr,uid,all_ids):
            bom_id = bom.id
            bom_lookup_by_id.setdefault(bom_id,[])
            if bom.bom_id:
                parent_id = bom.bom_id.id
                bom_lookup_by_id.setdefault(parent_id,[])
                bom_lookup_by_id[parent_id].append({ 
                                                'bom_id': bom_id, 
                                                'product_id': bom.product_id.id 
                                            })
            else:
                bom_lookup_by_product_id[bom.product_id.id] = {
                  'name': bom.product_id.name,
                  'default_code': bom.product_id.default_code,
                  'bom_id': bom_id
                }

        # Now iterate through through all the products (not that this ain't the 
        # most efficient way of going about it)
        circular_references = {}
        for product_id in bom_lookup_by_product_id.keys():
            try:
                product_data = bom_lookup_by_product_id[product_id]
                self._bom_cycle_check(
                              cr,
                              uid,
                              product_data['bom_id'],
                              bom_lookup_by_product_id,
                              bom_lookup_by_id
                          )
            except Exception as ex:
                circular_bom_id, stack = ex.args
                circular_references[circular_bom_id] = stack

        if not circular_references:
            return # horray, no weirdness

        problem_writeup = ""
        for bom_id, stack in circular_references.iteritems():
            # remove the prefix product entries on the stack to get just
            # the entries that represent the cycle
            first_instance = stack.index(bom_id)

            # Now lets get the products' names
            cycle_ids = stack[first_instance:]
            bom_data_for = {}
            for bom_data in self.browse(cr,uid,cycle_ids):
                bom_data_for[bom_data.id] = bom_data

            # Create human readable text describing the problem
            problem_writeup += "There is a cycle on: " + bom_data_for[bom_id].product_id.name+"\n\n"
            step_count       = 1
            for cycle_id in cycle_ids:
                bom_data = bom_data_for[cycle_id]
                problem_writeup += "%3i:%s (%i) [%s] %s\n" % (
                                      step_count,
                                      ( "  " * (step_count-1) ),
                                      bom_data.product_id.id,
                                      bom_data.product_id.default_code,
                                      bom_data.product_id.name, )

                step_count += 1
            problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                  cr, uid, context,
                  case_id,
                  problem_writeup
              )

    def _bom_cycle_check(self, cr, uid, bom_id, bom_lookup_by_product_id, bom_lookup_by_id, stack=None):
        if stack is None:
            stack = []

        # Circular reference check here. If the current bom_id is found in the stack,
        # it means we've already seen this bom_id and we've just looped back through
        # the bom back to it. So, throw error.
        if bom_id in stack:
            raise Exception(bom_id,stack)
        stack.append(bom_id)

        # Depth first handling here
        for component in bom_lookup_by_id[bom_id]:
            child_bom_id         = component['bom_id']
            component_product_id = component['product_id']

            # If there is no BoM associated with the bom id, check if we have
            # a bom for that particular product. Failing both we assume it's
            # a base component
            if child_bom_id not in bom_lookup_by_id or not bom_lookup_by_id[child_bom_id]:
                if component_product_id in bom_lookup_by_product_id:
                    child_bom_id = bom_lookup_by_product_id[component_product_id]['bom_id']
                else:
                    continue

            self._bom_cycle_check( 
                          cr,
                          uid,
                          child_bom_id,
                          bom_lookup_by_product_id,
                          bom_lookup_by_id,
                          list(stack)
                      )


    def bom_inactive_products(self, cr, uid, context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(cr,uid,context,
        'BoM with inactive products found',
        """ This is a stub to track incidents of BoMs that link to inactive products.
            Inactive products in BoMs can cause problems to certain statistics 
            generation calculations.
        """,
        'mrp_bom_inactive_products')

        # Find a list of all the products that are inactive yet found in BoMs
        cr.execute('''
          SELECT DISTINCT pp.id
          FROM  mrp_bom mb left join
                product_product pp on mb.product_id = pp.id 
          WHERE pp.active = false
          AND   mb.active = true
          ;
        ''')
        prod_ids = [row[0] for row in cr.fetchall()]
        if not prod_ids:
            return # horray, no weirdness

        # Find a list of all the BoMs that contain any of the inactive products
        #           0      1        2       3         4
        cr.execute('''
          SELECT    mb.id, mb.name, pmb.id, pmb.name, pp.id
          FROM      mrp_bom mb left join
                    product_product pp on mb.product_id = pp.id left join
                    mrp_bom pmb on mb.bom_id = pmb.id
          WHERE     pp.active = false
          AND       mb.active = true
          ORDER BY  pmb.name, mb.name
        ''')

        bad_boms_by_product = {}
        for row in cr.fetchall():
            bad_boms_by_product.setdefault(row[4],[])
            bom_description = ""
            if row[2]: # If we have a parent BoM
                bom_description = "\t* Parent BoM ID "+str(row[2])+": "+row[3]
            else: # If the BoM is for an inactive product...
                bom_description = "\t* Product BoM ID "+str(row[0])+": "+row[1]
            bad_boms_by_product[row[4]].append( bom_description )

        problem_writeup = "The following inactive products have been found in BoMs:\n\n"
        product_name_lookup = {}
        for prod in self.pool.get('product.product').browse(cr,uid,prod_ids):
            problem_writeup += "%i:(%s) %s\n" % (prod.id,prod.default_code,prod.name)
            product_name_lookup[prod.id] = "["+prod.default_code+"] "+prod.name
            problem_writeup += "\n".join(bad_boms_by_product[prod.id])
            problem_writeup += "\n"

        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                  cr, uid, context,
                  case_id,
                  problem_writeup
              )


mrp_bom()

class procurement_order(osv.osv):
    _inherit = 'procurement.order'

    def sanity_check(self, cr, uid, context=None):
        ''' Look for exception procurements
        '''
        self.procurement_exception_sanity(cr,uid,context)
        return 0

    def procurement_exception_sanity(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(cr,uid,context,
        'Exception Procurement to Fix Found',
        """ This is a stub to track incidents of Procurements that end up in Exception to Fix
        """,
        'mrp_procurement_exception_sanity')

        proc_ids = self.search(cr,uid,[('state','=','exception'),('message','<>','')])
        if not proc_ids:
            return # no errors

        problem_writeup = "The following procurements are in exception state:\n\n"
        for proc_data in self.browse(cr,uid,proc_ids):
            problem_writeup += "%i: %s '%s'\n\n" % ( proc_data.id, proc_data.name, proc_data.message )
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                  cr, uid, context,
                  case_id,
                  problem_writeup
              )
procurement_order()

class mrp_production(osv.osv):
    _inherit = 'mrp.production'

    def sanity_check(self, cr, uid, context=None):
        ''' Look for exception production orders
        '''
        self.production_exception_sanity(cr,uid,context)
        self.production_end_of_lifecycle(cr, uid, context)
        return 0

    def production_exception_sanity(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(cr,uid,context,
        'Exception Production Order to Fix Found',
        """ This is a stub to track incidents of production orders that end up in Exception.
        """,
        'mrp_production_exception_sanity')

        prodn_ids = self.search(cr,uid,[('state','=','picking_except')])
        if not prodn_ids:
            return # no errors

        problem_writeup = "The following production orders are in exception state:\n\n"
        for production in self.browse(cr,uid,prodn_ids):
            problem_writeup += "%i: %s\n\n" % ( 
                production.id, 
                production.name)
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
                  cr, 
                  uid, 
                  context,
                  case_id,
                  problem_writeup
              )

    def production_end_of_lifecycle(self,cr,uid,context):
        case_id = self.pool.get('crm.helpdesk').sanity_test_create_stub_case(
            cr,
            uid,
            context,
            'Production orders waiting for outdated components',
            """\
This is a stub to track incidents of production orders that are waiting for
components marked "end of lifecycle".
""",
            'mrp_production_end_of_lifecycle')

        # Find all production orders that are waiting for goods.
        prodn_ids = self.search(cr,uid,[('state','=','confirmed')])
        if not prodn_ids:
            return # no possible errors

        problems = {} # prodn name -> ["qty x [code] name (revision N)", ...]
        for production in self.browse(cr, uid, prodn_ids):
            component_problems = []
            for line in production.picking_id.move_lines:
                if (    line.product_id and 
                        line.product_id.state == 'end' and
                        line.state == 'confirmed'):
                    description = ('%.2f x [%s] %s (revision %d)' %
                                   (line.product_qty,
                                    line.product_id.default_code,
                                    line.product_id.name,
                                    line.product_id.revision))
                    component_problems.append(description)
            if component_problems:
                problems[production.name] = component_problems
        
        if not problems:
            return # no errors
        
        names = sorted(problems.keys())
            
        problem_writeup = """\
The following production orders are waiting for components marked 
"end of lifecycle".

"""
        for name in names:
            component_problems = problems[name]
            problem_writeup += "%s: %s\n" % (name, component_problems[0])
            for problem in component_problems[1:]:
                problem_writeup += "%s  %s\n" % (len(name) * ' ', problem)
        problem_writeup += "\n\n"

        # now that we've finished documenting the problem let's notify the 
        # appropriate users
        self.pool.get('crm.helpdesk').sanity_test_log(
            cr, 
            uid, 
            context,
            case_id,
            problem_writeup)
        
mrp_production()
