# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution	
#    Copyright (C) 2004-2012 Zaber Technologies Inc. (<http://zaber.com>). All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import fields,osv,orm
from pysolr import Solr, SolrError
from tools import config
import netsvc
"""

Helper module to:

- trap the instance when Solr has not been configured.
    (Instead of failing edits, just throw a warning message into the logs.)

- decorator for osv.osv classes to add support for solr actions

- an osv object to track changes to tables in the case that we don't want
  to immediately update the solr index

"""


def conn():
    solr_url = config.get('solr_url',False)

    if not solr_url:
        logger = netsvc.Logger()
        logger.notifyChannel(
            "addons.zerp_solr", 
            netsvc.LOG_WARNING,
            "zerp_solr requires a Solr installation with configuration. "
              + "See addons/zerp_solr/docs/usage.txt for more information."
        )
        return None

    return Solr(solr_url)


# Solr requires a limit to be set otherwise it defaults to 10. We'll default to 1 million. 
def search(dbname,record_type,query,limit=1000000):
    try:
        solr = conn()
        if not solr: return []
        results = solr.search(
                    query,
                    rows=limit,
                    fl="id",
                    fq='record_type:"%s-%s"'%(dbname,record_type),
                  )
    except SolrError as e:
        return []
    ids = [ r['id'] for r in results ]
    return ids


def add(records):
    solr = conn()
    if not solr: return
    to_index = []
    for record in records:
        if not ( record.get('id',False) and
                 record.get('record_type',False) ):
            continue
        to_index.append(record)
    if to_index: 
        solr.add(to_index)


def delete(query):
    solr = conn()
    if not solr: return
    return solr.delete(q=query)


def delete_byid(id):
    solr = conn()
    if not solr: return
    return solr.delete(id=id)

#################################################################################
## So we don't need to make lots of trips to the solr database while doing
## mass updates to records
#################################################################################

class zerp_solr_update_queue (osv.osv):
  _name = 'zerp.solr.update.queue'
  _columns = {
    'mode': fields.selection([
                ('update','Update requested'),
                ('delete', 'Delete requested'),
              ],'Update Mode'),
    'model': fields.char('Model',size=100,required=True),
    'res_id': fields.integer('ID'),
  }

  def _queue_updates(self,cr,uid,model,mode,ids):
      """ A quick way of queuing up a request to have records
          updated/deleted from the solr index in the future. This 
          allows us to hook into the ORM update code even in
          relatively time sensitive portions.
      """

      # Occasionally we might get straight up IDs rather than
      # arrays of IDs. So we convert them into something to be 
      # easily iterated here
      if isinstance(ids, (int, long)): ids = [ ids ]
      for res_id in ids:
          self.create(cr,uid,{
            'mode': mode,
            'model': model,
            'res_id': res_id
          })

  def _run_updates(self,cr,uid,context=None):
    ids = self.search(cr,uid,[])
    if not ids: return

    # We do this part so we can batch updates by model. This allows
    # use to be a bit more efficient in our loading and push of data
    # to solr
    records_for_update = {}
    records_for_delete = {}
    for req in self.browse(cr,uid,ids):
        model = req.model
        mode = req.mode
        if mode == 'update':
          records_for_update.setdefault(model,set()).add(req.res_id)
        elif mode == 'delete':
          records_for_delete.setdefault(model,set()).add(req.res_id)
        else: continue

    # Then, run the actual updates
    for model, res_ids in records_for_update.iteritems():
        res_ids = list(res_ids)
        obj = self.pool.get(model)
        obj.solr_reindex_records(cr,uid,res_ids)
        self.unlink(cr,uid,
            self.search(cr,uid,[
              ('model','=',model),
              ('mode','=','update'),
              ('res_id','in',res_ids),
            ])
        )

    # And seperately, remove records. We may reload 
    # the model twice but we assume that this function isn't 
    # called in the most time-sensitive portion of the code
    for model, res_ids in records_for_delete.iteritems():
        res_ids = list(res_ids)
        obj = self.pool.get(model)
        obj.solr_deindex_records(cr,uid,res_ids)
        self.unlink(cr,uid,
            self.search(cr,uid,[
              ('model','=',model),
              ('mode','=','delete'),
              ('res_id','in',res_ids),
            ])
        )

zerp_solr_update_queue()

#################################################################################
## Decorator to add solr support to objects
#################################################################################

def add_solr_support(opts=None,autoupdate=True):

    if opts == None:
        opts = {}

    def _add_solr_support(cls):
        """ Add the base functions/methods to an osv object to support
            solr based keyword searching. Just adding the functions isn't
            enough, however, a method "solr_record" must be defined to
            figure out how to convert the object into useful data for Solr
        """


        def extend_cls(f):
            # Use decorator on function to decorate the class
            fname = f.__name__
            if getattr(cls,fname, False):
                fname = "_"+fname
            setattr(cls,fname,f)
            return f

        def solr_record(self,cr,rec,solr_rec):
            print "Requires definition!"
            return {}
        if not getattr(cls,'solr_record',False):
            cls.solr_record = solr_record

        @extend_cls
        def _solr_record(self,cr,rec):
            solr_rec = {
              'record_type': '%s-%s' % (cr.dbname,self._name),
              'id': "%s-%s:%d"%(self._name,cr.dbname,rec.id,),
            }
            return self.solr_record(cr,rec,solr_rec)

        @extend_cls
        def solr_reindex(self,cr,uid,context=None):
            delete('record_type:"%s-%s"'%(cr.dbname,self._name))
            batch_size = 1000 # reindex in batches of 1000
            offset = 0

            # Knock out any queued updates we were going to perform for this table
            update_queue_obj = self.pool.get('zerp.solr.update.queue')
            update_queue_obj.unlink(cr,uid,
                update_queue_obj.search(cr,uid,[('model','=',self._name)])
            )

            # Then go through each record in the table and reindex it in batches
            while True:
                ids = self.search(cr,uid,[],offset=offset,limit=batch_size)
                if not ids: break
                if type(ids) == int: ids = [ids]
                offset += batch_size
                prod_records = [ self._solr_record(cr,prod) for prod in self.browse(cr,uid,ids) ]
                add(prod_records)

        @extend_cls
        def solr_reindex_records(self, cr, uid, ids):
            if isinstance(ids, (int, long)):
                ids = [ids]
            filtered_ids = self.search(cr,uid,[
                                ('id','in',ids)
                              ])
            prod_records = [ 
              self._solr_record(cr,prod) for prod in self.browse(cr,uid,filtered_ids) 
            ]
            add(prod_records)

        @extend_cls
        def solr_deindex_records(self, cr, uid, ids):
            for id in ids:
                delete_byid("%s-%s:%d"%(self._name,cr.dbname,id))

        # Update Solr as well when records are modified
        def _autoupdate_write(self, cr, uid, ids, *args, **kwargs):
            res = super(cls,self).write(cr,uid,ids,*args,**kwargs)
            self.solr_reindex_records(cr,uid,ids)
            return res

        def _autoupdate_create(self, cr, uid, *args, **kwargs):
            id = super(cls,self).create(cr,uid,*args,**kwargs)
            if id: self.solr_reindex_records(cr,uid,[id])
            return id

        def _autoupdate_unlink(self, cr, uid, ids, *args, **kwargs):
            res = super(cls,self).unlink(cr,uid,ids,*args,**kwargs)
            self.solr_deindex_records(cr,uid,ids)
            return res

        def _queue_write(self, cr, uid, ids, *args, **kwargs):
            res = super(cls,self).write(cr,uid,ids,*args,**kwargs)
            self.pool.get('zerp.solr.update.queue')._queue_updates(cr,uid,self._name,'update',ids)
            return res

        def _queue_create(self, cr, uid, *args, **kwargs):
            id = super(cls,self).create(cr,uid,*args,**kwargs)
            if id:
              self.pool.get('zerp.solr.update.queue')._queue_updates(cr,uid,self._name,'update',id)
            return id

        def _queue_unlink(self, cr, uid, ids, *args, **kwargs):
            res = super(cls,self).unlink(cr,uid,ids,*args,**kwargs)
            self.pool.get('zerp.solr.update.queue')._queue_updates(cr,uid,self._name,'delete',ids)
            return res


        if autoupdate:
            cls.write = _autoupdate_write
            cls.create = _autoupdate_create
            cls.unlink = _autoupdate_unlink
        else:
            cls.write = _queue_write
            cls.create = _queue_create
            cls.unlink = _queue_unlink

        # Hook for pulling Solr results
        @extend_cls
        def _keyword_search(self, cr, uid, obj, name, args, context=None):
            query_terms = []
            for arg in args:
                if arg[1] == 'ilike':
                    query_terms.append(arg[2])

            query_string = " ".join(filter(None,query_terms))

            res = search(cr.dbname,self._name,query_string)

            if not len(res):
                return [('id', '=', '0')]

            id_list = [int(r.partition(':')[2] or 0) for r in res]

            return [('id', 'in', id_list)]

        cls._columns = {
            'keyword': fields.function( 
                                    lambda *x: {},
                                    fnct_search=_keyword_search,
                                    string='Advanced Search',
                                    type='char' )
        }

        return cls
    return _add_solr_support

if __name__ == '__main__':
    import doctest
    failures, tests = doctest.testmod()

