import netsvc
from osv import fields, osv

class confirm_warning(osv.osv_memory):
    _name = "confirm.warning"
    _description = "Confirm Warning"

    def check_warning(self, 
                      cr, 
                      uid, 
                      title, 
                      partner_name, 
                      message,
                      model,
                      res_id,
                      signal,
                      warning_level, 
                      missing_taxes=False, 
                      context=None):
        ''' Create a warning record with extra details in the message.
        
            Usually sends user down workflow to confirm order 
            but 
            If the user has created a Canadian order that doesn't 
            have any taxes on it, or if the partner has some
            warnings, sends the workflow into
            the warning state or the blocked state.
            Also loads the form view with any warning messages in it.
            
            title is the title for the dialog box.
            partner_name is for the partner on the order.
            message is the basic warning message. It will get added to.
            model is the model to trigger workflow for.
            res_id is the record id to trigger workflow for.
            signal is the signal to send to the workflow.
            warning_level is 'no-message', 'warning', or 'block'.
            missing_taxes is True if you want to display a warning about 
                Canadian taxes, otherwise false.
        '''
        can_confirm = (warning_level == 'no-message')
        if can_confirm and missing_taxes:
            can_confirm = False
            
        if can_confirm:
            wf_service = netsvc.LocalService("workflow")
            wf_service.trg_validate(uid, 
                                    model, 
                                    res_id, 
                                    signal, 
                                    cr)
            return False

        full_message = ''
        if not partner_name:
            full_message += message + '\n'
        else:
            if warning_level == 'warning':
                full_message += '%s has a warning:\n    %s\n' % (partner_name, message)
            elif warning_level == 'block':
                full_message += '%s is blocked:\n    %s\n' % (partner_name, message)
            
        if missing_taxes:
            warning_level = 'warning'
            full_message += ('Missing taxes:\n    Some or all lines on this ' +
                             'Canadian order do not have any taxes ' +
                             'associated.\n')
        
        data = dict(message=full_message,
                    level=warning_level,
                    model=model,
                    res_id=res_id,
                    signal=signal)
        warning_id = self.create(cr, uid, data, context=context)
        return {'name': title,
                'view_mode': 'form',
                'view_id': False,
                'view_type': 'form',
                'res_model': self._name,
                'res_id': warning_id,
                'type': 'ir.actions.act_window',
                'nodestroy': True,
                'target': 'new',
                'domain': '[]',
                'context': context}

    def confirm_order(self, cr, uid, ids, context=None):
        ''' Confirm the order since we've got the okay to do so
        '''
        wf_service = netsvc.LocalService("workflow")
        for warning in self.browse(cr, uid, ids, context):
            wf_service.trg_validate(uid, 
                                    warning.model, 
                                    warning.res_id, 
                                    warning.signal, 
                                    cr)
            
        return {'type': 'ir.actions.act_window_close'}
    
    _columns = {'message': fields.text('Message', readonly=True),
                'level': fields.selection(
                    (('no-message', 'No Message'),
                     ('warning', 'Warning'),
                     ('block', 'Blocking Message')),
                    'Warning Level'),
                'model': fields.char('Model', size=64),
                'res_id': fields.integer('Resource Id'),
                'signal': fields.char('Signal', size=64)}
    _defaults = {'level': 'none'}

confirm_warning()
